#include <machine.h>
#include <errno.h>
#include <stdio.h>
#if	UNIX
#	include <unistd.h>
#else	//UNIX
#	include	<io.h>
#endif	//UNIX
#include <stdlib.h>
#include <error.h>
#include <win_io.h>

#define MINOF(a, b)	((a) <= (b) ? (a) : (b))

#ifndef APLTIME_DEFAULT
#define APLTIME_DEFAULT 2
#endif	// APLTIME_DEFAULT 

#ifdef	XAPL
int		jwm_proc_key;
#endif	// XAPL 

#if defined(MAINWIN) || WINE
#	if	GUI
#		undef BELL
#		undef BSD
#	endif//GUI
#endif // MAINWIN 

#ifdef	BELL
#include <fcntl.h>
#if	defined(LINUX)
#include <termios.h>
#else //defined(LINUX)
#include <termio.h>
#endif//defined(LINUX)
#include <sys/ioctl.h>
#endif	// BELL 

#ifdef	BSD
#include <sys/time.h>
#endif	// BSD 

#ifdef VERSION7
#include <sys/signal.h>
#include <setjmp.h>
static jmp_buf	tmbuf;
#endif // VERSION7 

#include <jumps.h>
#include <writloss.h>

#include <unix.h>
#if	AP_BUILD
	extern gchar* W3getenv(const gchar *envname);
	extern void W3Error(gchar *buf,...);
	extern int W3_timed_read(int fd,int wait);
	extern void W3_timed_ansi(int fd,int wait,int *value,int *ansi);
	extern void trans_message_box(gchar *message);
	extern size_t getcmdprofile(const gchar *name,gchar * buf,size_t buflen);
	extern size_t getglobalenv( const gchar *name,gchar * buf,size_t buflen);
#else //AP_BUILD
#	include <w3fsio.h>
#	include <embed.h>
#	include <util.h>
#	include <thread.h>
#endif//AP_BUILD

static int uninterrupted_read(int fd, UCH *buf, unsigned wanted);
extern int indq(void);

#ifdef	BSD

int timed_read(fd, wait)
int fd;
int wait;
{
	/*
		some machines show problems here especially when the
		keyboard is connected remotely via a network.
		In these circumstances it is not possible to set a time
		in which we can guarrentee to have collected all of the
		characters transmitted by a single key stroke.
	Alpha error:
		Users single keystroke is interpretted as multiple key strokes.
	Beta error:
		Users multiple key strokes are interpretted as a single stroke.

		The intention is to err on the side of the beta error on the
		basis that there is a low probability of actually typing
		the character sequence transmitted by a special key. Some
		terminals do however use 'ESC A' which is a common thing
		to type at vi.
	*/
	static struct timeval timeout;
	static UCH time_out;		// delay in 1/10ths sec		
	register int count = 0;
	int input = 1 << fd;			// mask for stream	
	UCH ch;

#ifdef XAPL
	if(isanX(fd))
		{
		int	rslt;
		jwm_proc_key = 0;		// not processing 
		rslt =  X_timed_read(fd, wait);
		jwm_proc_key = 1;		// go'n'process key 
		return rslt;
		}
#endif // XAPL 

	if(!time_out)	// first time through 
		{
		UCH *ep;

		ep = (void *)getenv("APLTIME");
		if(ep)time_out = atoi(ep);
		if(!time_out)time_out = APLTIME_DEFAULT;
		timeout.tv_sec = time_out / 10;
		timeout.tv_usec = (time_out % 10) * 100000;
		}
	if(wait || select(1, &input, 0, 0, &timeout) > 0)
		{
		if(1 == uninterrupted_read(fd, &ch, 1))
			return (int)ch & 0xff;
		else
			{
			switch(errno)
				{
			default:
				return -1;	//EOF
			/* case EINTR: */ #include	<eintr.h>
				break;
				}
			return -3;	// INTERRUPTED 
			}
		}
	return -2;			// TIMEOUT 
}

#endif	// BSD 

#ifdef	BELL

/*
 *	The BELL (System III or System V) version is much trickier.
 *	Firstly, if 'wait' is true then we can just issue a read for
 *	one character (the caller must already have set -ICANON if
 *	stdin is a tty).
 *	Otherwise, ('wait' is false) we have three alternative methods
 *	of implementing a non-blocking read.
 *	(1)	If there is a way of finding how many characters are
 *		queued we can use this to find whether a read would
 *		block or not, and if there are some characters ready
 *		then we just issue a read.
 *		NOTE: this section of code is currently commented out
 *		because it doesn't work on the CADMUS and the ALTOS.
 *		On the CADMUS, using FIONREAD causes characters to be
 *		lost; on the ALTOS, TCRDCHK appears to return a random
 *		number 0..15.
 *	(2)	We set O_NDELAY mode and issue a read.
 *	(3)	If stdin is a tty, then as a last resort we set MIN to
 *		0 and TIME to 0.1 sec, and issue a read.  This method
 *		may be necessary on a slow TTY line, where the read with
 *		O_NDELAY set may report no characters because the next
 *		character of the key sequence has not yet arrived from
 *		the terminal.  This method allows up to 100 ms between
 *		characters before a sequence is considered finished.
 */

#ifdef	ATT3B2			// has problems if TCSETA is used	
#define NBUF	1		// this may need to be increased to	
#define TCSETOP TCSETAW		// compensate for using TCSETAW		
#endif	// ATT3B2 

#ifndef TCSETOP			// default if not defined above		
#define NBUF	1		// buffer size 1 gives smooth response	
#define TCSETOP TCSETA		// use TCSETA unless it gives trouble	
#endif	// TCSETOP 

int timed_read(int fd, int wait)
{
	static UCH time_out;		// delay in 1/10ths sec		
	static UCH	buff[NBUF], *bfptr;
	static int		count = 0;
	int			flagcontext;
	struct termio		ttycontext,
				ttynodelay;
	int fail = -1;	// set to EOF until we are sure we fail on a tty 

#ifdef XAPL
	if(isanX(fd))
		{
		int	rslt;
		jwm_proc_key = 0;		// not processing 
		rslt =  X_timed_read(fd, wait);
		jwm_proc_key = 1;		// go'n'process key 
		return rslt;
		}
#endif // XAPL 

	if(count)
	{
		--count;
		return *bfptr++;
	}

	if(wait)				// willing to wait	
	{
#if	AP_BUILD
#else //AP_BUILD
		if(0 == pipelen(fd))
			{
			errno = 0;
			tswitch_at_descalc(fd, TSWITCH_ON);
			if
				(
				strong
			||	(weak && indq())
				)switch(errno)
				{
			/* case EINTR: */ #include	<eintr.h>
				count = 0;
				goto DONE;	// EINTR from the select()
				}
			}
		tswitch_at_descalc(fd, TSWITCH_OFF);
#endif//AP_BUILD
		count = uninterrupted_read(fd, buff, 1);
		goto DONE;
	}

	flagcontext = fcntl(fd, F_GETFL, 0);
	fcntl(fd, F_SETFL, flagcontext|O_NDELAY);
	count = uninterrupted_read(fd, buff, NBUF);// read with no delay
	fcntl(fd, F_SETFL, flagcontext);
	if(count)				// got some characters	
		goto DONE;

	if(ioctl(fd, TCGETA, &ttycontext) == 0)	// then it's a tty	
	{
		fail = -2;	// set to TIMEOUT 
		if(!time_out)	// first time through 
			{
			UCH *ep;

			ep = (void *)getenv("APLTIME");
			if(ep)time_out = atoi((char*)ep);
			if(!time_out)time_out = APLTIME_DEFAULT;
			}
		ioctl(fd, TCGETA, &ttynodelay);	// copy tty structure	
		ttynodelay.c_lflag &= ~ICANON;	// make sure ICANON off 
		ttynodelay.c_cc[VMIN] = 0;		// minimum characters	
		ttynodelay.c_cc[VTIME] = time_out;	// timeout		
		ioctl(fd, TCSETOP, &ttynodelay); // set nodelay modes	
		count = uninterrupted_read(fd, buff, NBUF);// get some chars
		ioctl(fd, TCSETOP, &ttycontext); // restore normal modes 
	}

DONE:
	bfptr = buff;
	if(count > 0)
		return  (int)(--count, *bfptr++) & 0xff;
	switch(errno)
		{
	default:
		return fail;
	/* case EINTR: */ #include	<eintr.h>
		break;
		}
	return -3;
}

#endif	// BELL 

#if	defined(DOS)
#	include <hardware.h>
#	if	defined(HIGHC)
#		include <system.cf>
#		include <msdos.cf>
#		define	R_AL	Registers.AX.LH.L
#		define	R_AH	Registers.AX.LH.H
#		define	R_AX	Registers.AX.W
#		define	R_EAX	Registers.AX.R
#		define	R_BL	Registers.BX.LH.L
#		define	R_BH	Registers.BX.LH.H
#		define	R_BX	Registers.BX.W
#		define	R_EBX	Registers.BX.R
#		define	R_CL	Registers.CX.LH.L
#		define	R_CH	Registers.CX.LH.H
#		define	R_CX	Registers.CX.W
#		define	R_ECX	Registers.CX.R
#		define	R_DL	Registers.DX.LH.L
#		define	R_DH	Registers.DX.LH.H
#		define	R_DX	Registers.DX.W
#		define	R_EDX	Registers.DX.R
#		define	R_ESI	Registers.SI.R
#	elif	defined(WATCOM)
#		include <dos.h>
		union REGPACK	Registers;
#		define		callint(x)	intr(x, &Registers)
#		define		calldos()	intr(0x21, &Registers)
#		define	R_AL	Registers.h.al
#		define	R_AH	Registers.h.ah
#		define	R_AX	Registers.w.ax
#		define	R_EAX	Registers.x.eax
#		define	R_BL	Registers.h.bl
#		define	R_BH	Registers.h.bh
#		define	R_BX	Registers.w.bx
#		define	R_EBX	Registers.x.ebx
#		define	R_CL	Registers.h.cl
#		define	R_CH	Registers.h.ch
#		define	R_CX	Registers.w.cx
#		define	R_ECX	Registers.x.ecx
#		define	R_DL	Registers.h.dl
#		define	R_DH	Registers.h.dh
#		define	R_DX	Registers.w.dx
#		define	R_EDX	Registers.x.edx
#		define	R_ESI	Registers.x.esi
#	endif	// defined(WATCOM) 

#if	defined(HIGHC)
static int COM_lcom(void)
{
// checks for the presense of the LCOM.EXE TSR 
// see the code for iscom_a1.c in the Blaise stuff 

struct
	{
	unsigned short offset;
	unsigned short seg;
	} real_farp;
unsigned real_off;
unsigned str;

	{
	FARP from = init_FARP( 4*0x66, 0x60 ); 	// the real mode interrupt 0x66 
	FARP to   = init_FARP( (unsigned)&real_farp, MYDATA() );

	COPYSEG(&to, &from, 4);
	real_off = (real_farp.seg << 4) + real_farp.offset;
	}

	{
	FARP from = init_FARP( real_off+2, 0x60 );// pointer 2 bytes into Blaise code 
	FARP to   = init_FARP( (unsigned)&real_farp, MYDATA() );

	COPYSEG(&to, &from, 4);
	real_off = (real_farp.seg << 4) + real_farp.offset;
	}

	{
	static char match[] = "Blaise Computing Inc";
	char test[20];
	FARP from = init_FARP( real_off + 21, 0x60 ); // follow that pointer (taxi?) 
	FARP to   = init_FARP( (unsigned)test, MYDATA() );

	COPYSEG(&to, &from, sizeof(test));
	return strncmp(match, test, sizeof(test)) ? 0 : 1;
	}
}

static int COM_chk(int in, int out)
{
	int total = in + out;

	if(total & 0xf) total += 16; total &= ~0xf; // whole paragraphs 
	if(in & 0xffff0000) LIMIT_ERROR;	// Blaise cant handle > 64K 
	if(out & 0xffff0000) LIMIT_ERROR;	// Blaise cant handle > 64K 
	if(total & 0xffff0000) LIMIT_ERROR;	// Blaise cant handle > 64K 
	if(!COM_lcom()) DOMAIN_ERROR;		// LCOM.EXE not there 
	return total;
}

static int COM_buf = 0;

void COM_open(void)
{
	int aplcomi = atoi((void *)getenv("APLCOMI"));
	int aplcomo = atoi((void *)getenv("APLCOMO"));
	int total = COM_chk(aplcomi, aplcomo);
	int realbuf = INT21_25C0(total>>4);
	PARAM_2511 param_2511 = init_PARAM_2511( 0x66, realbuf, 0, 0, 0, 0x100, 0 );

	/*
	* This gets dirty - Blaise will assume that IRQ 4 is mapped onto 
	* interrupt 0xc. Phar-Lap has moved it to interrupt 0x7c (usually).
	*
	* For this code to run properly then it is essential that 
	*	cfig386 apl.exe -intmap 8
	* has been executed.
	*
	* I check for it and issue a DOMAIN_ERROR if this is not the case.
	*
	* When APL exits it will have to check if COM1 has been opened and if so
	* it will have to close it
	*/

	if(realbuf == -1) LIMIT_ERROR;	// not enough real memory 
	COM_buf = realbuf;
	R_EAX = 0x2511;
	R_EDX = (unsigned)&param_2511;
	R_ESI = 0;
	R_EBX = aplcomi;	// input queue  
	R_ECX = aplcomo;	// output queue 
	calldos();
	if(param_2511.eax & 0xff) domain();

	{
	unsigned irq0, dummy1, dummy2;
	int dummy3 = INT21_250C(&irq0, &dummy1, &dummy2); // where is IRQ0 ? 

	if(irq0 != 8)	// apl is running in normal PharLap mode or under DESQVIEW 
		domain();
	}

	R_AX = 0x104;		// set option 
	R_BX = 0x106;		// enable local flow control 
	callint(0x66);
	if(R_AL) domain();
	R_AX = 0x104;		// set option 
	R_BX = 0x105;		// enable remote flow control 
	callint(0x66);
	if(R_AL) domain();
	return;
}

void COM_close(void)
{
if(COM_buf)
	{
	retry:
		/*
		 * Is there anything waiting to go out ?
		 */
		R_AX = 0x102;		// write COM1 
		R_BX = 0;		// transfer a string 
		R_CX = 0;		// of 0 bytes 
		callint(0x66);
		switch(R_AL)
			{
		case 3:		// port not yet open 
				// nobody has opened it 
			break;
		case 0:		// result OK 
			if(R_BX) goto retry;	// output not flushed 
			R_AX = 0x101;	// close COM1 
			callint(0x66);
			break;
		default:
			// strange condition - ignore it 
			R_AX = 0x101;	// close COM1 
			callint(0x66);
			break;
			}
	INT21_25C1(COM_buf);	// free it 
	COM_buf = 0;
	}
}

static void COM_ok(void)
{
if(COM_buf)		// already open 
	{
	int MCR = INP(0x3fc);	// get value of MODEM control register 
/*
 * If I use COM1 from apl and then do something with COM1 from another process
 * then this bit (OUT2) of the MCR register on the 8250 seems to get clobbered.
 * Why does Blaise need it ?
 * Why does it get clobbered ?
 *
 * I have chosen to completely reset the communication if and when this happens.
 * I considered just forcing the bit ON but felt that resetting was more 
 * defensive.
 */

	if(0 == (MCR&8)) // clobbered by something else ? 
		{
		COM_close();
		COM_open();
		}
	}
else
	COM_open();
}

int COM_status(int fd)
{
	COM_ok();
	R_AX = 0x103;		// read COM1 
	R_BX = 0;		// transfer a string 
	R_CX = 0;		// of 0 bytes 
	callint(0x66);
	switch(R_AL)
		{
	case 0:		// result OK 
		return 0 != R_BX;	// test queue size 
	case 10:	// input queue is empty 
		return 0;
	default:
		domain();
		}
}

int COM_input(int fd)
{
	COM_ok();
	R_AX = 0x103;		// read COM1 
	R_BX = 0x100;		// transfer single byte 
	callint(0x66);
	switch(R_AL)
		{
		case 0:
			return R_AH & 0xff;
		case 7: case 10:
			/*
			 * this should not happen if COM_status() has said
			 * that there is something to read
			 */
			domain(); // probably more useful than syserror() 
		case 3:
			COM_open();
			return COM_input(fd);
		}
}

int COM_output(int byte)
{
	int ret;

	COM_ok();
	R_AX = 0x102;		// write COM1 
	R_BH = 1;		// transfer single byte 
	R_BL = byte;
	callint(0x66);
	switch(ret = R_AL)
		{
		case 0: case 7: return ret;
		case 3:
			COM_open();
			return COM_output(byte);
		}
/*
 * return codes
 *	0 - OK
 *	7 - Output Queue is full
 */
}

#endif	// defined(HIGHC) 

int FD_status(int fd)
{
	R_AH = 0x44;
	R_AL = 0x6;
	R_BX = fd;
	calldos();
	return R_AL != 0;
}

int FD_input(int fd)
{
	char res = 0;
	read(fd, &res, 1);
	return res & 0xff;
}

int KB_status(int fd)
{
	R_AH= 0xB;
	calldos();
	return  0 != R_AL;
}

int KB_input(int fd)
{
	UCH res;

	R_AH = 0x8;
	calldos();
	res = R_AL;
	return res ? res & 0xff : 128;
}

void KB_flush(int fd)
{
	if(is_user_io(fd))
		while(KB_status(fd))
			KB_input(fd);
}

int timed_read(fd, wait)
int fd;
int wait;
{
int c = -2;	// mark it as time out 
static unsigned fromdisk;
static int lastfd = -1;
static int (*status)(int fd), (*input)(int fd);

if(lastfd != fd)
	{
	fromdisk = isadisk(lastfd = fd);
	switch(fd)
		{
	default:
		status = FD_status;
		input = FD_input;
		break;
	case 0:
		status = KB_status;
		input = KB_input;
		break;
	case 3:
		status = COM_status;
		input = COM_input;
		break;
		}
	}
if(wait)for(; !(*status)(fd) ;)	// do a busy wait 
	{ 			// - who else wants the CPU anyway 
	if(fromdisk) return -1;	// not the keyboard -> EOF 
	(void)strong_test();
	if(alarm_test() | weak_test()) return -3; // NB calls both funs 
	}
	
if(wait || (*status)(fd)) c = (*input)(fd);
return c;
}

#endif	// DOS 

#ifdef VERSION7

int timed_read(fd, wait)
int fd;
int wait;
{

	/*
	 *	Routines written using signals and alarm calls only.
	 *	We hope this will work on Version 7 machines.
	 *	This works on the Apollo.
	 */
	register int count = 0;
	UCH ch;
	unsigned curalrm;
	SIGTYPE (*curfn)(),alarmfn();
	static UCH time_out;		// delay in 1/10ths sec		

	if(!time_out)	// first time through 
		{
		UCH *ep;

		ep = (void *)getenv("APLTIME");
		if(ep)time_out = atoi(ep);
		if(!time_out)time_out = APLTIME_DEFAULT;
		}

	if(wait)
		{
		count = uninterrupted_read(fd, &ch, 1);
		if(0 == count) return -1;	// EOF 
		}
	else
		{
		if (setjmp(tmbuf))
			{
			signal(SIGALRM, curfn);
			if (curalrm >  1)
				alarm(curalrm-1);
			else if(curalrm == 1)
				// signal kill to calling process 
				kill(getpid(), SIGALRM);
			}
		else
			{
			curfn = signal(SIGALRM, alarmfn);
			curalrm = alarm(1);
			count = uninterrupted_read(fd, &ch, 1);
			alarm(curalrm);	// my clock loses time 
			signal(SIGALRM, curfn);
			}
		}

	if(count > 0)
		return  (int)ch & 0xff;
	switch(errno)
		{
	default:
		return -2;
	/* case EINTR: */ #include	<eintr.h>
		break;
		}
	return -3;
}

SIGTYPE alarmfn()
{
longjmp(tmbuf, 1);
}
#endif // VERSION7 

#if	GUI && defined(WINDOWS3)

int	timed_read(int fd, int wait)
	{
	int c = -2;	// mark it as time out 

	if
		(
		1
#if	AP_BUILD || APL_IME
#else //AP_BUILD || APL_IME
	&&	0 == RuntimeIsConsole()
#endif//AP_BUILD || APL_IME
	&&	is_user_io(fd)
		)
		{
		c = W3_timed_read(fd, wait);
		if(EOF == c) return -1;
		}
	else
		{
#if USES_IME && !AP_BUILD && defined(WINDOWS3)
		c = WindowsFileRead(fd);
#else
			{
			UCH t;
			if(0 == read( fd, &t, 1)) 
				return -1;
			c = t;
			}
#endif
		}
	return c;
	}
#endif//GUI && defined(WINDOWS3)

#if !defined(WATCOM)
static int uninterrupted_read(int fd, UCH *buf, unsigned wanted)
{
int count;

	errno = 0;
	count = read(fd, buf, wanted);
	return count<0 ? 0 : count;
}
#endif
