
//	Copyright (c) 1982 Dyadic Systems Limited	

#include	<apl.h>
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<errno.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>

#include	<w3fsio.h>
#include	<gfns.h>


#if !unix
#include <unix.h>
#endif

#if defined (MSC)
#pragma  warning(disable : 4311)

// we cover read() and write() with win_read and win_write.
// win_read and win_write take an HWND as the first parameter (the "file").
// Problem is that on WIN64 an HWND is 64bits and won't fit in the int parameter 
// to read or write, and warning 4311 is raised on the truncation.
// Current (25/08/2005) thinking is that this is ok, as on a 64 bit machine
// HWNDs will be restricted to the bottom 32 bits to allow 32-64 interprocess
// communication, although this is just a theory.
// JD 25/08/2005

#if defined(_DEBUG)
#include  <crtdbg.h>
#endif

#include  <malloc.h>
#include  <process.h>

#endif // defined (MSC)

#if	defined(UNIX)
#	include	<unistd.h>
#endif//defined(UNIX)

#if	defined(DOSWIN)
#	include	<windows.h>
#	include	<winmain.h>
#endif//defined(DOSWIN)

#include	<support.h>

#if	defined(MASSCOMP)
#	include	<sys/ioctl.h>	// to get FIONREAD on masscomp... 
#endif	// MASSCOMP 

#if	defined(AIXPS270)
#	define FIONREAD 0x4004667f// cant include ioctl.h - name clashes 
#endif	// AIXPS270 

#if	defined(BSD)
#	include	<sgtty.h>
#endif	// BSD 

#if	defined(DOS)
#	include	<stdio.h>
#	include	<language.cf>
#	include	<hardware.h>
#	include	<task.h>		// need these to shadow malloc() 
#endif	// DOS 

#if	defined(DOS)
#	define		PIPESIZE	0x7fffffff
#elif	defined(DOSWIN)
#	define		PIPESIZE	(64*1024)
#else
#	define		PIPESIZE	4096
#endif//  

#define AP_UNICODE_MARKER	~(MEMORYUNS)0


unsigned long	XsessionWindowID;

sigjmp_buf		errbuff;
#if	UNIX
	unsigned parentid;
#elif	DOSWIN
	void * parentid;
	extern void task_error(HANDLE, int);
#endif//UNIX DOSWIN

void		setSigHandShake(void);

static void	(*errorhandler)(int);// allows user-defined error action 
static void	sigerror(int);		// default error handler	    
static void	gobble(unsigned pipe);	// consume any bytes queued on pipe	
typedef ssize_t (LIBCALL *IOFN)(int, void *, size_t); // read or write
void pipewds(IOFN iofn,unsigned pipe,MEMORYUNS *buff,size_t size);
void error(int errnumber);
void toapl(MEMORYUNS n);

unsigned normal_exit = 0;

//    s t a r t u p    
//	startup protocol as in "apl.h"					

#if	DOSWIN

static void LIBCALL chk_normal_exit(int ignored)
{
	switch(normal_exit)
		{
	case 0:
		normal_exit = 1;		// make sure we dont recur
		task_error(parentid, AP_EXIT);
		break;
		}
}

static void LIBCALL call_chk_normal_exit(void)
	{
	chk_normal_exit(SIGHUP);
	}

#endif//DOSWIN

#if !UNIX
	extern HANDLE task_begin(void);
	extern int task_init(void *);
	extern WINMAIN win_main;
	extern void terminate(void);
	extern void LIBCALL do_terminate(int);
	extern void init_version(void);
#endif

void startup(unsigned n_fns)
{
#if	UNIX

	parentid = fromapl();
	toapl(getpid());
	setSigHandShake();
	signal(SIGINT, SIG_IGN);
	signal(SIGQUIT, SIG_IGN);
	signal(SIGHUP, SIG_DFL);

#elif	DOSWIN

	init_version();		// set winNT so we don't get PROC_SETJMP

	(void)task_init(win_main.hInst);
	parentid = task_begin();
	if(0 == parentid)
		{
		printf("An AP must have a parent\n");
		exit(0);
		}
	signal(SIGHUP, do_terminate);		// clean exit
	signal(SIGABRT, chk_normal_exit);	// unclean exit
	atexit(call_chk_normal_exit);		// maybe unclean exit

#endif//UNIX DOSWIN

#if HAS_UNICODE
	toapl(AP_UNICODE_MARKER);

	#if 32==xxBIT				// force 64 bits of AP_UNICODE_MARKER
		toapl(AP_UNICODE_MARKER);
	#endif
#endif
	toapl(n_fns);
}


//	d e f i n e	
void define
	(
	gchar *apfn	// function name to appear in workspace	
,	int syntax	// 'syntax'  as defined in "apl.h".	
,	int fncode
	)
{
	unsigned length;
	size_t ignore;

	toapl(length = (unsigned)gstrlen(apfn));

	ignore = write(RSLTPIPE, apfn, length*sizeof(gchar));
	toapl(syntax);
	toapl(fncode);

	if (syntax & XWID)
		XsessionWindowID = (unsigned long)fromapl();
}


//	g e t a r r a y   -   p u t a r r a y	

ARRAY *getarray(void)		// get(possibly nested) array from apl	
{
	BOUND rank, *shape;
	ARRAY *array, **ptr;
	unsigned nwords;	// number of words remaining to read	
	ARRAY temp;

	if((nwords = fromapl()) == 0)	// read word count		
		return(0);		// left arg omitted		
	fields(&temp) = fromapl();	// read type/rank/eltype	

#if	DALIGN
	if(temp.eltype == APLDOUB)
	{
		unsigned pad, *tmp;

		pad = (temp.rank+2)%DOUBWORDS;
		tmp = (unsigned *)alloc(--nwords+pad);
		array = (ARRAY *)(tmp+(DOUBWORDS-pad)%DOUBWORDS);
	}
	else
#endif//DALIGN 
	array = alloc(--nwords);	// adjust word count &		
					// allocate space for array	
	shape = &(array->shape[0]);	// take pointer to shape	
	array->length = nwords;		// fill in array length		
	fields(array) = fields(&temp);	// and type/rank/eltype		

	nwords -= 2;			// decr words to be read	

	if(temp.type == SIMPLE)
		pipewds((IOFN)read, ARGSPIPE, shape, nwords);	// simple array 
	else
	{
		rank = temp.rank;
		pipewds((IOFN)read, ARGSPIPE, shape, rank);	// read shape	
		nwords -= rank;			// ... and decr count	

		/*	now recursively read in sub-arrays and
			install pointers in parent array.
		*/
		for(ptr = data(array, ARRAY *); nwords--;)
			*ptr++ = getarray();
	}
	return array;			// return pointer to array	
}


void putarray		// pass (possibly nested) array to apl
	(
	ARRAY *array	// pointer to array to be transferred	
	)
{
	register ARRAY **ptr;
	register unsigned rank, nwords;

	if(array == 0)
		toapl(0);			// null result		
	else
	{
		toapl(nwords = array->length+1);	// pipe length word 
		if(array->type == SIMPLE)
			pipewds((IOFN)write, RSLTPIPE, &fields(array), nwords-2);
		else
		{
			rank = array->rank;
			// pipe and shape 
			pipewds((IOFN)write, RSLTPIPE, &fields(array), rank+1);
			nwords -= rank+1+2;	// words remaining	
			// recursively put each subarray		
			for(ptr = data(array, ARRAY *); nwords--;)
				putarray(*ptr++);
		}
	}
}


//  t h r o w a r r a y   -   f r e e a r r a y  

ARRAY *throwarray(ARRAY *thing)			// send the array to	
{						// APL & free the space	
	putarray(thing);
	return(freearray(thing));
}


ARRAY *freearray(ARRAY *thing)			// return an ARRAY to	
{						// the free space pool	
	unsigned n, n_elts = 1;
	ARRAY **ptr;
	UCH *malloc_pntr = (UCH *)thing;

	if(thing == 0)				// don't try to free	
		return(0);			// if arg is null ptr	

#if     DALIGN                                  // undo fudged alignment
	malloc_pntr = ((UCH *)((IPTR)malloc_pntr & -sizeof(double)));
#endif//DALIGN

	switch(thing->type)
	{
	default:
		// hopefully already freed
		break;
	case NESTED: 				// recursively free	
						// subarrays first	
		for(n = thing->rank; n--;)
			n_elts *= thing->shape[n];

		ptr = &data(thing, ARRAY *)[n_elts];
		while(n_elts--)			// free subarrays in	
			freearray(*--ptr);	// reverse order	
	case SIMPLE:
		thing->type = 0; // we hope that free() does not mess with this
				 // enough to give NESTED or SIMPLE
		free(malloc_pntr);	// free this array	
	}
	return 0;
}

//	u t i l i t i e s	


/*	handler

	install error handler - routine called by error().
	default action is to signal the error to APL and siglongjmp(errbuff).

	can be overridden by user routine allowing call to error() to be used
	by asynchronous tasks.

	if errorfn returns, default action taken.
*/

void handler(void (*errorfn)(int))
{
	errorhandler = errorfn;
}

/*	error

	call error handler with error number.

	if errorfn returns, default action taken.
*/

void error(int errnumber)
{

	if(errorhandler)
		(*errorhandler)(errnumber);	// if this returns ...	

#if	defined(DOS)
	dos_comms_error(errnumber, 1);	// set the error return 
	siglongjmp(errbuff,errnumber);	// go back for next object 
#elif	defined(DOSWIN)
	switch(errnumber)
		{
	default:
		{
		task_error(parentid, errnumber);// make a noise about it
		siglongjmp(errbuff,errnumber);	// go back for next object
		}
	case -1:				// -1 -> termination
		raise(SIGHUP);			// request task end
	case 0:					// ensure new function
		siglongjmp(errbuff,errnumber);	// go back for next object
						// quietly
		}
#else
	sigerror(errnumber);		// ... take default action	
#endif	// DOS , DOSWIN 
}

#if	!defined(WATCOM)
static int errno2dyalog(int err) { return err; }
#endif

void filerror(void)
{
	error(100 + errno2dyalog(errno));
}



#if	!(defined(DOS) || defined(DOSWIN))
/*	signal error to APL - default error handler
*/

static sigjmp_buf envSigHandShake;		// buf for handshake 
static int	inerror;	// flag for handshake control 

static void sigerror(errnumber)
int errnumber;
{
	SIGTYPE (*savequit)(int);	// save current strong interrupt 
	SIGTYPE (*saveint)(int);	// save current weak interrupt	 

	savequit=signal(SIGQUIT,SIG_IGN); // dare not interrupt handshake 
	saveint=signal(SIGINT,SIG_IGN);
	
	inerror=1;		// so takeSigHandShake() knows who instigated 
	if(!sigsetjmp(envSigHandShake, 1))
		{
		 kill(parentid,SIGHANDSHAKE);	// stop apl		
		 pause();	// wait for apl to clean the result pipe
		}
	
	if(!sigsetjmp(envSigHandShake, 1))
		{
		 toapl(errnumber);	// write one word error code 
		 pause();		// wait for signal from apl  
		}
	gobble(ARGSPIPE);		// clean input pipe	
	kill(parentid,SIGHANDSHAKE);	// tell apl you have finished 

	signal(SIGQUIT,savequit);	// restore interrupts	
	signal(SIGINT,saveint);

	inerror=0;
	siglongjmp(errbuff,errnumber);	// go back for next object 
}


static void takeSigHandShake(int sig)
{
	setSigHandShake();
	if(inerror)siglongjmp(envSigHandShake, 1);	// return to error 
	gobble(ARGSPIPE);			// apl issued an error 
	kill(parentid,SIGHANDSHAKE);		// tell apl you have finished 
	siglongjmp(errbuff,1);			// go get next input	
}

void setSigHandShake(void)
{
	signal(SIGHANDSHAKE, takeSigHandShake);
}


static void gobble(unsigned pipe)	// consume any bytes queued on pipe	
{
	struct stat buf;
	UCH c;
	unsigned count;
	size_t ignore;

	while(1)
	{
#ifdef	FIONREAD
		if(ioctl(pipe,FIONREAD,&count) == -1)
			return;			// illegal ioctl 
#else
	 	if(fstat(pipe, &buf) == -1)
			return;			// can't stat pipe	
	 	count = buf.st_size;
#endif	// FIONREAD 
		if(count == 0)
			return;			// finished		
		while(count--)
			ignore = read(pipe, &c, 1);	// throw away bytes	
	}
}


#endif	// DOS , DOSWIN 


void pipewds(IOFN iofn,unsigned pipe,MEMORYUNS *buff,size_t size) // pipe words to or from ap 
{
	size_t chunk, pipesize;
	UCH *cbuf;
	int nbytes;

	size *= sizeof(MEMORYINT);			// convert to bytes	
	cbuf = (UCH *)buff;
	pipesize = PIPESIZE - sizeof(MEMORYINT);

	for
		(
	;	chunk = size < pipesize ? size : pipesize
	;
		) switch(nbytes = (*iofn)(pipe, cbuf, chunk))
		{
	default:
		// transfer a chunk at a time 

		size -= nbytes;
		cbuf += nbytes;
		pipesize = PIPESIZE;
		break;
	case 0:
#if	defined(DOS)	// if apl is finished on DOS then I never get control
#else //defined(DOS)
		kill(getpid(), SIGHUP);
#endif//defined(DOS)
		exit(0);		// EOF => apl finished 
	case -1:
		switch(errno)
			{
		default:
			exit(errno);	// give up
#if	defined(EINTR)
		case EINTR:
#endif//defined(EINTR)
#if	defined(EAGAIN)
		case EAGAIN:
#endif//defined(EAGAIN)
#if	defined(ERESTART)
		case ERESTART:
#endif//defined(ERESTART)
			continue;	// io interrupted try again 
			}
		}
}


void toapl(MEMORYUNS n)
{
	pipewds((IOFN)write, RSLTPIPE, &n, 1);
}


MEMORYUNS fromapl()
{
	MEMORYUNS rslt;
	pipewds((IOFN)read, ARGSPIPE, &rslt, 1);
	return(rslt);
}

#ifdef	DOS

pragma	Calling_convention(PASCAL);
extern void *	_mwmalloc(unsigned size);
extern void	_mwfree(void *ptr);
pragma	Calling_convention();

UCH *malloc(unsigned size)
{
return (UCH *)_mwmalloc(size);
}

void free(void * ptr)
{
_mwfree(ptr);
}

#endif	// DOS 

void domain(void)	// return a domain error to apl	
{
	error(EDOMAIN);
}

void nonce(void)		// return a nonce error to apl	
{
	error(NONCE);
}

void limit(void)		// return a limit error to apl	
{
	error(LIMIT);
}

#if 0 

#define CPYNAME(a,b)	cpy_##a##_##b

#define CPYFN(a,b)					\
static void CPYNAME(a,b)(a *to, b*fm, BOUND n)		\
	{						\
	while (n--)					\
		{					\
		a t = (a)*fm;				\
		if (t!=*fm)				\
			error(DOMAIN);			\
		*to++=*fm++;				\
		}					\
	}						

#define CPYSAME(a,b)	same_##a##_##b

#define SAMEFN(a,b)					\
static void CPYSAME(a,b)(a *to, b*fm, BOUND n)		\
	{						\
	memcpy(to,fm,n*sizeof(a);			\
	}					

static void (*cpyfns)(void *,void *,BOUND)[4][4]=
		{
		CPYSAME(byte,byte)  ,CPYNAME(byte,short)  ,CPYNAME(byte,long)  ,CPYNAME(byte,double),
		CPYNAME(short,byte) ,CPYSAME(short,short) ,CPYNAME(short,long) ,CPYNAME(short,double),
		CPYNAME(long,byte)  ,CPYNAME(long,short)  ,CPYSAME(long,long)  ,CPYNAME(long,double),
		CPYNAME(double,byte),CPYNAME(double,short),CPYNAME(double,long),CPYSAME(double,double),
		};

void et_memcpy_et(void *to, int e1, void *fm, int e2, BOUND n)
{
cpyfns[ELTYPE_SHFTI(e1)][ELTYPE_SHFTI(e2)](to,fm,n);
}

#endif

#ifdef DOS
void exit(int code)	// cover the exit function otherwise APL terminates 
{
// APL does not run APs as programs with their own PSPs 
UCH emess[] = "dont attempt to reuse an AP after an AP_EXIT\n";
for(;;)
	{
	error(AP_EXIT);
	write(2, emess, strlen(emess));
	}
}

void syserror(unsigned n)
{
fprintf(stderr, "syserror %d\n", n);
exit(n);
}
#endif // DOS 

void terminate(void)
{
	normal_exit = 1;
	exit(0);
}

void LIBCALL do_terminate(int ignored)
{
	terminate();
}


#if	XAPL
                                                                  
int xerror(UCH *msg)
	{
	write(2, msg, strlen(msg));
	write(2, "\r\n", 2);
	}

void setsignals(void)
	{
	/*      use the interrupt handlers already set up. */             
	}

void resetsignals()
	{
	}
                                                                  
void wait_select_input(void)
	{
	/* null function as not using select() to listen on pipe
	   for events signalled after interrupts */
	}

#else //XAPL

void syserror(unsigned n)// hack for Xprefect: it supplies its own and is the only XAPL
{
	printf("via: syserror %u errno %u\n\r",n,errno);
	fflush(stdout);
	exit(3);
}

#endif//XAPL

//	Copyright (c) 1982 Dyadic Systems Limited	
