
//	Copyright (c) 1982 Dyadic Systems Limited	

/*
	Translate tables and support functions
*/
#include <machine.h>

#if WINCE
#	include	<aplole.h>
#	include	<winsock.h>
#	include	<io.h>
#	include	<fcntl.h>
#elif	UNIX
#	include	<unistd.h>
#	include <sys/time.h>
#	include	<arpa/inet.h>
#elif	WINDOWS3
#	include	<windows.h>
#	include	<io.h>
#	include	<fcntl.h>
#endif//UNIX

#include <stdlib.h>
#if	ALLOCA_H
#	include <alloca_h.h>
#endif//ALLOCA_H
#include <ctype.h>
#include <wctype.h>	// iswalpha(), towlower(), ...
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#if	MSC
#	define S_ISREG(m)	(S_IFREG & (m))
#endif//MSC
#include <setjmp.h>
#include <chars.h>
#include <error.h>
#if	MAINWIN
#	include <mw.h>
#endif//MAINWIN

#include <outcodes.h>
#include <datecodes.h>
#include <trans.h>

#define ILINECODES_DEFINE
#include <keystrokes.h>

#include <t_tables.h>		// Standard translate tables		

#include <jumps.h>

#include <unicode.h>
#include <win_io.h>
				// extra tables are defined below	
#if	defined(BELL) || defined(BSD)
#include <fcntl.h>
#define	FCNTL
#endif//defined(BELL) || defined(BSD)

#if	defined(O_BINARY)
#else //defined(O_BINARY)
#	define	O_BINARY	0	// will make no difference
#endif//defined(O_BINARY)

#ifndef	INPIPE			// allow machines to override (APOLLO)	
#define	INPIPE	3
#define	OUTPIPE	4
#endif	// INPIPE 

#if	_NFILE
#else //_NFILE
#	define	_NFILE	20	// should have got this from stdio.h
#endif//_NFILE

#if	32 < _NFILE		// IBM define it as 32767 which is silly
#	undef	_NFILE		// for what we are trying to do
#	define	_NFILE	32	// make it more sensible but still larger than
#endif//32 < _NFILE		// the default of 20. Constrains the nfiles AP

#define	MAX_BUF		(128*1024)
#define T_WKBFSZ	(32*1024)

#define BIT_READ(vec, i)	(vec[(i) >> 3] & (0x80 >> ((i) & 7)))
#define BIT_FLIP(vec, i)	(vec[(i) >> 3] ^= (0x80 >> ((i) & 7)))

#ifdef	WINCE
#	include	<wcestd.h>
#endif//WINCE

#include <getch.h>
#include <unix.h>
#include <writloss.h>

#if	AP_BUILD
	extern gchar* W3getenv(const gchar *envname);
	extern void W3Error(gchar *buf,...);
	extern int W3_timed_read(int fd,int wait);
	extern void W3_timed_ansi(int fd,int wait,int *value,int *ansi);
	extern void trans_message_box(gchar *message);
	extern size_t getcmdprofile(const gchar *name,gchar * buf,size_t buflen);
	extern size_t getglobalenv( const gchar *name,gchar * buf,size_t buflen);
#else //AP_BUILD
#	include <w3fsio.h>
#	include <embed.h>
#	include <colordefs.h>
#endif//AP_BUILD

static int m_inkey(int fd, Mmap *mp);
static int m_ansi(int fd, Mmap *mp);
static void eno(unsigned n);
static int out_stream(ubyte *cp, unsigned count);
static int out_raw(ubyte *cp, unsigned count);
static void up_cols(Imap *t, Mmap *mode, unsigned count);
static void up_rows(Imap *t, Mmap *mode, unsigned offset);
static void copy_up(UCH *to, UCH *from, unsigned n);
static void io_error(gchar *mess);
static void ReadScanMap(void);
static unsigned av_out(struct Translate *t,unsigned row,UCH **bufpt);
static int lc_str2num(UCH *cp,UCH ** cprslt);
static unsigned value2line(struct Translate *t, achar value);

void special(int code);
void o_init(void);
void o_filbuf(void);
void i_filbuf(void);
#if APL_IME
static int WIN_out(ubyte *cp, unsigned count);
#elif WINDOWS3
extern int WIN_out(ubyte *cp, unsigned count);	//get it from w3fsio.c
#endif //apl_ime
static UCH *fmtnum(int num, UCH* cptr);
static int checkrealloc(void);

void ansi2av(UCH *cptr, size_t size);
void av2ansi(UCH *cptr, size_t size);

int  UNICODEtoKEY(int key);

ubyte      FOREGROUND[256][3],      BACKGROUND[256][3];
ubyte  SESHFOREGROUND[256][3],  SESHBACKGROUND[256][3];
ubyte PRINTFOREGROUND[256][3], PRINTBACKGROUND[256][3];


UCH WINCHARS[256];		// cached screen output translate
				// also supports avtounix()

UCH BODGECHARS[256];		// support backwards translation of output
				// translate file. Has all sorts of problems
				// but needed until we write our own edit
				// box widget.
static unsigned cached;


unsigned unixstr2av(gchar * dest, gchar * src, unsigned maxchars)
	{
	gchar *d = dest;
	gchar *s = src;
	unsigned count = 0;

	for
		(
		--d
	,	--s
	,	++maxchars
	;	--maxchars && *++s
	;	*++d = unixtoav(*s)
	,	++count
		);
	return count;
	}

#if APL_IME

static int t_fopen(gchar *name,gchar * error2);

void error(unsigned n)
{
}

void W3Error(gchar *buf,...)
{
}

void binary_stream(FILE *stream)
{
}

// fix bug 4707 []SM scrolling by restoring use of WIN_out@w3fsio.c for
// classic Dyalog when we call Avtoasc.  
// Somehow, in v12.0, WIN_out crept out of the #if APL_IME case.  I Vince have
// put it back here, June 2009
static int WIN_out(ubyte *cp, unsigned count)
{
return count;
}


void trans_message_box(gchar * message)
{
W3Error(message);
}

int is_user_io(int fd)
{
return 1;
}

int LoadKeyboard(gchar *file)
{
if (-1!=t_fopen(file,""))
	{
	i_filbuf();
	return 1;
	}
return 0;
}

static unsigned *Key;
static unsigned KeyPos;

int ime_key(int fd , int wait)
{
return Key[KeyPos++];
}

int GetKey(ubyte vk,ubyte shift,int allow_mode)
{
unsigned key[3], *k=key;
int gui_Inkey
	(
	int fd	// file descriptor 
,	int up	// USER or PROG 
,	unsigned (*mc_fn)
		(
		UCH *
	,	Imap *
	,	unsigned
	,	unsigned
	,	unsigned
		)
,	int ansi
,	int allow_mode
	);


if (shift) 	*k++=shift;
		*k++=vk;
		*k++=MTQUEUE;
KeyPos=0;
Key=key;

return gui_Inkey(0,PROG,mc_dflt,0,allow_mode);
}

int IsKey(ubyte vk,ubyte shift)
{
int key=GetKey(vk,shift,0);

return key;//!=MTQUEUE;
//return (0<key&&key<=255) || (key == MC);
}

#endif // APL_IME

#if 0==HAS_UNICODE

/*	asctoav()
 *
 * Maps the input from an ascii device.  Note that on a UTS system,
 * it is the calling codes responsibility to ensure aplmode == 0
 * if the terminal is a tube (3278-type device).
 *
 */

/*	N.B.
 *	this code supports oddball translates now that the input
 *	mapping supports keyboards. This code is still used for
 *	things like converting ap function names when they are
 * 	defined into the workspace by quadSH.
 */

gchar asctoav(gchar a,unsigned int aplmode)
{
	return aplmode ? aplmap[a] : unixtoav(a);
}


/*	unixtoav()
 *
 * Converts input known to be ASCII to APL internal form (QuadAV)
 *
 */

UCH unixtoav(UCH a)
{
	return cached ? BODGECHARS[a] : asciimap[a];
}

/*	avtounix()
 *
 * Converts an APL character to an ASCII equivalent graphic if
 * possible.
 *
 */

static unsigned hexvalue(UCH hc)
	{
	if('0' <= hc && '9' >= hc) return hc - '0';
	if('a' <= hc && 'f' >= hc) return 10 + hc - 'a';
	if('A' <= hc && 'F' >= hc) return 10 + hc - 'A';
	error(11);
	return 0;	// keep the compiler happy
	}

UCH avtounix(UCH c)
{
	if(cached)
		{
#if defined(WINCE)	//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ recusion in getenv
		return WINCHARS[c];
#else
		static unsigned first = 1;
		static UCH map[256];

		if(first)
			{
			UCH *mapchars = (void *)getenv("MAPCHARS");
			unsigned i;
			size_t len = mapchars ? strlen(mapchars) : 0;

			for( i = 256; --i, map[i] = i, i;); // init identity map
			len &= ~3;		// must must be a multiple of 4
			for(;len;)
				{
				unsigned mapval;
				unsigned location;

				mapval = hexvalue(mapchars[--len]);
				mapval += 16 * hexvalue(mapchars[--len]);
				location = hexvalue(mapchars[--len]);
				location += 16 * hexvalue(mapchars[--len]);
				if
					(
					256 > location
				&&	256 > mapval
					)
					{
					map[location] = mapval;
					}
				}
			first = 0;
			}
		return WINCHARS[map[c]];
#endif
		}
	else
		{
		return unixtab[c];
		}
}


void avtow3(UCH *av)
{
UCP op;
for (op = av ; *av ; *op++ = avtounix(*av++));
}

void w3toav(UCH *av)
{
UCP op;
for (op = av ; *av ; *op++ = unixtoav(*av++));
}

void AVtoW3(UCH *av,size_t len)
{
while (len--)
	{
	*av=avtounix(*av);
	av++;
	}
}

void W3toAV(UCH *av,size_t len)
{
for (;len--;*av=unixtoav(*av),av++);
}

#endif // 0==HAS_UNICODE

void ToLower(achar *acp, size_t len)
{
// note: non of the languages currently defined in the upper unicode plains
// have the concept of "case". This concept is actually rare. It feels
// otherwise because the Latin based languages have it.
for (; len--; ++acp)
	{
	gchar gcp = *acp;

	if
		(
		sizeof(achar) == sizeof(gchar)	// help compiler at compile time
	||	gcp == *acp			// in range for to(w)lower()
		)
		*acp =
#if	HAS_UNICODE
			towlower
#else //HAS_UNICODE
			tolower
#endif//HAS_UNICODE
				(gcp);
	}
}                        

void ToUpper(achar *acp, size_t len)
{
for (; len--; ++acp)
	{
	gchar gcp = *acp;

	if
		(
		sizeof(achar) == sizeof(gchar)	// help compiler at compile time
	||	gcp == *acp			// in range for to(w)upper()
		)
		*acp =
#if	HAS_UNICODE
			towupper
#else //HAS_UNICODE
			toupper
#endif//HAS_UNICODE
				(gcp);
	}
}                       

achar in_word_specials[] =
	{
	AV_POINT, AV_QUAD, AV_UNDER, AV_DELTA, AV_DELTAU, AV_HIGH,
	AV_ON, AV_OVER, AV_POWER, AV_CONV,
	AV_AV86, AV_AV87, AV_AV88, AV_AV89, AV_AV8a, AV_AV8b, AV_AV8c, AV_AV8d, AV_AV8e, AV_AV8f,
	AV_AV90, AV_AV91, AV_AV92, AV_AV93, AV_AV94, AV_AV95, AV_AV96, AV_AV97, AV_AV98, AV_AV99,AV_AV9a,
	AV_AVD3, AV_AVD4, AV_AVD5, AV_AVD6,
	AV_AVE8, AV_AVE9, AV_AVEA, AV_AVEC,
	0
	}; //Added national language chars for prob 1865, cvc,aug99

#if HAS_UNICODE
#undef AV_AU
#undef AV_ZU
#define AV_AU 0x24b6
#define AV_ZU 0x24cf
#endif // default []avu as seen in uniav.h AV_AU, AV_ZU are not circle letters
//This fixes bug 4636

int char_in_word(achar c, WITHPOINT_t withpoint)
{	//Function added with fix to problem 1865, (RW not working)
return (
	(c>=AV_ZERO && c<=AV_NINE)	||
	(c>=AV_AL   && c<=AV_ZL)	||
	(c>=AV_AU   && c<=AV_ZU)	||
	(c>=AV_A    && c<=AV_Z)		||
	astrchr(in_word_specials + withpoint, c)
	);
}

int in_word(achar c)
{
return char_in_word(c, WITHPOINT);
}


int avtoupper(achar c)
{
if (c >= AV_AL && c <= AV_ZL)
	c += (AV_A-AV_AL);

return c;
}

void AVtoUpper(achar *cp,size_t len)
{
while (len--)
	*cp++ = avtoupper(*cp);
}

int avstrnicmp(achar *s1, achar *s2, size_t len)
{
while (len--)
	{
	if (avtoupper(*s1) < avtoupper(*s2))
		return -1;
	else if (avtoupper(*s1) > avtoupper(*s2))
		return 1;
	s1++;
	s2++;
	}
return 0;
}


int LIBCALL avstricmp(achar *s1, achar *s2)
{
while (*s1 || *s2)
	{
	if (avtoupper(*s1) < avtoupper(*s2))
		return -1;
	else if (avtoupper(*s1) > avtoupper(*s2))
		return 1;
	s1++;
	s2++;
	}
return 0;
}

int LIBCALL avstrcmp(achar *s1, achar *s2)
{
while (*s1 || *s2)
	{
	if (*s1 < *s2)
		return -1;
	else if (*s1 > *s2)
		return 1;
	s1++;
	s2++;
	}
return 0;
}


/*	avstrncmp
	Same as strncmp, only compares from (assumed to be quadAV string)
	with to (assumed to be UNIX string).
	Returns -1 if from is sorted before to, +1 if from is sorted
	after to, and 0 if from and to are equal.
*/
int avstrncmp(achar *from, achar *to, size_t max)
{
	while (max--)
	    {
	    UCH sch, dch;
	    sch = *from++;
	    dch = *to++;
	    if (sch < unixtoav(dch))
		return -1;
	    else if (sch > unixtoav(dch))
		return 1;
	    }
	return 0;		// strings are equal	
}




/*
 *	Support for input and output mapping initialisation 
 */

static int trans_file;			// fd for translate file	
static unsigned trans_size;		// size of translate file	
static UNIQUE_FID trans_unique_fid;	// file id of translate file
int test_file;				// done up to this descriptor	
int trans_no_error_messages_please = 0;	// nfiles wants to turn them off 
static int suppress_complaint = 0;	// "redirected" is not there (normally)

typedef void FNAMEFN
	(
	gchar *tp		// null or target file name 
,	gchar *buf		// where to put the file name 
,	unsigned errflag		// an error condition ? 
	);
static FNAMEFN d_fname, s_fname;
extern FNAMEFN i_fname, o_fname;
FNAMEFN *fname;

static UCH *curr_start_of_line;

#define MAXST 18	// keeps 4 byte allignment 

#define FORMFEED 12

#ifdef	INITBYTES_FUNCTION
static void initbytes(start, count, value)
UCP start;
unsigned count;
UCH value;
{
while(count--)*start++ = value;
}
#else//  
#define	initbytes(start, count, value)	memset(start, value, count)
#endif//  

gchar *fextend(achar *fname, gchar *fext)
	{
	static gchar nullstr[] = GSTR("");
#ifdef	FEXTEND
	achar * slash = astrrchr(fname, '/');
	achar * back  = astrrchr(fname, '\\');
	achar * dot   = astrrchr(fname, '.');
	achar * dir   = slash > back ? slash : back;

	if(dot <= dir) return fext;
#endif//FEXTEND 
	return nullstr;
	}

static gchar *gfextend(gchar *fname, gchar *fext)
	{
	static gchar nullstr[] = GSTR("");
#ifdef	FEXTEND
	gchar * slash = gstrrchr(fname, '/');
	gchar * back  = gstrrchr(fname, '\\');
	gchar * dot   = gstrrchr(fname, '.');
	gchar * dir   = slash > back ? slash : back;

	if(dot <= dir) return fext;
#endif//FEXTEND 
	return nullstr;
	}

achar * ftrunc			// truncates name
	(
	achar **this
,	achar **lim
,	gchar ext[4]
,	CHARSET charset
	)
	{
#if	defined(FEXTEND)
// normalise the slashes and truncate the extension 
	static struct
		{
		gchar lowera;
		gchar uppera;
		gchar slash;
		gchar slope;
		gchar point;
		} map[2] = 
		{
			{	// ASCII
			'a', 'A', '/', '\\', '.'
			}
		,	{	// QUADAV
			AV_AL, AV_A, AV_SLASH, AV_SLOPE, AV_POINT
			}
		};
	gchar lower_ext[4] = GSTR(".");
	gchar *gfrom, *gto;
	achar *afrom, *ato;
	int char_shift = map[charset].uppera - map[charset].lowera;
	gchar slash = map[charset].slash;
	gchar slope = map[charset].slope;
	static achar buf[MAXFCHARS];
	unsigned i;

	for
		(
		gfrom = ext + 4
	,	gto = lower_ext + 4
	;	gfrom > ext + 1
	;	*--gto = *--gfrom - char_shift
		);
	for
		(
		afrom = *this
	,	ato = buf
	;	afrom < *lim
	;	++afrom
	,	++ato
		)
		{
		if(slash == (*ato = *afrom)) *ato = slope;
		}
	if((i = 4) < *lim - *this) for
		(
		afrom -= 4
	;	i--
	&&	
			(
			afrom[i] == ext[i]
		||	afrom[i] == lower_ext[i]
			)
	;
		);
		
	if(0 == ++i) ato -= 4;

	*ato = 0;	// always null terminate
	*this = buf;
	*lim = ato;
#endif//defined(FEXTEND)
	return *this;
	}


#if	UNIX
#	define PUNCTUATION	GSTR("/")
#else //UNIX 
#	define PUNCTUATION	GSTR("/\\:")
#endif//UNIX 

static UNIQUE_FID null_unique_fid = { 0 }; 
static gchar pathdelim[] = { PATHDELIM, 0 };
static gchar dirdelim[] = { DIRDELIM, 0 };

static void t_fname(	// develop a filename 
gchar *tp,		// null or target file name 
gchar *envdir,		// "APLTRANS" or "APLKEYS" 
gchar *dirname,		// "apltrans" or "aplkeys" 
gchar *envname,		// "APLT"     or "APLK"	   
gchar *ext,		// ".DOT"     or ".DIN"	   
gchar *buf,		// where to put the file name 
unsigned errflag 	// called by error handler 
)
{
gchar *ep, *dp;
static unsigned skip;
static UNIQUE_FID unique_fid;
gchar tbuf[127];	// somewhere to copy enviroment vars to 

if(!buf)
	{
	// initialise static data 
	skip = 0;		// skip no levels of APLKEYS 
	unique_fid = null_unique_fid;	// set last device read
	return;			// to impossible values 
	}
if(!tp)
	{
	ep = envname + 4;
	if(test_file>9)*ep++ = '0' + test_file/10;
	*ep++ = '0' + test_file%10;
	*ep=0;

	ep = ggetenv(envname);
	if(0 == ep && is_user_io(test_file))
		{
		*(envname+4) = 0;	// "APLK" 
		ep = ggetenv(envname);
		if(0 == ep) ep = ggetenv(GSTR("TERM"));
		if(0 == ep) ep = GSTR("default");
		}
	if(0 == ep) switch(envname[3])
		{
	case 'K': case 'T':
#if	AP_BUILD || APL_IME || 0 != _CONSOLE
		ep = GSTR("file");
#else //AP_BUILD || APL_IME || 0 != _CONSOLE
		ep = GSTR("redirected");
		suppress_complaint = 1;
#endif//AP_BUILD || APL_IME || 0 != _CONSOLE
		break;
	case 'F':
		ep = GSTR("default");
		break;
	case 'S':
		ep = GSTR("us.scn");	// as in default.reg
		break;
	default:
		ep = GSTR("");		// defensive
		break;
		}
	tp = ep;
	}
gstrcpy(tbuf, tp);
tp = tbuf;
gstrcat(tbuf, gfextend(tbuf, ext));
if(gstrcspn(tp, PUNCTUATION) == gstrlen(tp))	// relative pathname ? 
	{
	int s = 0 != (dp = ggetenv(envdir));
	switch(s)
		{
		default:
			{
			GSTAT sbuf;
			unsigned count;

			sbuf.st_dev = 0; // in case we dont succesfully stat 
			sbuf.st_ino = 0; // a file 
			buf[0] = 0;	 // in case APLKEYS exists but is 0
			for(count = 0 ;;skip += count++ == skip, dp += *dp != 0)
				{
				gchar *bp = dp;
				unsigned n;

				dp += gstrcspn(dp, pathdelim);
				if(bp == dp)break;
				if(count == skip)
					{
					gstrncpy(buf, bp, n = dp - bp);
					buf[n] = 0;
					gstrcat(buf,dirdelim);
					gstrcat(buf,tp);
					if
						(
/* got a file */				-1 != gstat(buf, &sbuf)
/* is a plain file */			&&	(S_ISREG(sbuf.st_mode))
					&&		(
/* is an error or */					errflag
/* different from last one read */		||	0 == match_unique_fid
								(
								filename_to_unique_fid(buf)
							,	unique_fid
								)
							)
						)break;
					}
				}
			unique_fid = filename_to_unique_fid(buf);
			}
			break;
		case 0:		// build the default 
			if(dp = ggetenv(GSTR("DYALOG")))
				{
				gstrcpy(buf, dp);
				gstrcat(buf, dirdelim);
				gstrcat(buf, dirname);
				gstrcat(buf,dirdelim);
				gstrcat(buf,tp);
				}
			else
				buf[0] = 0;	// return empty
			break;
		}
	}
	else gstrcpy(buf, tp);
}

static int t_fopen(gchar *name,gchar * error2)
						// file name 
						// sub error mesage 
{
struct stat statbuf;
int ret;

if
	(
	(-1 == (trans_file = gopen(name, O_RDONLY | O_BINARY))) ||// cant open 
	(-1 == fstat(trans_file, &statbuf))	// no file 
	) 
	{
	if(trans_no_error_messages_please == 0 && suppress_complaint == 0)
		{
		static gchar error0[] = GSTR("Stream ");
		static gchar error1[] = GSTR(": Can't access ");
		static gchar last_error_name[MAXFCHARS] =GSTR("");

		if(0 == gstrcmp(name, last_error_name)) return -1;
		gstrncpy(last_error_name, name, MAXFCHARS);
		last_error_name[MAXFCHARS - 1] = 0;
#if	GUI && defined(WINDOWS3)
		{
		W3Error
			(
			GSTR("%ls %i %ls %ls %ls")
		,	error0
		,	test_file
		,	error1
		,	name
		,	error2
			);
		}
#else //GUI && defined(WINDOWS3)
		{
		size_t ignored;

		ignored = write(2, error0, gstrsize(error0));
		eno(test_file);
		ignored = write(2, error1, gstrsize(error1));
		ignored = write(2, name  , gstrsize(name));
		ignored = write(2, error2, gstrsize(error2));
		}
#endif//GUI && defined(WINDOWS3)
		}
	ret = -1;
	}
else
	{
	trans_size = statbuf.st_size;
	trans_unique_fid = fd_to_unique_fid(trans_file);

#if	DOSWIN && !WINCE && !APL_IME
		binary_file(trans_file);
#endif//DOSWIN && !WINCE && !APL_IME
	ret = trans_file;
	}

suppress_complaint = 0;
return ret;
}

void	(*x_error)(int);		// helps quadSR & nfiles avoid exiting 

	//	values passed to (*x_error)()	

typedef enum
	{
	NULLERROR
,	NUM_WOT
,	NUM2BIG
,	MINUS_NO_M
,	UNKNOWN_LC
,	FILE2BIG
,	CANT_OPEN
,	CANT_READ
,	NOT_ASCII
,	SYNTAX
,	STRING2BIG
,	EMBEDDED_NULL
,	DUP_STRING
,	NO_COLON
,	PROG_MODE
,	USER_MODE
,	UTF_CODE
,	STRING2MANY
,	TC_CAPS
,	BOTH_LOWER
,	NOSCANFILE
,	MALLOC_FAIL
,	MOD2LONG
,	TOOMANYVAL

#ifdef	INTERRUPTS
,	WEAK2BIG
,	STRONG2BIG
,	WEAK_MODES
,	STRONG_MODES
#endif	// INTERRUPTS 
	} INIT_ERR;

static void	init_err(int code)
{
	static gchar *msgs[] =
			{
			GSTR("")
		,	GSTR("digit expected")
		,	GSTR("number exceeds 255")
		,	GSTR("non-mode line starts with '-'")
		,	GSTR("unrecognised line code")
		,	GSTR("file too big")
		,	GSTR("can't open")
		,	GSTR("can't read")
		,	GSTR("file not ASCII / not terminated by newline")
		,	GSTR("invalid syntax")
		,	GSTR("string exceeds 30 chars!")
		,	GSTR("file contains embedded nulls")
		,	GSTR("duplicate string")
		,	GSTR("colon expected")
		,	GSTR("invalid PROG mode")
		,	GSTR("invalid USER mode")
		,	GSTR("invalid utf code")
		,	GSTR("too many strings")
		,	GSTR("caps exchange value must exceed 4")
		,	GSTR("already specified as upper case")
		,	GSTR("can't read scan file")
		,	GSTR("can't allocate memory")
		,	GSTR("MOD: filename too long")
		,	GSTR("too many characters in unicode table")
#ifdef	INTERRUPTS
		,	GSTR("weak interrupt must be one character")
		,	GSTR("strong interrupt must be one character")
		,	GSTR("weak interrupt differs across modes")
		,	GSTR("strong interrupt differs across modes")
#endif	// INTERRUPTS 
		,	GSTR("unknown error")
			};
	unsigned maxcode = (sizeof(msgs) / sizeof(msgs[0])) -1;

	if(code > maxcode) code = maxcode;
	io_error(msgs[code]);
}

static void	quadsr_lc_err(int code)
{
	error(61);
}

/*
	Support for input mapping
	i_init()	get the keyboard map table
	inkey()		get an input keystroke
	inlab()		returns pointer to apl string
	keyset()	returns current mode and sets new mode
	v_inkey()	kludge for via - do inkey() but dont allow mode change
*/


extern void ReadKeyboardShortcuts(void);

struct Fd_map in_fd[_NFILE];		// the input maps		

	// forward references - this file
static Imap redirected_input_map;
static Mmap redirected_input_mode;
static UCH *RawAnso;

static int LIBCALL sort_line_code(const void *vp1, const void *vp2)
{
const s2n_map *s1 = vp1;
const s2n_map *s2 = vp2;

return strcmp(s1->string,s2->string);
}

static void sort_line_codes(void)
{
static int done;

if (!done)
	{
	qsort(i_line_codes,NUM_LINE_CODES,sizeof(s2n_map),sort_line_code);
	done=1;
	}
}


/*	key logging:

	logkey: whether this task logs keys: 0 in ap's, apl's main() sets
		 flag if DEBUG_K.

*/

int		logkey = 0;		// default don't log		
static int	keylog = -1;		// opened in (recursive) i_init	
void d_error(int);

static void  LIBCALL i_finish(void)
	{
	unsigned i, j;
	Imap *map;

	for(i = _NFILE; i--;)if (map = in_fd[i].map)
		{
		for(j = i; j--;) if(map == in_fd[j].map) in_fd[j].map = 0;
		if (map->freefn)
			map->freefn(map);
		}
	}

static void i_init(void)
{
sort_line_codes();

#if 0==USES_IME		// no translate tables when using "external" keyboards

#if !WINCE
	if
		(
		logkey				// want to log		
	&&	keylog == -1				// first time through	
		)
		{
		keylog = creat( "apllog", 0666);

		if(-1 == keylog)		// couldn't open	
			logkey = 0;		// can't log		
		}
#else
	if
		(
		logkey				// want to log		
	&&	keylog == 0				// first time through	
		)
		{
		keylog = gopen( GSTR("apllog"), O_WRONLY | O_CREAT);

		if(-1 == keylog)		// couldn't open	
			logkey = 0;		// can't log		
		}
		
#endif   //WINCE


if(!x_error || x_error == d_error) x_error = init_err;
fname   = i_fname;

// while(_NFILE>++test_file) all current code uses fd = 0 
for(test_file = 0; 1 > test_file; ++test_file) switch(test_file)
	{
	case 1: case 2: // outlaw these for input 
#if	!defined(WATCOM)
	case INPIPE: case OUTPIPE:
#endif//!defined(WATCOM)
		break;
	default:
		if(i_get(0))i_filbuf();
		break;
	}

if(in_fd[0].map)	// we have a keyboard 
#ifdef INTERRUPTS
// set up the weak and strong interrupts so that inittty() can install them 
	{
	extern UCH tty_weak, tty_strong; // see tty.c 

	tty_weak = in_fd[0].map->weak_int;
	tty_strong = in_fd[0].map->strong_int;
	}
#else //INTERUPTS
	;
#endif//INTERRUPTS 
else
#if	GUI && defined(WINDOWS3)
	//	we cope with this using m_ansi()
#else //GUI && defined(WINDOWS3)
	//	no hope just exit and allow user to read the error message
	exit(1)
#endif//GUI && defined(WINDOWS3)
	;
#if	GUI && (defined(MAINWIN) || WINE)
	ReadScanMap();
#endif//GUI && (defined(MAINWIN) || WINE)

atexit(i_finish);


#endif // USES_IME
}

//	I am assuming that it is enough to check if reallocmoves for this one
//realloc.  Normally, realloc does not move when shrinking the size, but Simcorp
// paul gurtler turned on "page heap" option (2 june 2008)
//gflags.exe -i dyalog.exe +hpa
//	This debugging option affects all dyalog.exe on the system and has the
//effect of making a shrink realloc move.  this then caused an assertion failure
//in the d/o/i_filbuf functions.  
static int checkrealloc(void)
{
	static int reallocmoves = -1;
	UCH *buf, *new;
	if (-1 == reallocmoves)
	{
		buf = malloc(MAX_BUF);
		if(0 == buf) (*x_error)(MALLOC_FAIL);
		new = realloc(buf, (MAX_BUF/2));
		if (new != buf)
			reallocmoves = 1;
		else 
			reallocmoves = 0;
		free(new);
	}
	return reallocmoves;
}



void i_fname	// develop an input file name 
	(
	gchar *tp		// null or target file name 
,	gchar *buf		// where to put the file name 
,	unsigned errflag		// is this an error 
	)
	{
	t_fname(tp, GSTR("APLKEYS"), GSTR("aplkeys"), GSTR("APLK  "), GSTR(".DIN"), buf, errflag);
	}

static void	prestarterr(int err) // error before curr_start_of_line set 
{
	UCH	formfeed = FORMFEED;

	curr_start_of_line = 1 + &formfeed;

	(*x_error)(err);
}

int i_get(gchar *tp)
{
struct stat statbuf;	// for testing file attributes 
gchar envbuf[MAXFCHARS]; // for building the name of the map file 
int i;

#ifndef WINCE
// test: Is the descriptor open for read 

if(-1 == fstat(test_file, &statbuf)) return 0; // descriptor not open 

#ifdef FCNTL
	i = fcntl(test_file, F_GETFL, 0);
	if(-1 == i) return 0;		// unsuitable descriptor 

	i &= O_RDONLY | O_RDWR | O_WRONLY;
	if(i != O_RDONLY && i != O_RDWR) return 0; // not read 
#else//  
	if
		(
		0 == 
			(
			statbuf.st_mode
#if	defined(MSC)	// see dispareging comment in o_get()
		&	(_S_IFCHR | _S_IFIFO | _S_IREAD)
#else //defined(MSC)
		&	0444
#endif//defined(MSC)
			)
		) return 0; // not read 
			// note: cant test that it is actually read 
#endif // FCNTL 
#endif //WINCE

// is there an enviroment variable for this descriptor 

(*fname)(0, 0, 0); 		// init statics for top level request 
(*fname)(tp, envbuf, 0);		// get file name into envbuf 

if(-1 == t_fopen(envbuf, GSTR("\n")))
	{
#if	0 == _CONSOLE
#if	AP_BUILD
	(*x_error)(CANT_OPEN);		// don't do this sub in AP's
#endif//AP_BUILD
	in_fd[test_file].map = 0;	// forces use of the m_ansi() code
	if
		(
		0
#if	AP_BUILD || APL_IME
#else //AP_BUILD || APL_IME
	||	RuntimeIsConsole()
#endif//AP_BUILD || APL_IME
	||	0 == is_user_io(test_file)
		)
		{
		in_fd[test_file].map = &redirected_input_map;
		// I need some runtime initialisations here.
		for(i = 256; i--;)
			redirected_input_mode.char_or_index[RawAnso[i]] = i;
		redirected_input_mode.char_or_index[10] = ER;// direct translation
		redirected_input_mode.char_or_index[13] = 0; // row 0
		}
	return 0;
#else //0 == _CONSOLE
	prestarterr(CANT_OPEN);
#endif//0 == _CONSOLE
	}

// Have we already got this keyboard map 

for(i = -1; test_file > ++i;)
	{
	if
		(
		in_fd[i].map		// we have a map for fd == i 
	&&	match_unique_fid
			(
			trans_unique_fid
		,	in_fd[i].map->header.unique_fid
			)
		)
		{
		in_fd[test_file] = in_fd[i];
		close(trans_file);
		return 0;	// need not read the file 
		}
	}
return 1;	// caller must read the file 
}

static UCH *currstr; // address for next string 
static UCH *currbuflim;// space for current translate table 
static unsigned i_any_get(Imap *t);

void i_filbuf(void)
{
void *buf = malloc(MAX_BUF);
Imap *t = (Imap *)buf;
Imap *new;
int reallocmoves = checkrealloc();

sort_line_codes();

if(0 == buf) (*x_error)(MALLOC_FAIL);
currbuflim = (UCH *)buf + MAX_BUF;
initbytes(buf, MAX_BUF, 0);
t->header.unique_fid = trans_unique_fid;
t->freefn=free;		// to release later

if (reallocmoves)
	i_any_get(t);
else
	new = realloc(t, i_any_get(t));
}

// replicate this from pocket.h. trans.c can't include pocket.h (APs)
// define a macro to round the size of a structure in bytes to the appropriate
// MEMORYUNS boundary. Same style for alignment of bits in bitfields.

#define	UNSMASK	(sizeof(MEMORYUNS)-1)
#define	UNSEXTRA(bytes)	((1 + ((bytes) ^ UNSMASK)) & UNSMASK)
#define	UNSROUND(bytes)	((bytes) + UNSEXTRA(bytes))

static void * w_align(void *ptr, void *from)
{
UCP cp = ptr;
UCP bp = from;
unsigned len = cp - bp;

/*
 *	Problem to solve:
 *	Align i to 4|8 byte boundary measured from t which may not be
 *	4|8 byte aligned
 */

return bp + UNSROUND(len);
}



#define ILLEGAL -87024	// use random number as flag for illegal 

/*	default mode change action - adopt the new mode
*/

unsigned	mc_dflt
	(
	ubyte		*cmode
,	Imap	*imp
,	unsigned	up
,	unsigned	thismode
,	unsigned	newmode
	)
	{
	if(!imp->mode[newmode]->istemp)
		cmode[up] = newmode;

	return newmode;
	}

/*	mode change action for via - change the OTHER mode, retain current.
*/

static unsigned	mc_via(
ubyte *cmode,
Imap	*imp,
unsigned	up,
unsigned	thismode,
unsigned	newmode 
)
{
	if(!imp->mode[newmode]->istemp)
		cmode[!up] = newmode;	// change THE OTHER	

	return thismode;		// retain THIS		
}

#if APL_IME

int ime_key();
int (*readfn)() = ime_key;	// allow substitution by nfiles 

#else

int (*readfn)(int,int) = timed_read;	// allow substitution by nfiles 

#endif // APL_IME

int inkey(		// read a keystroke from file fd using currmode[up]
int fd,			// file descriptor 
int up)			// USER or PROG 
{
	return Inkey(fd, up, mc_dflt);	// default mode change action 
}

int v_inkey(int fd,int  up)	// read a keystroke from file fd using currmode[up]
				// file descriptor 
				// USER or PROG 
{
	return Inkey(fd, up, mc_via);	// use via mode change action 
}

static void log_key(int key)
	{
#if	AP_BUILD
#else //AP_BUILD
	// too many variants of time_t suseconds_t to produce
	// consistant output
	// so define our own: friendly to APL structure.
	size_t ignored;
	typedef struct
		{
		int32_t gigasecs;
		int32_t	secs;
		int32_t nanosecs;
		int32_t key;
		} LOG_ENTRY;
	LOG_ENTRY log_entry;
	const int64_t gbase = 1e9;
#if	UNIX
	struct timeval timeval;

	gettimeofday(&timeval, 0);
	log_entry.gigasecs = (int32_t)(timeval.tv_sec / gbase);
	log_entry.secs = timeval.tv_sec - (log_entry.gigasecs * gbase);
	log_entry.nanosecs = (int32_t)(timeval.tv_usec * 1000);

#elif	WINDOWS3
	SYSTEMTIME sys;
	FILETIME file;
	uint64_t time;	// not sure that this should be unsigned
			// two DWORDS are unsigned. For this 
			// purpose we are only interested in 
			// unsigned anyway.
	const uint64_t nbase = 1e7;

	GetSystemTime(&sys);
	SystemTimeToFileTime(&sys, &file);
	time = file.dwHighDateTime;
	time <<= 32;
	time += file.dwLowDateTime; // time units are 100 nano
	log_entry.gigasecs = (int32_t)(time / (gbase * nbase));
	time -= log_entry.gigasecs * (gbase * nbase);
	log_entry.secs = (int32_t)(time / nbase);
	time -= log_entry.secs * nbase;
	log_entry.nanosecs = (int32_t)(time * 100);
#endif//UNIX
	log_entry.key = key;

	log_entry.gigasecs = htonl(log_entry.gigasecs);
	log_entry.secs = htonl(log_entry.secs);
	log_entry.nanosecs = htonl(log_entry.nanosecs);
	log_entry.key = htonl(log_entry.key);

	ignored = write(keylog, &log_entry, sizeof(log_entry));
	fsync(keylog);
#endif//AP_BUILD
	}

/*	read a keystroke from file fd using currmode[up]
	handle mode changes with mc_fn.
*/

int Inkey
	(
	int fd	// file descriptor 
,	int up	// USER or PROG 
,	unsigned (*mc_fn)
		(
		ubyte *
	,	Imap *
	,	unsigned
	,	unsigned
	,	unsigned
		)
	)
	{
	unsigned thismode;
	int key;
	Imap *imp = in_fd[fd].map;

	if(0 == imp)		// no  keyboard translate file so do our best
#if	defined(GUI) && defined(WINDOWS3) && !APL_IME
		{		// ANSI keyboard has no temp modes
		for
			(
		;	key = m_ansi(fd, 0)
		,	ILLEGAL == key
		;
			);
		return key;
		}
#else //defined(GUI) && defined(WINDOWS3)
		{
		return ILLEGAL;
		}
#endif//defined(GUI) && defined(WINDOWS3)

	up = up == USER;

	getperm:		// get current permanent mode	
	thismode = in_fd[fd].curr_mode[up];

	for(;;)
		{
		key = m_inkey(fd, imp->mode[thismode]);

		if(key == ILLEGAL)
			goto getperm;		// illegal key leaves temp mode 

#if USES_IME
			{	
			int av = key;
#else
		if (key == UNICODEFOLLOWS)
			{	
			int av = key = (*readfn)(fd, 0);
#endif
			if (key==0x9) 
				return HT;
#if HAS_UNICODE
			else if (IsLikeNewLine(key))
				return ER;
#endif
			else if (key==0x20) 
				return DP;
			else if (key==0) 
				return 0;
			else if (av = UNICODEtoKEY(key))
				return av;
			goto getperm;
			}

#if !APL_IME
		if(logkey)
			log_key(key);
#endif //APL_IME

		if
			(
			key <= M0		// change to specific mode	
		||	key == MC		// change to next mode in cycle	
			) 
			{
			thismode = (*mc_fn)
				(
				in_fd[fd].curr_mode
			,	imp
			,	up
			,	thismode
			,	(key <= M0)
				?	M0 - key
				:	(thismode+1) % imp->n_modes
				);
			}
		else
			{
			int subst;

			return 
				(
				imp->caps_on
			&&	key > 4
			&&	(subst = imp->caps_xchg[key])
				) ? subst : key;
			}
		}
	}


// Late on in 10.1 and in 11.0 we added support for
// the user dynamically changing his windows keyboard layout
// when (s)he does this we stuff (for example) 7 7 e n _ U K
// into the input buffer.

// if the appropriate codes are NOT in the translate table
// m_inkey will absort the illegal character
// m_ansi however does not and we end with "junk" characters
// in the input stream (and then into the GUI)

// We tried making m_ansi absorb the illegal characters
// but that removed some messages from the APL event list
// and therefore could break code

// so. We detect the use of m_ansi and don't bother puttin
// the layout sequences into the input stream.

// the value of the ansi flag to gui_Inkey is irrelevant
// in each case we want the same behaviour

bool handles_layout_switch(int fd)
{
Imap *imp = in_fd[fd].map;

return 0!=imp;
}

int gui_Inkey
	(
	int fd	// file descriptor 
,	int up	// USER or PROG 
,	unsigned (*mc_fn)
		(
		ubyte *
	,	Imap *
	,	unsigned
	,	unsigned
	,	unsigned
		)
,	int ansi
,	int allow_mode
	)
	{
	static unsigned thismode;
	int key;
	Imap *imp = in_fd[fd].map;
	static int get_mode = 1;

	if(ansi || 0 == imp)	// no  keyboard translate file so do our best
#if	defined(GUI) && defined(WINDOWS3) && !APL_IME
		{
		key = m_ansi(fd, 0);
		return ILLEGAL == key ? 0 : key;
		}
#else //defined(GUI) && defined(WINDOWS3)
		{
		return ILLEGAL;
		}
#endif//defined(GUI) && defined(WINDOWS3)

	up = up == USER;

	if (get_mode)
		{
		thismode = in_fd[fd].curr_mode[up];
		get_mode = 0;
		}

	for(;;)
		{
		key = m_inkey(fd, imp->mode[thismode]);

		if(key == ILLEGAL)
			{
			// illegal key leaves temp mode 
			thismode = in_fd[fd].curr_mode[up];
			return 0;
			}

		if (key == MTQUEUE)
			return key;

		if (key == UNICODEFOLLOWS)
			{	
			int av=key=(*readfn)(fd, 0);
			if (key==0x9) 
				return TB;
			else if (av = UNICODEtoKEY(key))
				return av;
			return 0;
			}

#if !APL_IME
		if(logkey)
			{
			union
				{
				short	s;
				UCH	c[sizeof(short)];
				} sc;
			size_t ignored;

			sc.s = key;

			ignored = write(keylog, sc.c, sizeof(short));
			fsync(keylog);
			}
#endif //APL_IME

		if
			(
			key <= M0		// change to specific mode	
		||	key == MC		// change to next mode in cycle	
			) 
			{
			int mode = (*mc_fn)
				(
				in_fd[fd].curr_mode,
				imp,
				up,
				thismode,
				(key <= M0)
				?	M0 - key
				:	(thismode+1) % imp->n_modes
				);
			if (allow_mode)
				thismode=mode;
#if APL_IME
			return MC; 
#endif
			}	
		else
			{
			int subst;
			get_mode = key;

			return 
				(
				imp->caps_on
			&&	key > 4
			&&	(subst = imp->caps_xchg[key])
				) ? subst : key;
			}
		}
	}

int upper_of(int fd, achar key)
{
	Imap *imp = in_fd[fd].map;

	return
		(
		   imp		// there is a keyboard translate 
		&& key > 4	// sensible key to translate 
		&& BIT_READ(imp->is_lower, key)	// specified as lower case 
		) ? imp->caps_xchg[key] : key;
}

int lower_of(int fd, achar key)
{
	Imap *imp = in_fd[fd].map;
	int subst;

	return
		(
		   imp		// there is a keyboard translate 
		&& key > 4	// sensible key to translate 
		&& ( subst = imp->caps_xchg[key] ) // something there 
		&& BIT_READ(imp->is_lower, subst) // target is lower case 
		) ? subst : key;
}

void set_key_caps_lock(int flag)
{
	Imap *mp = in_fd[0].map;

	if(mp) mp->caps_on = flag;
}


#ifdef	XAPL
int	XAPLmouse_event;
#endif	// XAPL 

static UCH * copy_left(UCH *fifo, UCH *fifo_ptr)
{
	UCH *fp;

	/*
	 * consider the following hypothetical situation - map expects 
	 * key to send abc, key actually sends abd, a key which sends da is
	 * defined - user continually hits key which sends abd
	 */

	if(--fifo_ptr != fifo) for
		(
		fp = fifo
	;	*fp++ = *fifo_ptr
	,	0 != *fifo_ptr++
	;
		);
	return fifo;
}

#if	AP_BUILD || APL_IME
	static unsigned testresize(void) { return 0; }
#else //AP_BUILD
	extern unsigned testresize(void);
#endif//AP_BUILD

static int m_inkey(int fd, Mmap *mp)	// read a moded keystroke 
{
int k, v;
UCP candp;	// candidate pointer 
ubyte *rp;	// row pointer 
short candres;			// candidate result 
unsigned stripped = mp->doistrip;
unsigned timed_out = 0;

static UCH fifo[32]; // space for when we have read too many chars 
static UCH *fifo_ptr = fifo;


#ifdef	XAPL
XAPLmouse_event = 0;	// X_timed_read can send back mouse events 
#endif	// XAPL 

k = *fifo_ptr++;
if(!k)	// no UCH in buffer 
	{
	*(fifo_ptr = fifo) = 0;	// buffer is empty 
	testresize();

	k = (*readfn)(fd, 1);

#if	USES_IME
	if (k==-1)
		return 0;
	return k;
#endif//USES_IME
	switch(k)
		{
	case -1:
		*fifo_ptr++ = k;
		*fifo_ptr = 0;
		return 0;
	case MTQUEUE:
		return k;
	case UNICODEFOLLOWS:
		return k;
	case -3:
#if	0  //==GIVING_PRESENTATIONS
		return WI;	// interrupted 
#endif//0  //==GIVING_PRESENTATIONS
	testresize();
	default:
		if(k < 0)
			{
			// I would like to test for M0 - nmodes, but I don't
			// have that value within the moded structure.
			// The test is necessary because of a dodgy HP
			// intellimouse driver that does strange things with
			// WM_MOUSEWHEEL
			if(k < M0 - 255) goto ret_utf;
			return k;
			}
		*fifo_ptr++ = k;
		*fifo_ptr = 0;
		break;
		}
	}
#ifdef	XAPL
if (XAPLmouse_event) return k;		// already a code	
#endif	// XAPL 
k &= stripped ? 0x7f : 0xff;
	
v=mp->char_or_index[k];

if(BIT_READ(mp->direct, k))
	{
	if (v)
		return v;	// normal exit 
	fifo_ptr = copy_left(fifo, fifo_ptr);
	goto ret_utf;
	}

rp = mp->strs + v*mp->cols;

if(*rp != k)
	{
	fifo_ptr = copy_left(fifo, fifo_ptr);
	goto ret_utf;	// no entries at all for this first UCH 
	}

// special case single character special keys 

if(!rp[1])
	return mp->result[v];

fifo_ptr = copy_left(fifo, fifo_ptr);

candp = rp; rp += mp->cols;
candres = mp->result[v++];

timed_out = 0;
				
for(;;)
	{
	UCH f = *fifo_ptr;

	if(stripped) f &= 0x7f;
	switch(((*candp != 0) << 1 ) | (f != 0))
		{
		case 0:
			// perfect match 
			*(fifo_ptr = fifo) = 0; // throw the buffer away 
			return candres;
		case 1:
			// matched but the buffer is not empty 
#if	0
			// original code special cased ESC here. No idea why!
#else
                        if(fifo[0] == 27 && mp->isatty)break;
#endif//0
			return candres;
		case 2:
			// no more characters in the buffer 
			if(!timed_out)
				{
				int tr_val = (*readfn)(fd, 0);

				switch(tr_val)
					{
				case -3:
					*fifo_ptr = 0;
					fifo_ptr = fifo;
					return WI;	// interrupt 
				case MTQUEUE:
				case -2:
					timed_out = 1;
					*fifo_ptr = 0;
					break;
				default:
					*fifo_ptr = tr_val;
					break;
					}

				*(fifo_ptr+1) = 0;
				continue;
				}
			/* no more characters to come so fall through to
			   definite mismatch code
			 */
		case 3:
			// characters in both 

			if(*candp == f){candp++; fifo_ptr++; continue;}
			
			// definate mismatch 

			candp = rp; rp += mp->cols;
			candres = mp->result[v++];
			fifo_ptr = fifo;
			if(*candp == k)
				{
				candp++;
				fifo_ptr++;
				continue;
				}

			// no match possible 

			break;
		}
	break;
	}
ret_utf:
switch(mp->utfwidth)
	{
	ubyte first;
	unsigned nbyte;
	signed char mask;
	unsigned shift;
	UCP fp;
	unsigned sixbit;
case 0:
case 16:
	break;
case 8:
	first = fifo[0];
	if(0x80 & first) for
		(
		nbyte = 0
	;	0x80 & first
	;	first <<= 1
	,	++nbyte
		)
		{
		if(timed_out)			// no more to come
			goto ret_illegal;
		else if(*fifo_ptr)
			{
			++fifo_ptr;
			continue;
			}
		switch(k = (*readfn)(fd, 0))
			{
		case -3:
			*(fifo_ptr = fifo) = 0;
			return WI;
		case -2:
		case MTQUEUE:
			timed_out = 1;
			*fifo_ptr = 0;
			goto ret_illegal;
		default:
			*fifo_ptr++ = k;
			*fifo_ptr = 0;
			continue;
			}
		}
	else
		{
		nbyte = 1;
		++fifo_ptr;
		}
	for
		(
		fp = fifo
	,	mask = 0x80
	,	shift = 0
	,	sixbit = 0xc0
	,	k = 0
	,	mask >>= nbyte - 1
	;	nbyte--
	;
		)
		{
		k += (fp[nbyte] & ~(0 == nbyte ? mask : sixbit)) << shift;
		shift += 6;
		}
	return k;
	}
ret_illegal:
if(!timed_out)	// wait for time out, absorb file if not tty
	while(0 <= (*readfn)(fd, 0));
*(fifo_ptr = fifo) = 0; // throw buffer away 
return ILLEGAL;
}


#if	defined(GUI) && defined(WINDOWS3) && !APL_IME

static int m_ansi(int fd, Mmap *mp)
	{
	static int control_map[33] =
		{
		0  			// 00	 (EOF)
	,	ILLEGAL			// 01
	,	ILLEGAL			// 02
	,	CP			// 03
	,	ILLEGAL			// 04
	,	ILLEGAL			// 05
	,	ILLEGAL			// 06
	,	ILLEGAL			// 07
	,	DB			// 08
	,	TB			// 09
	,	ILLEGAL			// 10
	,	ILLEGAL			// 11
	,	ILLEGAL			// 12
	,	ER			// 13
	,	ILLEGAL			// 14
	,	ILLEGAL			// 15
	,	ILLEGAL			// 16
	,	ILLEGAL			// 17
	,	ILLEGAL			// 18
	,	ILLEGAL			// 19
	,	ILLEGAL			// 20
	,	ILLEGAL			// 21
	,	PT			// 22
	,	ILLEGAL			// 23
	,	CT			// 24
	,	ILLEGAL			// 25
	,	ILLEGAL			// 26
	,	EP			// 27
	,	ILLEGAL			// 28
	,	ILLEGAL			// 29
	,	ILLEGAL			// 30
	,	ILLEGAL			// 31
	,	DP			// 32
		};

	int ansi;
	int value;
	int shift = 0;
	int code;

#if USES_IME && !AP_BUILD
	if (0==is_user_io(fd))
		{
		ansi=WindowsFileRead(fd);
		return (ansi <= 32) ? control_map[ansi] : ansi;
		}
#endif

	W3_timed_ansi(0, 1, &value, &ansi);

	if (value==UNICODEFOLLOWS)
		{
		W3_timed_ansi(0, 1, &value, &ansi);
		if (value==AV_SP)value=DP;
		return UNICODEtoKEY(value);
		}

	if(0 > value) return value;

	for(;-1 == ansi;)
		{
		shift = value;
		W3_timed_ansi(0, 1, &value, &ansi);
		if(0 > value)	// defensive
			{
			// defend against HP intellimouse driver
			if(M0 > value) return ILLEGAL;
			return value;	
			}
		}

	if(ansi)
		{
#if	HAS_UNICODE
		return ansi;
#else //HAS_UNICODE
		if(ansi<=32)
			return control_map[ansi];
		else			// normal character
			{
			UCH c = ansi;

			ansi2av(&c, 1);
			return c;
			}
#endif//HAS_UNICODE
		}
#if	AP_BUILD
	code = ILLEGAL;
#else //AP_BUILD
	code = vk2key(value, shift);
#endif//AP_BUILD

#if 0
// if you need to put this back you'll need to fix
// Mantis [3980]

#if	HAS_UNICODE
	if(ILLEGAL == code)
		return value;
#endif//HAS_UNICODE
#endif

	return code;
	}
#endif//defined(GUI) && defined(WINDOWS3)

achar *inlab(int fd, achar * name)
{
	Imap *mp = in_fd[fd].map;
	static achar anull[1] = { 0 };		// null string
	achar *lab = anull;
	
	if(mp && mp->labinfo)lab = mp->labinfo;

	while('\0' != *lab)			// not labinfo terminator 
		{
		if(astrcmp(lab, name))
			{
			lab += 1+astrlen(lab);	// skip wrong name	  
			lab += 1+astrlen(lab);	// skip its label	  
			}
		else
			{
			lab += 1+astrlen(lab);	// skip right name	  
			return lab;		// return its label	  
			}
		}

	return lab;				// pointer to empty string 
}

	// return current programmer mode and set new mode 
unsigned keyset(int fd,unsigned new,int up)
				// file descriptor 
				// new mode 
				// USER or PROG 
{
int old;

up = up == USER;
old = in_fd[fd].curr_mode[up];
in_fd[fd].curr_mode[up] = new;
return old;
}



static struct Out_fd	*out;		// relevent output structure 
int	(*out_fn)(ubyte *,unsigned int);// output function to use GLOBAL for prt.c 
void	(*out_fail)(void);	// what to do on failure global for prt.c 
static ubyte 	state;		// current state 
static ubyte	used;		// anything gone out for ... 

unsigned out_fld = 1;	// output file descriptor global for sm.c 
FILE	 *out_sd;	// output stream (if used)global for nfiles.c 

static struct Out_fd out_fd[_NFILE];
       struct Out_fd *o_out[_NFILE];	// one for each file descriptor 
					// nfiles ap needs this

	// forward references - this file
static struct Out_fd raw_fd;
static struct Translate raw_output;
static struct Translate redirected_output;


void	(*o_fn)();	// communication function 

UCH fbt_buff[4096];	// was 256 but multi screens will be bigger
UCH syncol_buff[4096];	// was 256 but multi screens will be bigger
UCP  fbt_buffptr;

int fbt_out_buf(gchar * cp, unsigned count)
{
	// imbedded carriage ^s may be a problem 

	unsigned ret = count;
	for(;count--;)*fbt_buffptr++ = *cp++;
	return ret;
}

static UCH *wincharsptr;

static int WINCHARS_out(ubyte *cp, unsigned count)
	{
	unsigned ret = count;

	if(count) *wincharsptr = (UCH)*cp; // only take one character - translate
	return ret;			// should be 1 to 1 mapping
	}

static ubyte *colorptr;

static int COLOR_out(ubyte *vp, unsigned count)
	{
	unsigned ret = count;
	ubyte *cp=(ubyte*)vp;
	
	if(count > 3) count = 3;	// defensive
	for(;count--;)*colorptr++ = (ubyte)*cp++;
	return ret;
	}


static void cache_tab(void)	// cache the translate table    
	{
	unsigned i;

	out = o_out[1];
	if(0 == out || 0 == is_user_io(1))
		{
		out = o_out[2];
		if(0 == is_user_io(2)) out = &raw_fd;
		}
	out_fn = COLOR_out;

	memset(BACKGROUND,0xFF,sizeof(BACKGROUND));

	for(i = 256; i--;)
		{
		colorptr = &FOREGROUND[i][0]; special(C0);
		colorptr = &FOREGROUND[i][0]; special(C0 - i);
		colorptr = &BACKGROUND[i][0]; special(B0);
		colorptr = &BACKGROUND[i][0]; special(B0 - i);
		}
#if	GUI && defined(WINDOWS3)
#if	AP_BUILD || APL_IME
#else //AP_BUILD || APL_IME
	memcpy(SESHFOREGROUND, FOREGROUND, sizeof(FOREGROUND));
	memcpy(SESHBACKGROUND, BACKGROUND, sizeof(BACKGROUND));
	memcpy(PRINTFOREGROUND, FOREGROUND, sizeof(FOREGROUND));
	memcpy(PRINTBACKGROUND, BACKGROUND, sizeof(BACKGROUND));
#endif//AP_BUILD || APL_IME
	out_fn = WINCHARS_out;
	for(i = 256; i--;)
		{
		wincharsptr = &WINCHARS[i]; avtoasc((UCH)i);
		BODGECHARS[WINCHARS[i]] = i;
		}
#else //GUI && defined(WINDOWS3)
	for(i = 256; i--;)		// doing it here makes MAPCHARS work
		{
		WINCHARS[i] = i;
		BODGECHARS[i] = i;
		}
	av2ansi(WINCHARS, 256);		// uses internal win.dot
	ansi2av(BODGECHARS, 256);	// uses internal win.dot
#endif//GUI && defined(WINDOWS3)
#if !HAS_UNICODE
	BODGECHARS[WINCHARS[AV_SQUISH]] = AV_SQUISH; // special case this one 
#endif // HAS_UNICODE
	cached = 1;
	}


void o_fname	// develop an output file name 
	(
	gchar *tp		// null or target file name 
,	gchar *buf		// where to put the file name 
,	unsigned errflag		// an error condition ? 
	)
	{
	t_fname(tp, GSTR("APLTRANS"), GSTR("apltrans"), GSTR("APLT  "), GSTR(".DOT"), buf, errflag);
	}

void  LIBCALL o_finish(void);

void o_init(void)
{
/*
 *	allocate necessary translate tables at run time (malloced)
 *	if d_init() is used instead of o_init() then a date map
 *	and a keyboard map are installed as well.
 *	if i_init() is used instead of o_init() then keyboard map
 *	is installed as well.
 */

if(!x_error || x_error == d_error) x_error = init_err;
fname   = o_fname;

for(test_file = 0; _NFILE > test_file; ++test_file) switch(test_file)
	{
	case 0:		// outlaw using 0 for output 
#if	!defined(WATCOM)
	case INPIPE: case OUTPIPE:
#endif//!defined(WATCOM)
		break;
	default:
		if(o_get(0)) o_filbuf();
		break;
	}
cache_tab();
atexit(o_finish);
}


static unsigned o_any_get(struct Translate *t);

void o_filbuf(void)
{
UCP buf = malloc(MAX_BUF);
struct Translate *t = (struct Translate *)buf;
struct Translate *new;
int reallocmoves = checkrealloc();

if(0 == buf) (*x_error)(MALLOC_FAIL);
currbuflim = (UCH *)buf + MAX_BUF;
initbytes(buf, MAX_BUF, 0);
t->unique_fid = trans_unique_fid;

if (reallocmoves)
	o_any_get(t);
else
	new = realloc(t, o_any_get(t));
}


static void t_abort(void)
{
(*x_error)(FILE2BIG);
}

int o_get(gchar *tp)
{
struct stat statbuf={0};
gchar envbuf[MAXFCHARS];
int i;
struct Out_fd *this;

#ifndef	WINCE
// test is the descriptor open for write 

if(-1 == fstat(test_file,&statbuf)) return 0;	// descriptor not open 

#ifdef	FCNTL
i = fcntl(test_file, F_GETFL, 0);
if(-1 == i) return 0;			// unsuitable descriptor 

i &= O_RDONLY | O_RDWR | O_WRONLY;
if(i != O_RDWR && i != O_WRONLY) return 0; // not write 

#else  
if
	(
	0 == 
		(
		statbuf.st_mode
#if	defined(MSC)
	// only Microsoft could be this daft. If you have a descriptor connected
	// to NUL then it is _S_IFCHAR but has no permissions.
	&	(_S_IFCHR | _S_IFIFO | _S_IWRITE) 
#else //defined(MSC)
	&	0222
#endif//defined(MSC)
		)
	) return 0;    // not write 
		// cant test that it is actually write 
#endif	// FCNTL 
#endif //WINCE

// do we already know about this device 

for(i= -1;test_file> ++i;)
	{
	if
		(
		o_out[i]
	&&	match_unique_fid
			(
			o_out[i]->unique_fid
		,	fd_to_unique_fid(test_file)
			)
		)
		{
		o_out[test_file] = o_out[i];	// already know it 
		return 0;
		}
	}

// initialise an output struct 

this = o_out[test_file] = &out_fd[test_file];
this->unique_fid = fd_to_unique_fid(test_file);
this->curr_col = 0;

o_fname(0, 0, 0);		// init statics for top level request 
o_fname(tp, envbuf, 0);	// put file name to read in envbuf 

if(-1 == t_fopen(envbuf, GSTR("\n")))
#if	0 == _CONSOLE
	{			// substitute table GUI specific
	struct Translate *choice[2] = {&raw_output, &redirected_output};
	unsigned index = 
		0
#if	AP_BUILD || APL_IME
#else //AP_BUILD || APL_IME
	||	RuntimeIsConsole()
#endif//AP_BUILD || APL_IME
	||	0 == is_user_io(test_file)
	;

	++(this->t_out = choice[index])->refcount;
	return 0;
	}
#else //0 == _CONSOLE
	prestarterr(CANT_OPEN);
#endif//0 == _CONSOLE

// do we already know about this translate table 

for(i= -1;test_file > ++i;)
	{
	if
		(
		o_out[i]
	&&	match_unique_fid
			(
			trans_unique_fid
		,	o_out[i]->t_out->unique_fid
			)
		)
		{
		(o_out[test_file]->t_out = o_out[i]->t_out)->refcount++;
		close(trans_file);
		return 0;
		}
	}

// we need to read the file 

return 1;
}

void  LIBCALL o_finish(void)
{
/* put all files into known state for exit -
 * use RESET: string
 */
unsigned i, j;
	
if(out_fn == out_stream)
	fflush(out_sd);

for(i=_NFILE;i--;)if(out = o_out[i])
	{
	if(out->out_used)
		{
		out_fld = i;
		out_fn = out_raw;
		out_fail = 0;
		special(RESET);
		}
	for(j=i;j--;)if(o_out[j] == out)
		{
		o_out[j] = 0;
		}
	if(0 == --out->t_out->refcount)
		free(out->t_out);
	}
}

static void eno(unsigned n)	// print n 
{
UCH c;
size_t ignored;
if(n>=10)eno(n/10);
c = '0' + n%10;
ignored = write(2, &c, 1);
}

int err_eq_out(void)
{
	/*
	 * is device connected to file descriptor 2 the same device
	 * as that connected to file descriptor 1
	 */
return o_out[1]==o_out[2];
}


int init_out(UPTR describe, void (*fail)(void))
{
unsigned fd;	// the file descriptor 
static unsigned recur = 0;	// defend against recursion during (*fail)()
				// fail ... error ... errormess ... init_out
				// can happen when handling SIGHUP ... EOF_INTERRUPT

if(recur++) return 0;
if(describe>_NFILE)	// stream 
	{
	FILE *stream = (FILE *)describe;

/* make output exactly match the output translate i.e.
	if we put out 13 10 we do not want DOS to make it
	13 13 10
*/
#if	DOSWIN && !WINCE
	binary_stream(stream);
#endif//DOSWIN && !WINCE
	fd = (int)fileno(stream);
	out_sd = stream;
	out_fn = out_stream;
	}
else
	{		// file descriptor 
	fd = describe;
	out_fn = out_raw;
	}

#if defined(WINCE)
return 0;
#endif

if(out = o_out[fd])
	{
#ifdef	XAPL
		{
		extern unsigned iam_X_console;

		if(!out->unused)	// don't know yet if an X console 
			out->unused = isanX(fd);
		if(iam_X_console = out->unused) out_fn = X_out;
		if(iam_X_console) loadXcolours(out);
		}
#endif	// XAPL 
#if	GUI && defined(WINDOWS3)
#	if	AP_BUILD || APL_IME
#	else  //AP_BUILD
		if (RuntimeHasSession())
#	endif //AP_BUILD
			{

			if(is_user_io(fd))out_fn = WIN_out;
			}
#endif//GUI && defined(WINDOWS3)
	out->out_used = 1;	// flag this fd as used for o_finish() 
	out_fld = fd;
	out_fail = fail;
	special(INIT);
	state = 0;	// set to first 
	used  = 0;	// set to unused  
	}
else 
	{
	(*fail)();
	}
--recur;
return out ? 0 : -1;
}

void special(int code)
{
struct Translate *t = out->t_out;

if (t)	// not sure under what circumstances there won't be one.
	// somewhere in the UNICODE interpreter.
	{
	struct Sas *s = t->special;
	unsigned n = t->nsas;
	unsigned count;

	if(!n) return;			// no strings 

	#if defined(WINDOWS3)
	if (code==INIT)
		return;
	#endif

	if(code == INIT && s->code != INIT)return; // special case no INIT string 
	for(;n--; s++)
		{
		if(s->code == code)
			{
			count = s->len;
			if(!count) return;
			(*out_fn)(t->strstart + s->offset, count);
			return;
			}
		}
	}
return;
}

static int out_stream(ubyte *cp, unsigned count)
{
int sent;
// LARGE_ARRAY: (int) case below is naff

#if USES_IME && !AP_BUILD && defined(WINDOWS3)
	sent = WindowsFileWrite((int)_fileno(out_sd),cp,count);
#else
	sent = (int)fwrite(cp, 1, count, out_sd);
#endif

if(sent != count && out_fail)(*out_fail)();
return sent;
}

static int out_raw(ubyte *cp, unsigned count)
{
int sent = write(out_fld, cp, count);

if(sent != count && out_fail)(*out_fail)();
return sent;
}

static int g_count;	// communication from out_count() to t_count() 

static int out_count(ubyte *cp, unsigned count)
{
return g_count = count;
}

int t_count(void (*fn)(int), int param)
			// function to measure ascsp() or avtoasc() 
			// parameters to pass it 
{
int (*save)(ubyte *,unsigned int) = out_fn;		// shadow out_fn 

g_count = 0;			// in case (*out_fn)() not called 
out_fn = out_count;		// lie about output function 
(*fn)(param);			// call the requested function 
out_fn = save;			// unshadow 
return g_count;
}

/*
 * functions / globals to help nfiles go faster 
 */

UCH t_is_safe;
UCH t_safe_code;

void t_safe_init_out(UPTR fd)
{
	// cut down init_out() - I dont want to call SPECIAL() 
	out = o_out[fd];
	state = 0;
	used = 0;
}

int t_safe(UCP cp, unsigned count)
{
	t_is_safe = 1 == count && 1 == out->t_out->nmodes;
	t_safe_code = *cp;
	return count;
}

static UCH *buf_ptr;	// comm from X_buffer() to t_buffer() 

static int X_buffer(ubyte *cp, unsigned int count)
{
buf_ptr = cp;
return count;
}

UCH *t_buffer(void (*fn)(int),int p1)
			// function to measure special() 
			// parameters to pass it 
{
int (*save)(ubyte *,unsigned) = out_fn;		// shadow out_fn 

buf_ptr = 0;			// in case (*out_fn)() not called 
out_fn = X_buffer;		// lie about output function 
(*fn)(p1);			// call the requested function 
out_fn = save;			// unshadow 
return buf_ptr;
}

static unsigned av_out(struct Translate *t,unsigned row,UCH **bufpt)
{
short unsigned sum, n;
UCP rp = t->lengths + row*t->nmodes;

sum = 0;
if(state)for(n = 0; n++ < state; ) sum += *(rp++);
n = *rp;
rp = t->strstart + t->offset[row] + sum;
switch(n)
	{
case 1:
	**bufpt = *rp;
	break;
default:
	memcpy(*bufpt, rp, n);
	break;
	}
*bufpt += n;
return n;
}

void avtoasc(achar v)
{
UCH buf[32];		// output buffer 
UCH *cp = buf;

#if	0 == _CONSOLE	&& HAS_UNICODE
if(WINCHARS_out != out_fn)
	{
	(*(achar *)buf)=v;
	cp+=sizeof(v);
	(*out_fn)(buf, cp-buf);
	}
	else
#endif//0 == _CONSOLE	&& HAS_UNICODE
	{
	achar c = v;		// keep a copy
	struct Translate *t = out->t_out;	// translate table 

	#if	HAS_UNICODE
		c = value2line(t, v);		// the reason for the free slot
						// if c points there it can't go
						// out in any mode
	#else //HAS_UNICODE
		ubyte uc = (ubyte)c;
	if
		(
	#if	XAPL
		X_out == out_fn
	#elif	GUI && defined(WINDOWS3)
		WIN_out == out_fn
	#else//special cases
		0
	#endif//special cases
		)
		switch(uc)
		{
	case 1: case 2: case 3:		// cursor control character 
		(*out_fn)(&uc, -uc);
		return;
		}
	#endif//HAS_UNICODE

	if(!av_out(t, c, &cp)) // cant go out in current mode 
		{
		UCH *lp = t->lengths + c*t->nmodes;
		unsigned n;

		for(n = 0; n < t->nmodes; n++, lp++)
			{
			/*
			 * note that
			 * offset[c] is the start of the output string
			 * ( all previous counts are 0 ).
			 * This code does not use this fact.
			 */
			if(*lp)	// will go out in mode n 
				{
				// change to mode n from mode state 
				av_out(t, MODE0+n, &cp);
				state = n;
				// now output the character 
				av_out(t, c, &cp);
				break;
				}
			}
		if(8 == t->utfwidth)
			{
			if(v <= 0x7f)
				{		// only one byte needed
				*cp = v;
				(*out_fn)(cp, 1);
				}
			else
				{
				UCP bp = buf + sizeof(buf); 	// right to left
				UCP top = bp;
				int16_t lead =
#if	SOLARIS
					(int16_t)
#endif//SOLARIS
					0xff00;
			
				for
					(
				;	*--bp = 0x80 | (v & 0x3f) // rightmost six bits
				,	v >>= 6
				,	lead >>=1
				,	v
				;
					);
				*bp |= lead & 0xff;
				(*out_fn)(bp, top - bp);
				}
			return;
			}
		if(cp == buf)	// still failed to find a character 
			{
			if(v)    // avoid recursion
				{
				avtoasc(0);	// print character 0 instead 
				return;
				}
			// else put out nothing but still worry about backspace etc 
			}
		}

	(*out_fn)(buf, cp-buf);

	used = 1;

	switch(v)	// try to cope with cursor movement characters in quadav 
		{
		default:
			out->curr_col++;	break;
		case AV_BS:	// backspace 
			if(out->curr_col) out->curr_col--; break;
		case AV_LF:	// line feed 
			// character output above should have been newline 

			{
			unsigned i;

			for(i=out->curr_col; i--;) avtoasc(AV_SP); break;
			}
		case AV_NEWLINE:
			out->curr_col = 0; break;
		}
	}
}



void ascsp(int code)
/*
 * ensure a blank on screen in case screen supports an overstrike
 * which we dont recognise
 */
{
struct Translate *t = out->t_out;	// translate table 
UCH buf[32];				// output buffer 
UCP cp = buf;
UCP tp;

tp = cp;
if(!av_out(t, CC, &cp)) // cant go out in current mode 
	{
	UCH *lp = t->lengths + CC*t->nmodes;
	unsigned n;

	for(n = 0; n < t->nmodes; n++, lp++)
		{
		if(*lp)	// will go out in mode n 
			{
			// change to mode n from mode state 
			av_out(t, MODE0+n, &cp);
			state = n;
			// now output the character 
			av_out(t, CC, &cp);
			break;
			}
		}
	}
if(cp == tp)return;	// abort if cant be done 
(*out_fn)(buf, cp-buf);
used = 1;
}

#if	defined(XAPL)
UCH avtoXasc(c)
UCH c;
/* obtain output character corresponding to quadAV character 'c'
 * for X Window implementation
 */
{
struct Translate *t = out->t_out;	// translate table 
UCH buf[32];				// output buffer 
UCP cp = buf;
		
if(!av_out(t, c, &cp)) // cant go out in current mode 
	{
	UCH *lp = t->lengths + c*t->nmodes;
	unsigned n;

	for(n = 0; n < t->nmodes; n++, lp++)
		{
		if(*lp)	// will go out in mode n 
			{
			// change to mode n from mode state 
			av_out(t, MODE0+n, &cp);
			state = n;
			// now output the character 
			av_out(t, c, &cp);
			break;
			}
		}
	}
if(cp == buf)	// still failed to find a character 
	if(c)
		return 0;

return buf[0];		// I know it's only one character for quadAV  
}

#endif//defined(XAPL)


	/*
	 *	GENERAL ROUTINES FOR INPUT, OUTPUT AND DATE FILE INTERPRETERS
	 */

#define	IOD_DECLARE(name)	\
	static UCH *name	\
		(		\
		UCH *cp		\
	,	UCH *ep		\
		)		\

static int i_alphabetic(UCH c)
{
	return (c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z');
}

static int i_digit(UCH c, unsigned *value)
{
	if (c >= '0' && c <= '9')
		;
	else
		return 0;
	if(value)
		*value = c - '0';
	return 1;
}

static int i_hexdigit(UCH c, unsigned *value)
{
	unsigned val;

	if (c >= '0' && c <= '9')
		val = c - '0';
	else if (c >= 'a' && c <= 'f')
		val = 10 + c - 'a';
	else if (c >= 'A' && c <= 'F')
		val = 10 + c - 'A';
	else
		return 0;
	if(value)
		*value = val;
	return 1;
}

static int i_symbolic(UCH c)
{
	return	   c >= ' '
		&& c <= '~'
		&& !i_alphabetic(c)
		&& !i_digit(c, 0)
		&& c != ':';
}

static int i_ascii_char(UCH c)
{
	return i_alphabetic(c) || i_digit(c, 0) || i_symbolic(c);
}

IOD_DECLARE(white)
{
while(*cp == ' ' || *cp == '\t' || *cp == '\r')cp++;
return cp;
}

IOD_DECLARE(comment)
{
if(cp == ep || *cp == '+')return ep;
(*x_error)(SYNTAX);
return 0; 	// keep compiler happy 
}

static unsigned lastnumber;	// for number() to communicate 

IOD_DECLARE(number)
{
UCP wp;
unsigned n = 0;
unsigned dig_val;
enum { DEC, HEX } base_select = DEC;
struct
	{
	unsigned base;
	int (*digit)(UCH, unsigned *);
	} base[2] =
	{
		{ 10, i_digit }
	,	{ 16, i_hexdigit }
	};

	'0' == cp[0]
&&	'x' == cp[1]
&&	(base_select = HEX)
&&	(cp += 2)
;

for
	(
	n = 0
,	wp = cp
;	(*base[base_select].digit)(*cp, &dig_val)
;	n *= base[base_select].base
,	n += dig_val
,	++cp
	) if(n & -(1 << 21))
		{
		(*x_error)(NUM2BIG);
		}
if(wp == cp)(*x_error)(NUM_WOT);
lastnumber = n;
return cp;
}

IOD_DECLARE(byte_number)
{
	cp = number(cp, ep);
	if (lastnumber & -(1 << 8))
		(*x_error)(NUM2BIG);
	return cp;
}


IOD_DECLARE(colon)
{
cp = white(cp, ep);
if(*cp != ':')(*x_error)(NO_COLON);
cp++;
cp = white(cp, ep);
return cp;
}

static unsigned mod_flen;
static gchar mod_fname_buffer[MAXFCHARS];
static gchar *mod_fname;

IOD_DECLARE(modifies_desc)
{
UCP start;

if(memcmp(cp, "MOD", 3)) return cp;
cp += 3;
cp = colon(cp, ep);
mod_fname = mod_fname_buffer;
start = cp;
for(;;)
	{
	switch(*cp)
		{
		default:
			cp++;
			continue;
		case ' ': case '\t': case '\n': case '+':
			mod_flen = cp - start;
			if(MAXFCHARS <= mod_flen)
				(*x_error)(MOD2LONG);
			g_memcpy_i(mod_fname, start, mod_flen);
			break;
		}
	break;
	}

cp = white(cp, ep);
return comment(cp, ep);
}

IOD_DECLARE(mod_line)
{
UCP start;
UCP wp = cp;

for(;;)
	{
	wp = white(wp, ep);
	switch(*wp)
		{
		case '+':
			while(*++wp != '\n');
		case '\n':
			curr_start_of_line = ++wp;
			continue;
		default:
			break;
		}
	break;
	}

if(0 != memcmp(wp, "MOD", 3))
	return cp;	// first non - comment is not MOD 
wp = colon(wp + 3, ep);
start = wp;
mod_fname = mod_fname_buffer;
while(wp < ep)
	{
	switch(*wp)
		{
		default:
			wp++;
			continue;
		case ' ': case '\t': case '+': case '\n': case '\r':
			break;
		}
	break;
	}
mod_flen = wp - start;
if(MAXFCHARS <= mod_flen)
	(*x_error)(MOD2LONG);
g_memcpy_i(mod_fname, start, mod_flen);
return start;
}

typedef UCH *(*LINEFN)(void *,UCH *,UCH *);

static UCH	*line
	(
	void	*t
,	UCH *cp
,	UCH *ep
,	LINEFN lline_fn
	)
	{
	UCP wp;

	/*
	 * N.B. This code needs to know it is going backwards 
	 */
	ep--;
	for(;;)
		{
		if(*ep == FORMFEED)ep--;
		wp = ep;
		wp--;
		while(wp >= cp)
			{
			switch(*wp)
				{
				default: wp--; continue;
				case '\n': case FORMFEED: curr_start_of_line = ++wp;
				}
			break;
			}
		wp = white(wp, ep);
		if(wp != ep)(*lline_fn)(t, wp, ep);
		ep = curr_start_of_line - 1;
		if(ep <= cp)break;
		}
	return cp;
	}


static unsigned	setnmodes(UCH *cp,UCH * ep)	// common part of {io}_setnmodes 
{
unsigned to_date = 0;
unsigned thisline = 0;
unsigned counting = 1;

while(cp < ep)switch(*cp++)
	{
	case '+': counting = 0; break;
	case '\n':
		if(thisline > to_date)to_date = thisline;
		counting = 1;
		thisline = 0;
		break;
	case ':':
		if(counting) thisline = 0;	// consider ,ER: 13 10 * + ,: 
		break;
	case ',': thisline += counting; break;
	}

	return to_date + 1;
}


	/*
	 *	INPUT MAP INTERPRETER
	 */


#define RHOQUADAV	256
#define COMMAND_BASE	-4096
static int base = COMMAND_BASE;
#define GOT_IT(i)		BIT_READ(got_it, (i) - base)
#define ISDIRECT(m, i)		BIT_READ((m)->direct, i)

typedef UCH *IPARSEFN (Imap *,UCH *,UCH *);

#define I_DECLARE(name)		\
	static UCH *name	\
		(		\
		Imap *t		\
	,	UCH *cp		\
	,	UCH *ep		\
		)
/*
 B.N.F.
	<line>		::=<white><lline><newline>
	<lline>		::=,<statement>|
			   <statement>|
			   <nmodes desc>|
			   <modifies desc>|
			   <prog desc>|
			   <user desc>|
			   <label desc>|
			   <utf desc>|
			   <comment>

	<comment>	::=<null>|+<any non vertical spacing>

	<statement>	::=<header><colon><str def><white><comment>
	<white>		::=<null>|<white delim>|<white>
	<white delim>	::=<space>|<tab>|<cr>
	<header>	::=<header id><white><label>
	<header id>	::=-<white>M<number>|
			   M<byte number>|
			   F<byte number>|
			   <line code>|
			   <number><caps lock>
	<str def>	::=<strands>|<strands>*
	<strands>	::=<strand>|<strand><white>,<strands>
	<strand>	::=<white>|
			   <white><byte number>|
			   <white><byte number><white delim><white><strand>
	<number>	::=<dec number>|<hex number>
	<byte number>	::=<number>
	<decnumber>	::=<dec digit><dec digits>
	<dec digits>	::=<null>|<dec digit><dec digits>
	<hexnumber>	::=0x<hex digit><hex digits>
	<hex digit>	::=0|1|...|9|a|...|f|A|...|F
	<hex digits>	::=<null>|<hex digit><hex digits>
	<colon>		::=<white>:<white>
	<caps_lock>	::=<null>|<white>-<white><number>
	<label>		::=<null>|=<ascii string>
	<ascii string>	::=<null>|<ascii char><ascii string>
	<ascii char>	::=<alphabetic>|<symbolic>|<digit>|<space>

	<modifies desc>	::=MOD<colon><filename><white><comment>
	<filename>	::=<fullpathname>|<aplkeys relative pathname>

	<prog desc>	::=PROG<colon><number><white><comment>
	<user desc>	::=USER<colon><number><white><comment>
	<utf desc>	::=UTF<colon><utf number>
	<utf number>	::=8|16
	<line code>	::=
				WI|SI|QT|ER|TB|BT|EP|
				UC|DC|RC|LC|
				US|DS|RS|LS|
				UL|DL|LL|RL|
				HO|CT|PT|
				IN|II|DI|DP|DB|RD|
				TG|DK|OP|CP|MV|
				FD|BK|BP|
				ZM|
				SC|RP|NX|PV|RT|RA
				ED|TC|
				NB|NS|ST|EN|IF|
				HK|FX|LN|
				MC|MR|JP|
				HT|TH|
if XAPL, also			D1|D2|D3|D4|D5|U1|U2|U3|U4|U5|
if GUI also			Ll|Rl|LW|RW|Lw|Rw|Ul|Dl

	not expanded
		<fullpathname>
		<aplkeys relative pathname>
		<null>
		<any non vertical spacing>
		<space>
		<tab>
		<digit>		N.B. 0-9
		<cr>		N.B. not newline
		<alphabetic>	N.B. a-z A-Z
		<symbolic>	N.B. -=`[];'\,./!@#$%^&*()_+~{}"|<>? and space
			i.e. 7-bit ASCII printable less alphanumerics & ':'
 */

static IPARSEFN
	i_header	// see BNF 
,	i_header_id	// see BNF 
,	i_index		// fix up the char_or_index array 
,	i_label		// see BNF 
,	i_lline		// see BNF 
,	i_prog_desc	// see BNF 
,	i_setnmodes	// find how many modes have been used	
,	i_statement	// see BNF 
,	i_str_def	// see BNF 
,	i_strand	// see BNF this does the difficult bit 
,	i_user_desc	// see BNF 
,	i_caps_lock	// see BNF 
,	i_utf		// see BNF 
;

int		 getlinecode(UCH *);	// quadSR calls this for str2num 

static unsigned curr_mode;
static int curr_res;
static UCH *got_it;	// global pointer to local array 
				/* flags to ensure that the latest version
				 * is kept rather than using a modified 
				 * version
				 */
static unsigned lab_len;	// total space required for label info	 
static unsigned get_labs;	// are we filling in labels ? 
static unsigned get_mod_files;	// are we getting MOD files vs compiling 

static void up_cols(Imap *t, Mmap *mode, unsigned count)
{
// increase the number of columns in the strings matrix by count 
/*

	check  for buffer overrun here ---------------------------------+
									|
									v
	-----------------------------------------------------------------
	|strs			|result					|
	-----------------------------------------------------------------
		^		^
		|		|
		|		|
		|		+ preserve alignment here
		|
		+ insert count new bytes at row boundaries

	-----------------------------------------------------------------
	|	region 0	| region 1				|
	-----------------------------------------------------------------

region 0:
	each row moves count*row bytes
region 1:
	moves count*rows + align
 */

unsigned r = mode->rows;
unsigned c = mode->cols;
UCP addr;
unsigned move;
unsigned move_with_align;

addr = (UCH *)&mode->result[r];

move = r * count;
move_with_align = (UPTR)w_align((mode->strs + r*(c + count)), t) - (UPTR)mode->result;

if(addr + move_with_align >= currbuflim)t_abort();

copy_up(addr + move_with_align, addr, r*sizeof(result_t));

mode->result = (result_t *)((UCH *)mode->result + move_with_align);

addr = mode->strs + r*mode->cols;

for(; move -= count, r--; addr -= c)copy_up(addr + move, addr, c);
mode->cols += count;
}

static void up_rows(Imap *t, Mmap *mode, unsigned offset)
{
/*
 * increase the number of rows in the strings matrix by 1; inserting the
 * new one before the offset one
 *
 */


/*
 * we need to create a space in mode->result for the new row
 * we need to ensure the alignment of mode->result[]
 * we need to move all rows in strs with offsets >= offset by mode->cols
 * we need to update mode->rows and mode->result
 */

/*

	check  for buffer overrun here ---------------------------------+
									|
									v
	-----------------------------------------------------------------
	|strs			|result					|
	-----------------------------------------------------------------
		^		^		^
		|		|		|
		|		|		+ insert a new result here
		|		|
		|		+ preserve alignment here
		|
		+ insert a new row here

	-----------------------------------------------------------------
	|	|  region 0	| region 1	| region 2		|
	-----------------------------------------------------------------

region 0:
	moves cols
region 1:
	moves cols + align to 4 byte boundary relative to t
region 2:
	same as region 1 + sizeof(*mode->result)
 */
unsigned move[3];
UCP addr[3];
unsigned size[3];

move[0] = mode->cols;
move[1] = (IPTR)w_align((mode->strs + mode->cols*(mode->rows + 1)), t) - (IPTR)mode->result;

move[2] = move[1] + sizeof(result_t);

addr[0] = mode->strs + mode->rows*mode->cols;
addr[1] = (UCH *)&mode->result[offset];
addr[2] = (UCH *)&mode->result[mode->rows];

size[0] = (mode->rows - offset)*mode->cols;
size[1] = offset * sizeof(result_t);
size[2] = (mode->rows - offset)*sizeof(result_t);

if(addr[2] + move[2] > currbuflim)t_abort();

copy_up(addr[2] + move[2], addr[2], size[2]);
copy_up(addr[1] + move[1], addr[1], size[1]);
copy_up(addr[0] + move[0], addr[0], size[0]);

mode->result = (result_t *)((UCH *)mode->result + move[1]);
if
	(
	0 >= ++mode->rows		// overflow the count
||	mode->rows != (COI)mode->rows	// too big to use as index
	)
	(*x_error)(STRING2MANY);
}

static void copy_up(UCH *to, UCH *from, unsigned n)
{
/*
 * straight forward copy routine that performs its operation backwards
 * rather that forwards
 * N.B.
 * The addresses given are 1 byte higher than the address to be moved
 * Thus if the object being moved is actually an array of doubles then
 * the address of the next double can be coerced and given to copy_up()
 */
if(to != from)while(n--)*--to = *--from;
}

static void TrimDosFile(UCH *cp)
{
#if	DOSWIN	// DOS editors add rubbish to end of file 
	for(;;)
		{
		switch(cp[trans_size-1])
			{
			case 26:	// control Z 
			case ' ':
				trans_size--;
				continue;
			default:
				break;
			}
		break;
		}
#endif//DOSWIN
}

UCH *fillmod
	(
	UCH *workbuf	// start of buffer 
,	UCH * ep	// 1 UCH off the top of buffer 
	)
{
UCH *cp, *wp;
unsigned readfile = 1;

cp = ep;
get_mod_files = 1;	// affects error messages		     
while(readfile)
	{
	cp -= trans_size;
	wp = cp--;

	if(cp < workbuf)
		prestarterr(FILE2BIG);

	*cp = FORMFEED;	// mark file start points with Form Feed 
	curr_start_of_line = wp;

	if(trans_size != read(trans_file, wp, trans_size))
		(*x_error)(CANT_READ);
	close(trans_file);

	TrimDosFile(wp);

	if(cp[trans_size] != '\n')(*x_error)(NOT_ASCII);

	/*
	 * look for MOD: in area just read
	 */
	if(readfile = wp != mod_line(wp, &wp[trans_size]))	// got one ? 
		{
		gchar save;
		gchar envbuf[MAXFCHARS];

		save = mod_fname[mod_flen];
		mod_fname[mod_flen] = 0;	// put null trailing byte 
		(*fname)(mod_fname, envbuf, 0);
		mod_fname[mod_flen] = save;
		if(-1 == t_fopen(envbuf, GSTR(" from ")))(*x_error)(NULLERROR);
		}
	}
get_mod_files = 0;	// affects error messages		     
return cp;
}

static UCH workbuf[T_WKBFSZ];

static unsigned i_any_get(Imap *t)
{
UCH
	*ep = &workbuf[sizeof(workbuf)],
	*cp = ep,
	*startbuf;
UCH l_got_it		// 1 bit for each possible thing
	[
		(
#if	HAS_UNICODE
		(1 << 21)
#else //HAS_UNICODE
		RHOQUADAV
#endif//HAS_UNICODE
	-	COMMAND_BASE
		)
	/	8
	];
unsigned ret;

got_it = l_got_it;	// global reference for sub functions to use 

startbuf = fillmod(workbuf, &workbuf[sizeof(workbuf)]);
curr_start_of_line = startbuf + 1;
i_setnmodes(t, startbuf + 1, ep);
cp = (UCH * )t;
cp += sizeof(Imap);	// point to start of dynamic area	

get_labs = 0;			// not filling in labels		

for(curr_mode = 0; mod_fname = 0, curr_mode < t->n_modes ; curr_mode++)
	{
	Mmap *mp = (Mmap *)cp;

	t->mode[curr_mode] = mp;
	if(mp->strs >= currbuflim)t_abort();
	mp->rows = 1;
	mp->cols = 1;
	mp->result = (result_t *)
		(
		cp +			// start of structure 
		sizeof(Mmap) -		// size of fixed structure 
		sizeof(void *) +	// allocated dynamic size 
		UNSROUND(1)		// genuine size 
		);
	*mp->strs = 0;

	lab_len = 1;			// zap label info size count 

	// null all the got_it flags 

	initbytes(l_got_it, sizeof(l_got_it), 0);
	initbytes(mp->direct, sizeof(mp->direct), ~0);
	initbytes(mp->char_or_index, sizeof(mp->char_or_index), 0);

	mp->doistrip = 1;
	mp->istemp = 0;
	mp->isatty = is_user_io(test_file);

	cp = line(t, startbuf, ep, (LINEFN)i_lline); // tramp the buf for this mode 

	cp = i_index(t, cp, ep);	// fix the indx bytes 

	cp = (UCH *)(mp->result + mp->rows);
	cp = w_align(cp, t);
	if(mp->utfwidth = t->utfwidth)	// copy it into mode
		mp->doistrip = 0;	// utf forces it off
	}

// now make final pass to gather labels up 

if(cp + lab_len*sizeof(achar) > currbuflim)
	t_abort();

t->labinfo = (achar *)cp;

get_labs = 1;			// set flag for i_label 

initbytes(l_got_it, sizeof(l_got_it), 0);

line(t, startbuf, ep, (LINEFN)i_lline);	// go get 'em 

*t->labinfo++ = 0;		// null-terminate whole thing	

ret = ((UCH *)t->labinfo - (UCH *)t);

t->labinfo = (achar *)cp;		// 'n' point back to start	

in_fd[test_file].map = t;
in_fd[test_file].curr_mode[0] = t->prog_mode;
in_fd[test_file].curr_mode[1] = t->user_mode;
return ret;
}

static void io_error(gchar *mess)
{
gchar filename[MAXFCHARS];
gchar modname[MAXFCHARS];
UCP cp;
unsigned linecount;
gchar* mod = GSTR("");
gchar buf[1024];

modname[0] = 0;

(*fname)(0, filename, 1);
// go to end of this input file 

cp = curr_start_of_line;

for(linecount = 1;;)
	{
	if(*--cp == FORMFEED)break;
	else
		if(*cp == '\n')linecount++;
	}
if(!get_mod_files && mod_fname)
	{
	mod = GSTR(" MOD:");
	mod_fname[mod_flen] = 0;
	(*fname)(mod_fname, modname, 1);
	}
gsprintf
	(
	buf
,	gsizeof(buf)
,	GSTR("Stream %u: %ls%ls%ls line : %u %ls")
,	test_file
,	filename
,	mod
,	modname
,	linecount
,	mess
	);

#if	GUI && defined(WINDOWS3)

	trans_message_box(buf);

#else //GUI && defined(WINDOWS3)
	
	{
	size_t ignored;

	ignored = write(2, buf, gstrsize(buf));
	}

#endif//GUI && defined(WINDOWS3)

exit(1);
}

I_DECLARE(i_setnmodes)
{
t->n_modes = setnmodes(cp, ep);
return 0;
}

I_DECLARE(i_lline)
{
UCP wp = cp;

if('+' == *cp)cp = comment(cp, ep);
if(wp  ==  cp)cp = i_prog_desc(t, cp, ep);
if(wp  ==  cp)cp = i_user_desc(t, cp, ep);
if(wp  ==  cp)cp = modifies_desc(cp, ep);
if(wp  ==  cp)cp = i_utf(t, cp, ep);
if(wp  ==  cp)cp = i_statement(t, cp, ep);
return cp;
}


I_DECLARE(i_statement)
{
unsigned gotcomma = *cp == ',';

cp += gotcomma;
cp = i_header(t, cp, ep);
if(GOT_IT(curr_res))return ep;	// already processed this result 
cp = colon(cp, ep);
cp = i_str_def(t, cp, ep);
cp = white(cp, ep);
cp = comment(cp, ep);
if(!gotcomma)BIT_FLIP(got_it, curr_res - base);
return cp;
}

/*	getlinecode
	Does much the same job as i_header: parsing a line code.

	This is done for quadsr in parsing its left argument.
*/

int	getlinecode(UCH *cp)
{
	UCH	*newcp;

	x_error = quadsr_lc_err;	// don't want io_error from quadSR! 

	return lc_str2num(cp, &newcp);
}

/*	getsafelinecode
	Does much the same job as getlinecode parsing a line code.

	This is done for stuff that just needs to know if a code exists for
	the given string
*/

static void	dummy_err(int code)
{
}

int	getsafelinecode(UCH *cp)
{
	UCH	*newcp;
	void	(*old_x_error)(int);
	int ret;

	old_x_error = x_error;
	x_error = dummy_err;	// don't want io_error from quadSR! 

	ret = lc_str2num(cp, &newcp);

	x_error = old_x_error;
	return ret;
}


//	see inlab for a clue as to how labels are stored	

I_DECLARE(i_header)
{
	UCH	*hp, *wp, *lp;

	hp = i_header_id(t, cp, ep);

	wp = white(hp, ep);

	lp = i_label(t, wp, ep);		// new style labels	

	if(lp != wp)				// a label of sorts	
		{
		unsigned		llen;
		UCH	*sl, *el;

		sl = wp+1;			// first char of label	
		el = lp-1;			// last  char of label	

		while(sl <= el && ' ' == *sl)	// skip leading  spaces	
			++sl;

		while(sl <= el && ' ' == *el)	// skip trailing spaces	
			--el;

		llen = 1 + el - sl;		// trimmed label length	

		if(llen)			// a non-null label	
			{
			unsigned	nlen;

			while(!i_alphabetic(*cp) && !i_digit(*cp, 0))
				++cp;		// skip '-' and ','	     

			nlen = 1 + (hp-1) - cp;	// trimmed name length	     

			if(get_labs)		// installing labels ...     
				{
				achar	*heap;

				heap = t->labinfo;

				while(nlen--)
					*heap++ = unixtoav(*cp++);

				*heap++ = '\0';

				while(llen--)
					*heap++ = unixtoav(*sl++);

				*heap++ = '\0';

				t->labinfo = heap;
				}
			else			// just counting size ...    
				lab_len += 2 + llen + nlen;
			}
		}

	return lp;
}

I_DECLARE(i_header_id)
{
	UCH	*newcp;

	curr_res = lc_str2num(cp, &newcp);	// do || die 

	if(curr_res > 0)	// was a number 
		newcp = i_caps_lock(t, newcp, ep);

	if(curr_res <= M0)	// do tricky things if mode string parsed 
		if(!GOT_IT(curr_res) && curr_mode == (M0 - curr_res))
			t->mode[curr_mode]->istemp = (cp[0] == '-');

	return newcp;
}

I_DECLARE(i_caps_lock)
{
	UCH * oldcp = cp;

	cp = white(cp, ep);
	if('-' != *cp++) return oldcp;
	cp = white(cp, ep);
	cp = number(cp, ep);
	if(4 >= lastnumber)(*x_error)(TC_CAPS);
	if(!GOT_IT(curr_res))
		{
		int pair = t->caps_xchg[curr_res];

		if(pair)	// previous pass for another mode 
			{
			if(BIT_READ(t->is_lower, pair))	// or user error 
				(*x_error)(BOTH_LOWER);
			}
		else
			{
			t->caps_xchg[curr_res] = lastnumber;
			t->caps_xchg[lastnumber] = curr_res;
			BIT_FLIP(t->is_lower, curr_res);
			}
		}
	return cp;
}

IOD_DECLARE(utf)
{
	UCH *sp = cp;
	UCH *wp;

	if
		(
		0 != memcmp(cp, "UTF", 3)
		)
		return cp;
	cp += 3;
	wp = cp;
	cp = colon(cp, ep);
	if(wp == cp)
		return sp;
	cp = number(cp, ep);
	switch(lastnumber)
		{
	default:
		(*x_error)(UTF_CODE);
		break;
	case 8: case 16:
		break;
		}
	return cp;
}

I_DECLARE(i_utf)
{
	UCP wp = cp;

	cp = utf(cp, ep);
	if(wp != cp)
		t->utfwidth = lastnumber;
	return cp;
}

I_DECLARE(i_label)			// new style labels		
{
	cp = white(cp, ep);

	if(*cp != '=')			// not a label introducer	
		return cp;		// no label			

	while(cp < ep && i_ascii_char(*++cp))
		;

	return cp;
}


/*	lc_str2num
	Parse a line code. cp points to the first character of the candidate
	line code.

	If cp is a bad line code, an error function is called.

	The numeric code for the string is returned, and the post-line code
	position of cp stored indirect thru the ptr param.
*/

#define	NILt	(Imap *)0   // Pass an ignored input translate table 
#define	NILep	(UCH *)0 // Pass an ignored end of line pointer   

static int lc_str2num(UCH *cp,UCH ** cprslt)
{
	int		lo;
	int		hi;
	int		ix;
	int		code;
	UCH	*str;
	s2n_map *tabptr;

	switch(*cp)
		{
		case '0': case '1': case '2': case '3': case '4':
		case '5': case '6': case '7': case '8': case '9':
			cp =
#if	HAS_UNICODE
				number
#else //HAS_UNICODE
				byte_number
#endif//HAS_UNICODE
					(cp, NILep);
			code = lastnumber;

			break;

		case '-':
			cp = white(cp+1, NILep);

			if('M' != *cp) // only mode lines can begin with - 
				(*x_error)(MINUS_NO_M);

			// ... else drop thru to return Mode code ... 

		case 'M':
			if (i_digit(cp[1], 0))
			{
				cp = byte_number(cp+1, NILep);

				code = M0 - lastnumber;

				break;
			}

		case 'F':
			if (i_digit(cp[1], 0))
			{
				cp = byte_number(cp+1, NILep);

				code = F0 - lastnumber;

				break;
			}

		default: // Not special? search thru 2-character codes ... 
			 // ... table sorted so binary search is best ...  

			tabptr = i_line_codes;

			lo = 0;
			hi = NUM_LINE_CODES - 1;

			while(lo <= hi)
				{
				int	rslt;

				ix = (lo + hi) / 2;

				str = tabptr[ix].string;

				if(str[0] != cp[0])
					rslt = (int)str[0] - (int)cp[0];
				else
					rslt = (int)str[1] - (int)cp[1];

				if(rslt < 0)		// too soon in table 
					lo = ix + 1;
				else if(rslt > 0)	// too late in table 
					hi = ix - 1;
				else
					break;
				}
			
			if(hi < lo)
				(*x_error)(UNKNOWN_LC);

			code = tabptr[ix].code;	// found	
			cp += 2;	// skip 2-char line code 

			break;
		}

	*cprslt = cp;
	return code;
}

/*	lc_num2str
	convert a special function code (line code) into a string.

	At the time of writing, no-one but quadSR uses this fn, so I know
	that code DEFINITELY identifies a key, as returned from inkey().

	The pointer is to a bit of static data which is overwritten each call.
*/

achar *lc_num2str(int code)
{
	static achar abuf[5]; // room for null-terminated "F255" 
	achar *aptr = abuf;
	UCH cbuf[5],*cptr=cbuf;

	cptr = cbuf;

	if(code <= M0)		// It's a mode setting key		
		{
		*cptr++ = 'M';

		cptr = fmtnum(M0 - code, cptr);
		}
	else if(code <= F0)	// It's a function key			
		{
		*cptr++ = 'F';

		cptr = fmtnum(F0 - code, cptr);
		}
	else if(code < 0)	// It's a plain-jane 2-char line code	
		{
		s2n_map	*tabptr;

		for(tabptr = i_line_codes;
		    tabptr < &i_line_codes[NUM_LINE_CODES];
		    ++tabptr
		   )
			if(code == tabptr->code)
				break;

		if(tabptr < &i_line_codes[NUM_LINE_CODES])
			{
			*cptr++ = unixtoav(tabptr->string[0]);
			*cptr++ = unixtoav(tabptr->string[1]);
			}
		}
	else			// It's a value from quadAV		
		{
		*aptr++ = (achar)code;
		*aptr++ = 0;
		}

	if (code < 0)
		{
		*cptr = (char)0;	// null-terminate string 
		a_strcpy_i(abuf, cbuf);
		}

	return abuf;
}

/*	fmtnum
	Format a small number (0-255) into the char buf cptr, and return ptr
	to next char.
*/

static UCH	*fmtnum(int num, UCH * cptr)
{
	int	was;
	int	div;

	was = num;

	if(was >= 100)
		{
		div = num / 100;
		*cptr++ = '0' + div;
		num -= 100 * div;
		}

	if(was >= 10)
		{
		div = num / 10;
		*cptr++ = '0' + div;
		num -= 10 * div;
		}

	*cptr++ = '0' + num;

	return cptr;
}

/*	END OF HEADER PARSING:	LINE CODES.
 *	THAT CODE IS USED BY I_INIT AND QUADSR.
 */
//***************************************************************************

I_DECLARE(i_str_def)
{
unsigned i;
UCP wp = cp;
UCP last = cp;

for(i = 0; i < t->n_modes; i++)
	{
	if(i == curr_mode)wp = i_strand(t, wp, ep);
	else		// skip strings not for this mode 
		for
			(
		;	*wp==' '
		||	*wp=='\t'
		||	*wp=='x'
		||	i_hexdigit(*wp, 0)	// decimal will pass this
		;	++wp
			);
	last = wp > last ? wp : last;
	switch(*wp)
		{
		case ',':
			wp++;
			break;
		case '*':
			last += wp == last; wp = cp;
			break;
		default:
			return last;
		}
	}
return last;
}

void assertion_failure(void)
	{
	return;
	}

I_DECLARE(i_strand)
{
UCP wp;
unsigned len, r;
unsigned firstnumber;
COI *c_or_ip;
Mmap *mp = t->mode[curr_mode];
unsigned dir;
UCH string[32];

// get the string into string[] 

wp = string;
for(len = 0;;len++)
	{
	cp = white(cp, ep);
	if(!i_digit(*cp, 0)) break;
	if(len == 30)(*x_error)(STRING2BIG);
	cp = number(cp, ep);
	if(lastnumber > 127)mp->doistrip = 0;
	if(lastnumber == 0)(*x_error)(EMBEDDED_NULL);
	*wp++ = lastnumber;
	}
if(len == 0)
	{
	return cp;	// no string 
	}
*wp = 0;
firstnumber = *string;
#ifdef	INTERRUPTS
if(curr_res == WI)
	{
	if(len != 1)(*x_error)(WEAK2BIG);
	if(curr_mode == 0)t->weak_int = firstnumber;
	else if(firstnumber != t->weak_int)
		(*x_error)(WEAK_MODES);
	}
if(curr_res == SI)
	{
	if(len != 1)(*x_error)(STRONG2BIG);
	if(curr_mode == 0)t->strong_int = firstnumber;
	else if(firstnumber != t->strong_int)
		(*x_error)(STRONG_MODES);
	}
#endif	// INTERRUPTS 

dir = ISDIRECT(mp, firstnumber);
c_or_ip = &t->mode[curr_mode]->char_or_index[firstnumber];
if
	(
	len == 1
&&	curr_res >= 0
#if	HAS_UNICODE
&&	curr_res <= UINT16_MAX 
#endif//HAS_UNICODE
&&	dir
	)
	{
	if(!*c_or_ip)
		{
		// simple case - maps directly 
		// new entry - no problem 
		
		*c_or_ip = curr_res;
		return cp;
		}
	else
		{
		/*
		 * complex case - either
		 * 	already indirect or
		 *	currently direct and needs to be indirect
		 *
		 * in each of the above it is like a multi byte string
		 * so fall through
		 */
		}
	}
// need to change strs matrix 
if(mp->cols <= len)up_cols(t, mp, 1 + len - mp->cols);
if(dir)
	{
	/*
	 * Find the entry which starts with a number bigger than 
	 * firstnumber or the entry that starts with 0
	 */

	for(r=0,wp=mp->strs; *wp && *wp <= firstnumber; r++, wp += mp->cols);
	if(*c_or_ip)
		{
		/*
		 * we need to take the current direct entry and insert it into
		 * the matrix. The matrix will already be wide enough and the
		 * entry will be the only one starting with firstnumber.
		 */

		if(len == 1) (*x_error)(DUP_STRING);	// bug 701 
		up_rows(t, mp, r);
		
		wp = mp->strs + r*mp->cols;
		*wp++ = firstnumber; *wp = 0;
		mp->result[r] = *c_or_ip;
		*c_or_ip = 0;
		}
	BIT_FLIP(mp->direct, firstnumber);
	}
else
	{
	for(r = 0;;r++)
		{
		size_t sl;
		UCH *tp;

		/*
		 * matrix is kept sorted by string within string length within
		 * first character
		 */

		wp = mp->strs + r*mp->cols;
		if(!*wp)break;
		if(firstnumber < *wp)break;
		if(firstnumber > *wp)continue;
		sl = strlen(wp);
		if(sl < len)break;
		if(sl > len)continue;
		// sl == len 
		if(0 == strcmp(wp, string))(*x_error)(DUP_STRING);
		// cant rely on sign of strcmp - we are using UCHs 
		for(tp = string; *wp == *tp; wp++, tp++);
		if(*wp > *tp) break;
		}
	}
// r now points to the row we wish to insert in front of 
up_rows(t, mp, r);
strcpy(mp->strs + r*mp->cols, string);
mp->result[r] = curr_res;
return cp;
}

I_DECLARE(i_index)
{
unsigned i,j,r,c;
UCP wp;
COI *c_or_ip;
Mmap *mp;

mp = t->mode[curr_mode];
r = mp->rows;
c = mp->cols;
wp = mp->strs + r*c;
c_or_ip = mp->char_or_index;

for(i=r; wp -= c, i--;)
	{
	c_or_ip[*wp] = i;
	for(j = 0; j<c ; j++)if(!wp[j])break;
	for(;j<c; j++)wp[j] = 0;
	}
return cp;
}

static UCH *i_pu_desc(
UCH n_modes,
UCH *cp, 
UCH *ep,
UCH *modep,			// &t->prog_mode or &t->user_mode 
UCH *pustr,			// "PROG" or "USER"		
int		puval,			//  PROG  or  USER		
int		puerr 			// PROG_MODE or USER_MODE	
)
{
if(0 != memcmp(cp, pustr, 4)) return cp;
if(GOT_IT(puval))return ep;
cp = colon(cp + 4, ep);
cp = number(cp, ep);
*modep = lastnumber;
if(lastnumber >= n_modes)(*x_error)(puerr);
cp = white(cp, ep);
cp = comment(cp, ep);
BIT_FLIP(got_it, puval - base);
return cp;
}

I_DECLARE(i_prog_desc)
{
return i_pu_desc(t->n_modes, cp, ep, &t->prog_mode, "PROG", PROG, PROG_MODE);
}

I_DECLARE(i_user_desc)
{
return i_pu_desc(t->n_modes, cp, ep, &t->user_mode, "USER", USER, USER_MODE);
}

	/*
	 *	OUTPUT MAP INTERPRETER
	 */


/*
 B.N.F.
	<line>		::=<white><lline><newline>
	<lline>		::=<statement>|
				<nmodes desc>|
				<modifies desc>|
				<comment>|
				<utf desc
	<comment>	::=<null>|+<anything>

	<statement>	::=<header><colon><str def><white><comment>|
			   <n_header><colon><strand>
	<white>		::=<null>|<white delim>|<white>
	<white delim>	::=<space>|<tab>|<cr>
	<header>	::=M<white><number>|<number>|CC
	<n_header>	::=INIT|
			   RESET|
			   V<white><byte number>|
			   C<white><byte number>|
			   B<white><byte number>
	<str def>	::=<strands>|<strands>*
	<strands>	::=<strand>|<strand><white>,<strands>
	<strand>	::=<white>|
			   <white><byte number>|
			   <white><byte number><white delim><white><strand>
	<number>	::=<dec number>|<hex number>
	<byte number>	::=<number>
	<decnumber>	::=<dec digit><dec digits>
	<dec digits>	::=<null>|<dec digit><dec digits>
	<hexnumber>	::=0x<hex digit><hex digits>
	<hex digit>	::=0|1|...|9|a|...|f|A|...|F
	<hex digits>	::=<null>|<hex digit><hex digits>
	<colon>		::=<white><colon><white>

	<modifies desc>	::=MOD<colon><filename><white><comment>
	<filename>	::=<fullpathname>|<apltrans relative pathname>
	<utf desc>	::=UTF<colon><utf number>
	<utf number>	::=8|16

	not expanded
		<fullpathname>
		<apltrans relative pathname>
		<null>
		<anything>
		<space>
		<tab>
		<cr>	N.B. not newline
 */

typedef UCH *OPARSEFN
	(
	struct Translate *t
,	UCH *cp
,	UCH *ep
	);

#define DECLARE(name)			\
	static UCH *name		\
		(			\
		struct Translate *t	\
	,	UCH *cp			\
	,	UCH *ep			\
		)			\

static OPARSEFN
	lline
,	statement
,	header
,	n_header
,	str_def
,	strand
,	utf_desc
;

#define NMODED	512		// allows for 256 modes + CC + lots of spare 
static int currline;		// current line number in translate output 
static int currcode;		// current code of non-moded string 

DECLARE(o_setnmodes)
{
t->nmodes = setnmodes(cp, ep);
t->nmodes += 0 == t->nmodes;
return 0;
}

DECLARE(lline)
{
UCP wp = cp;

if('+' == *cp)cp = comment(cp, ep);
if(wp  ==  cp)cp = modifies_desc(cp, ep);
if(wp  ==  cp)cp = utf_desc(t, cp, ep);
if(wp  ==  cp)cp = statement(t, cp, ep);
return cp;
}

#if	HAS_UNICODE
static unsigned value2line(struct Translate *t, achar value)
	{
	unsigned i;

	for
		(
		i = 0
	;	i < t->nvalues
	&&	value != t->values[i]
	;	++i
		);
	return i;
	}
#endif//HAS_UNICODE

DECLARE(statement)
{
UCP wp = cp;

wp = header(t, wp, ep);
if(wp != cp)		// moded statement 
	{
	if(GOT_IT(currline))return ep;
	BIT_FLIP(got_it, currline - base);
	wp = colon(wp, ep);
	wp = str_def(t, wp, ep);
	wp = white(wp, ep);
	wp = comment(wp, ep);
	return wp;
	}
wp = n_header(t, wp, ep);
if(wp != cp)		// non_moded statement 
	{
	UCH *start = currstr;

	if(GOT_IT(currcode)) return ep;
	BIT_FLIP(got_it, currcode - base);
	t->special--;	// allocate space strand() tests for overflow 
	t->nsas++;
	wp = colon(wp, ep);
	wp = strand(t, wp, ep);
	t->special->code = currcode;
	t->special->offset = start - t->strstart;
	t->special->len = currstr - start;
	wp = comment(wp, ep);
	return wp;
	}
wp = comment(wp, ep);	// will force error if neither case 
return wp;
}

DECLARE(header)
{
switch(*cp)
	{
	case 'M':
		cp++; cp = white(cp, ep); cp = byte_number(cp, ep);
		currline = MODE0 + lastnumber;
		break;
	case 'C':
		if('C' == cp[1])
			{
			currline = CC;
			cp += 2;
			}
		break;
	case '0': case '1': case '2': case '3': case '4':
	case '5': case '6': case '7': case '8': case '9':
		cp = number(cp, ep);
#if	HAS_UNICODE
		currline = value2line(t, lastnumber);
		if(MAX_NVALUES <= currline)
			(*x_error)(TOOMANYVAL);
		t->nvalues += currline == t->nvalues;	// new value
		t->values[currline] = lastnumber;
#else //HAS_UNICODE
		currline = lastnumber;
#endif//HAS_UNICODE
		break;
	}
return cp;
}

DECLARE(n_header)
{
switch(*cp)
	{
	case 'I':
		if(0 == memcmp(cp, "INIT", 4))
			{
			cp += 4;
			currcode = INIT;
			}
		break;
	case 'R':
		if(0 == memcmp(cp, "RESET", 5))
			{
			cp += 5;
			currcode = RESET;
			}
		break;
	case 'V':
		cp++;
		cp = white(cp, ep);
		cp = byte_number(cp, ep);
		currcode = V0 - lastnumber;
		break;
	case 'C':
		cp++;
		cp = white(cp, ep);
		cp = byte_number(cp, ep);
		currcode = C0 - lastnumber;
		break;
	case 'B':
		cp++;
		cp = white(cp, ep);
		cp = byte_number(cp, ep);
		currcode = B0 - lastnumber;
		break;
	}
return cp;
}

static void testbuf(struct Translate *t,UCH *addr)	// test for overrun of allocated buffer 
{
if(addr >= (UCH *)t->special) t_abort();
}

DECLARE(str_def)
{
UCP wp = cp;
UCP start;
UCP lp;
unsigned i = t->nmodes;
UCP last = cp;

lp = t->lengths + i*currline;
testbuf(t, (UCH *)(t->offset + currline));
t->offset[currline] = currstr -  t->strstart;
while(i--)
	{
	start = currstr;
	wp = strand(t, wp, ep);
	testbuf(t, (UCH *)lp);
	*lp++ = currstr - start;	// store the length 
	last = wp > last ? wp : last;
	switch(*wp)
		{
		case ',':
			wp++;			break;
		case '*':
			last += wp == last;	wp = cp;	break;
		default:
			return last;
		}
	}
return last;
}

DECLARE(strand)
{
for(;;)
	{
	cp = white(cp, ep);
	if(!i_digit(*cp, 0))return cp;
	cp = number(cp, ep);
	testbuf(t, (UCH *)currstr);
	*currstr++ = lastnumber;	// store it in strings 
	}
}

DECLARE(utf_desc)
{
	UCP wp = cp;

	cp = utf(cp, ep);
	if(wp != cp)
		t->utfwidth = lastnumber;
	return cp;
}

static unsigned o_any_get(struct Translate *t)
{
UCP cp;
UCP ep = &workbuf[sizeof(workbuf)];
UCH l_got_it
	[
		(
#if	HAS_UNICODE
		(1 << 21)
#else //HAS_UNICODE
		RHOQUADAV
#endif//HAS_UNICODE
	+	NMODED
	-	COMMAND_BASE
		)
	/	8
	];

got_it = l_got_it;
cp = fillmod(workbuf, ep);
o_setnmodes(t, cp, ep);

t->offset = (unsigned short *)t->dynamic;
t->lengths = t->dynamic + (MODE0 + t->nmodes)*sizeof(unsigned short);
currstr = t->strstart = t->lengths + (MODE0+t->nmodes)*t->nmodes;
t->special = (struct Sas *)currbuflim;

initbytes(l_got_it, sizeof(l_got_it), 0);
cp = line(t, cp, ep, (LINEFN)lline);
(o_out[test_file]->t_out = t)->refcount = 1;

// move the special strings down in memory
if(1 & (UPTR)currstr) ++currstr;	// get to an even boundary
memmove(currstr, t->special, t->nsas * sizeof(struct Sas));
t->special = (void *)currstr;

// find the INIT string (if any) and put it first 
	{
	unsigned i = t->nsas;

	while(i--)if(t->special[i].code == INIT)
		{
		struct Sas temp;

		temp = *t->special;
		*t->special = t->special[i];
		t->special[i] = temp;
		}
	}
return (UCH *)&t->special[t->nsas] - (UCH *)t;
}




	/*
	 *	DATE MAP INTERPRETER
	 */


/*
 B.N.F.
	<line>		::=<white><lline><newline>
	<lline>		::=<statement>|<comment>
	<comment>	::=<null>|+<anything>
	<statement>	::=<date_defn><white><comment>

	<date_defn>	::=<dm_header><colon><dm_desc>|
			   <fmt_header><fmt_label><colon><fmt_desc>

	<dm_header>	::= D<white>|
			    M<white>
	<dm_desc>	::=<dm_name>|
			   <dm_name><white space><dm_desc>
	<dm_name>	::=<alphabetic>|<alphabetic><alphabetic><alphabetics>

	<fmt_header>	::=F<white><number>
	<fmt_label>	::=<null>|=<ascii string>
	<fmt_desc>	::=<ascii char>|<ascii char><ascii char><ascii string>

	<ascii string>	::=<null>|<ascii char><ascii string>
	<ascii char>	::=<alphabetic>|<symbolic>|<digit>|<space>
	<colon>		::=<white>:<white>
	<alphabetics>	::=<null>|<alphabetic><alphabetics>
	<number>	::=0|1|...|255
	<white>		::=<null>|<white delim><white>
	<white space>	::=<white delim>|<white delim><white space>
	<white delim>	::=<space>|<tab>

	not expanded
		<null>
		<space>
		<tab>
		<digit>
		<alphabetic>	N.B. a-z A-Z
		<symbolic>	N.B. -=`[];'\,./!@#$%^&*()_+~{}"|<>? and space
			i.e. 7-bit ASCII printable less alphanumerics & ':'
 */

/*
	struct Datemap	used to store the local information of day names,
			month names, and formats requested in quadSM.
	int datemaperr	used to indicate whether a valid date map has been
			read for quadSM, and indicates which error if an
			error occured.
*/

static struct Datemap	*datemap;
int			datemaperr;
static jmp_buf		datebuf;
static unsigned		lastdayno;
static unsigned		lastmonthno;

typedef UCH *DPARSEFN
	(
	struct Datemap *t
,	UCH *cp
,	UCH *ep
	);

#define D_DECLARE(name)			\
	static UCH *name		\
		(			\
		struct Datemap *t	\
	,	UCH *cp			\
	,	UCH *ep			\
		)			\


#define BAD_DATE_ACCESS	70
#define BAD_DATE_SYNTAX	71
#define ESC_CHAR	0x5c

static DPARSEFN
	d_lline
,	d_statement
,	d_dm_header
,	d_dm_desc
,	d_dm_name
,	d_fmt_header
,	d_fmt_label
,	d_fmt_desc
;

void	d_error(int);

static void d_testbuf(struct Datemap *t, UCH *addr)
{				// test for overrun of allocated buffer 
if(addr >= (UCH *)t->strings) (*x_error)(BAD_DATE_ACCESS);
}

D_DECLARE(d_lline)
{
UCP wp = cp;

if('+' == *cp)cp = comment(cp, ep);
if(wp  ==  cp)cp = d_statement(t, cp, ep);
return cp;
}

D_DECLARE(d_statement)
{
UCP wp = cp;

wp = d_dm_header(t, wp, ep);
if(wp != cp)
	{
	wp = colon(wp, ep);
	wp = d_dm_desc(t, wp, ep);
	wp = white(wp, ep);
	wp = comment(wp, ep);
	return wp;
	}
wp = d_fmt_header(t, wp, ep);
if(wp != cp)
	{
	if(GOT_IT(currcode)) return ep;
	wp = d_fmt_label(t, wp, ep);
	wp = colon(wp, ep);
	wp = d_fmt_desc(t, wp, ep);
	wp = white(wp, ep);
	wp = comment(wp, ep);
	return wp;
	}
wp = comment(wp, ep);	// will force error if none of these cases 
return wp;
}

D_DECLARE(d_fmt_header)
{
if (*cp == 'F')
	{
	cp++;
	cp = white(cp, ep);
	cp = number(cp, ep);
	t->n_formats++;
	currcode = DF0 - lastnumber;
	}
return cp;
}

D_DECLARE(d_fmt_label)
{
UCP wp;
unsigned len = 0;
UCP start = currstr;
int formatcode = currcode;

	if (*cp == ':')
		return cp;
	if(*cp != '=') (*x_error)(BAD_DATE_SYNTAX);

	wp = ++cp;
	for(;wp<ep && *wp != ':';)
	    {
	    unsigned class = 0;

	    if(*wp == ESC_CHAR) class = 1;
	    else if(*wp >= ' ' && *wp <= '~') class = 2;
	    switch(class)
		{
		case 0: wp++;
			break;
		case 1: len++;
			wp+=4;
			continue;
		case 2: len++;
			wp++;
			continue;
		}
	    }
	
	t->n_labels++;
	currcode = DL0 - lastnumber;
	BIT_FLIP(got_it, currcode - base);
	t->strings--;
	while (len--)
	    {
	    d_testbuf(t, (UCH *)currstr);
	    if (*cp != ESC_CHAR)
		*currstr++ = unixtoav(*cp++);
	    else
		{
		cp = number(cp+1, ep);
		*currstr++ = lastnumber;
		}
	    }
	*currstr++ = 0;
	t->strings->code = currcode;
	t->strings->offset = start - t->strstart;
	t->strings->len = currstr - 1 - start;
	currcode = formatcode;
	return cp;
}

D_DECLARE(d_fmt_desc)
{
UCP wp = cp;
unsigned spaces = 0;
unsigned len = 0;
UCP start = currstr;

	for(;wp<ep && *wp != '+';)
	    {
	    unsigned class = 0;

	    if(*wp == ' ') class = 1;
	    else if(*wp == ESC_CHAR) class = 2;
	    else if(*wp > ' ' && *wp <= '~') class = 3;
	    switch(class)
		{
		case 0: wp++;
			break;
		case 1: spaces++;
			wp++;
			continue;
		case 2: len += spaces;
			spaces = 0;
			len++;
			wp+=4;
			continue;
		case 3: len += spaces;
			spaces = 0;
			len++;
			wp++;
			continue;
		}
	    }
	
	if(!len) (*x_error)(BAD_DATE_SYNTAX);
	BIT_FLIP(got_it, currcode - base);
	t->strings--;
	while (len--)
	    {
	    d_testbuf(t, (UCH *)currstr);
	    if (*cp != ESC_CHAR)
		*currstr++ = unixtoav(*cp++);
	    else
		{
		cp = number(cp+1, ep);
		*currstr++ = lastnumber;
		}
	    }
	*currstr++ = 0;
	t->strings->code = currcode;
	t->strings->offset = start - t->strstart;
	t->strings->len = currstr - 1 - start;
	return cp;
}

D_DECLARE(d_dm_header)
{
switch(*cp)
	{
	case 'D' :
		cp++;
		cp = white(cp, ep);
		currcode = D0 - lastdayno;
		break;
	case 'M' :
		cp++;
		cp = white(cp, ep);
		currcode = MN0 - lastmonthno;
		break;

	}
return cp;
}

D_DECLARE(d_dm_desc)
{
struct Sas	*s = t->strings;
int		currnum;
int	nread, num;

	currnum = currcode;

	for(;cp<ep && *cp != '+';)
		{
		cp = d_dm_name(t, cp, ep);
		cp = white(cp, ep);
		}
	nread = currnum - currcode;

	for (num = 0; num < nread; num++)
		{
		s--;
		if (currcode <= MN0)
			s->code = MN0 - (12 - lastmonthno) - num;
		else
			s->code = D0 - (7 - lastdayno) - num;
		}
return cp;
}

D_DECLARE(d_dm_name)
{
UCP wp = cp;
UCP start = currstr;

unsigned len = 0;

	while( ( i_alphabetic(*wp)	||
		 (*wp == ESC_CHAR)
	       )
		&& wp < ep)
		{
		if (*wp != ESC_CHAR)
			wp++;
		else
			wp+=4;
		len++;
		}

	if(!len) (*x_error)(BAD_DATE_SYNTAX);
	BIT_FLIP(got_it, currcode - base);
	t->strings--;
	while (len--)
	    {
	    d_testbuf(t, (UCH *)currstr);
	    if (*cp != ESC_CHAR)
		*currstr++ = unixtoav(*cp++);
	    else
		{
		cp = number(cp+1, ep);
		*currstr++ = lastnumber;
		}
	    }
	*currstr++ = 0;
	t->strings->code = currcode;
	t->strings->offset = start - t->strstart;
	t->strings->len = currstr - 1 - start;
	if(currcode <= MN0)
		{
		t->n_months++;
		lastmonthno++;
		}
	else
		{
		t->n_days++;
		lastdayno++;
		}
	currcode--;
	return cp;
}

static UNIQUE_FID date_unique_fid;

static void d_fname	// develop a date file name 
	(
	gchar *tp		// null or target file name 
,	gchar *buf		// where to put the file name 
,	unsigned errflag		// an error condition ? 
	)
{
	static gchar envdir[]  = GSTR("APLFMT");
	static gchar dirname[] = GSTR("aplfmt");
	static gchar envname[] = GSTR("APLF  ");
	static gchar ext[]     = GSTR(".DFT");
	gchar tbuf[127];
	gchar *ep, *dp;
	static unsigned skip;

if(!buf)
	{
	// initialise static data 
	skip = 0;
	date_unique_fid = null_unique_fid;
	return;
	}
if(!tp)
	{
	ep = envname + 4;
	if(test_file>9)*ep++ = '0' + test_file/10;
	*ep++ = '0' + test_file%10;
	*ep=0;

	ep = ggetenv(envname);
	if(0 == ep)
		{
		*(envname+4) = 0;	// "APLF" 
		ep = ggetenv(envname);
		if(0 == ep) ep = GSTR("default");
		}
	tp = ep;
	}
gstrcpy(tbuf, tp);
tp = tbuf;
gstrcat(tbuf, gfextend(tbuf, ext));
if(gstrcspn(tp, PUNCTUATION) == gstrlen(tp))	// relative pathname ? 
	{
	int s = 0 != (dp = ggetenv(envdir));
	switch(s)
	    {
	    default:
		{
		GSTAT sbuf;
		unsigned count;

		buf[0] = 0;	 // in case APLFMT exists but is 0
		for(count = 0 ;;skip += count++ == skip, dp += *dp != 0)
		    {
		    gchar *bp = dp;
		    unsigned n;

		    dp += gstrcspn(dp, pathdelim);
		    if(bp == dp)break;
		    if(count == skip)
			{
			gstrncpy(buf, bp, n = dp - bp);
			buf[n] = 0;
			gstrcat(buf, dirdelim);
			gstrcat(buf,tp);
			if
			    (
/* got a file */		-1 != gstat(buf, &sbuf) &&
#ifndef WINCE
/* is a plain file */		(S_ISREG(sbuf.st_mode)) &&
#endif   //WINCE
/* is an error */		errflag
			    )break;
			}
		    }
		date_unique_fid = filename_to_unique_fid(buf);
		}
		break;
	    case 0:		// build the default 
		if(dp = ggetenv(GSTR("DYALOG")))
			{
			gstrcpy(buf, dp);
			gstrcat(buf, dirdelim);
			gstrcat(buf, dirname);
			gstrcat(buf, dirdelim);
			gstrcat(buf,tp);
			}
		else
			buf[0] = 0;	// return empty
		break;
		}
	    }
    else gstrcpy(buf, tp);
}

static unsigned d_any_get(struct Datemap *t);

static void d_filbuf(void)
{
UCP buf = malloc(MAX_BUF);
struct Datemap *t = (struct Datemap *)buf;
struct Datemap *new;
int reallocmoves = checkrealloc();

if(0 == buf) (*x_error)(MALLOC_FAIL);
currbuflim = buf + MAX_BUF;
initbytes(buf, MAX_BUF, 0);
t->unique_fid = date_unique_fid;
currstr = t->strstart = t->dynamic;
t->strings = (struct Sas *)currbuflim;
datemap = t;

if (reallocmoves)
	d_any_get(t);
else
	new = realloc(t, d_any_get(t));
}

static void  LIBCALL d_finish(void)
	{
	if(datemap) free(datemap);
	}

void d_init(void)
{
struct stat statbuf;
gchar envbuf[MAXFCHARS];

if(!x_error) x_error = d_error;
datemaperr = 0;			// no errors to start with		
d_fname(0, 0, 0);		// init statics for top level request	
d_fname(0, envbuf, 0);		// put file name to read in envbuf	

if (
	(-1 == (trans_file = gopen(envbuf, O_RDONLY | O_BINARY))) ||// cant open 
	(-1 == fstat(trans_file, &statbuf))	// no file 
	) 
		datemaperr = BAD_DATE_ACCESS;
else
	{
	trans_size  = statbuf.st_size;
	trans_unique_fid = fd_to_unique_fid(trans_file);

#if	defined(DOSWIN) && !WINCE && !APL_IME
	binary_file(trans_file);
#endif
	if(setjmp(datebuf))
		{
		datemaperr = BAD_DATE_SYNTAX;
		}
	else
		{
		d_filbuf();

		if
			(
			datemap->n_days    !=  7
		||	datemap->n_months  != 12
		||	datemap->n_formats >  10
			)
			datemaperr = BAD_DATE_SYNTAX;
		}
	}
atexit(d_finish);

i_init();	// get the keyboard tables 

#if !APL_IME && !AP_BUILD && HAS_UNICODE && HAS_SKEYS
	ReadKeyboardShortcuts();
#endif

o_init();	// get the output tables
}



static unsigned d_any_get(struct Datemap *t)
{
UCP cp;
UCP ep = &workbuf[sizeof(workbuf)];
UCH l_got_it[(RHOQUADAV + NMODED - COMMAND_BASE)/8];
unsigned extra;

got_it = l_got_it;
cp = fillmod(workbuf, ep);
initbytes(l_got_it, sizeof(l_got_it), 0);
lastdayno = lastmonthno = 0;
cp = line(t, cp, ep, (LINEFN)d_lline);
currstr = w_align(currstr, t);
extra = currbuflim - (UCH *)t->strings;
memmove(currstr, t->strings, extra);
t->strings = (void *)currstr;
currstr += extra; 
return currstr - (UCH *)t;
}


static struct Sas *get_date_s(int code)
	{
	struct Sas *s = datemap->strings;
	unsigned n = datemap->n_days
	+	datemap->n_months
	+	datemap->n_formats
	+	datemap->n_labels
	;

	for(++n, --s; --n && code != (++s)->code;);
	return n ? s : 0;
	}

achar *get_date_string(int code)
	{
	struct Sas *s = get_date_s(code);
	UCP str = s ? datemap->strstart + s->offset : 0;

#if	HAS_UNICODE
	if(str)
		{
#	define	MAX_DATE_LEN	16		// I guess
		static achar result[MAX_DATE_LEN];

		if(MAX_DATE_LEN < s->len)
			nonce();
		a_memcpy_i(result, str, s->len);
		return result;
		}
	return 0;
#else //HAS_UNICODE
	return (achar *)str;
#endif//HAS_UNICODE
	}

short unsigned get_date_length(int code)
	{
	struct Sas *s = get_date_s(code);

	return s ? s->len : 0;
	}

void d_error(int err_no)
{
	datemaperr = (err_no == BAD_DATE_ACCESS)
			? BAD_DATE_ACCESS
			: BAD_DATE_SYNTAX;

	longjmp(datebuf, 1);
}


#ifdef getenv

#undef getenv
#undef ggetenv

extern char* LIBCALL getenv(const char *);

#endif

void * LIBCALL get_ours(const gchar *envname)
{

#define ADD(a,b)	{ GSTR(a), GSTR(b)	}

typedef struct
	{
	gchar * name;
	gchar * value;
	} ENV_DEF;
ENV_DEF our_defaults[] =
#if	GUI && (defined(MAINWIN) || WINE)
	{
		ADD( "APLFSCB_HOST","aplfscb"	),
		ADD( "APLK",	"apl_us.din" 	),
		ADD( "APLNID",	"0" 		),
		ADD( "CONFIRM_SESSION_DELETE", "1"),
		ADD( "DYALOG",	"/usr/mdyalog"	),
		ADD( "LOG_FILE",	"/dev/null" 	),
		ADD( "LOG_SIZE",	"80" 		),
		ADD( "MAPCHARS",	"db0deba7eec00be0" ),
		ADD( "MAXWS",	"4096" 		),
		ADD( "SESSION_FILE", "default.dse" ),
		ADD( "TERM",	"win" 		),
		ADD( "TMP",	"" 		),
		ADD( "WSPATH",	".:/usr/mdyalog/ws:/usr/mdyalog/xflib"),
		   { 0,		0 		}
	};
#elif	GUI && defined(WINCE)
	{
		ADD( "APLNID",	"0" ),
		ADD( "CONFIRM_SESSION_DELETE", "1"),
		ADD( "LOG_FILE",	"nul" ),
		ADD( "LOG_SIZE",	"80" ),
		ADD( "MAPCHARS",	"db0deba7eec00be0" ),
		ADD( "MAXWS",	"4096" ),
		ADD( "TERM",	"WIN" ),
		ADD( "TMP",	"" ),
		ADD( "WSPATH",	"\\Program Files\\Dyalog\\"),
		   { 0,		0 }
	};
#elif	GUI && defined(WINDOWS3)
	{
		ADD( "APLNID",	"0" ),
		ADD( "CONFIRM_SESSION_DELETE", "1"),
		ADD( "LOG_FILE",	"nul" ),
		ADD( "LOG_SIZE",	"80" ),
		ADD( "MAPCHARS",	"db0deba7eec00be0" ),
		ADD( "MAXWS",	"16384" ),
		ADD( "TERM",	"WIN" ),
		ADD( "TMP",	"" ),
		ADD( "WSPATH",	"."),
		   { 0,		0 }
	};
#elif	defined(MAINWIN)	
	{
		ADD( "APLFSCB_HOST",	"aplfscb"),
		ADD( "APLK",	"apl_us" 	),
		ADD( "CONFIRM_SESSION_DELETE", "0"),
		ADD( "DYALOG",	"/usr/mdyalog" ),
		ADD( "LOG_FILE",	"/dev/null" 	),
		ADD( "LOG_SIZE",	"80" 		),
		ADD( "MAPCHARS",	"db0deba7eec00be0" ),
		ADD( "MAXWS",	"4096" ),
		ADD( "TMP",	"" 		),
		ADD( "WSPATH",	".:/usr/mdyalog/ws:/usr/mdyalog/xflib" ),
		   { 0,		0 }
	};
#else
	{
		ADD( "APLFSCB_HOST",	"aplfscb"),
		ADD( "CONFIRM_SESSION_DELETE", "0"),
		ADD( "DYALOG",	"/usr/dyalog" ),
		ADD( "LOG_FILE",	"/dev/null" 	),
		ADD( "LOG_SIZE",	"80" 		),
		ADD( "MAPCHARS",	"db0deba7eec00be0" ),
		ADD( "MAXWS",	"4096" ),
		ADD( "TMP",	"" 		),
		ADD( "WSPATH",	".:/usr/dyalog/ws:/usr/dyalog/xflib" ),
		   { 0,		0 }
	};
#endif	

ENV_DEF *ours;

for(ours = our_defaults; ours->name ; ours++)
	{
	if(0 == gstrcmp(ours->name, envname))return ours->value;
	}
return 0;

}

gchar *map_environment_name(gchar *fm)
{
static gchar mapped[256];

gchar *to=mapped;

for (;*fm;fm++,to++) switch (*fm)
	{
	case '\\':		// Windows registry delimeter
		*to='_';
		break;
	default:
		*to=gtoupper(*fm);
		break;
	}
*to=0;
return mapped;
}


void * (*special_getenv)(const void *);	// hook for special cases 


gchar * LIBCALL w_shad_getenv(const gchar *envname)
{
gchar * res;

if(special_getenv && (res = (*special_getenv)(envname)))
	return res;

#if	defined(WINDOWS3) && !APL_IME
	{

	if(res = W3getenv(envname))	// implements environment, command line,
		return res;		// registry search order
	}
#else //defined(WINDOWS3)

// here it should use a ggetenv() defined in gfns.c that would do the switch...
#if HAS_UNICODE
if(res = _wgetenv(envname))
	return res;
#else
if(res = getenv(envname))
	return res;
#endif

#endif//defined(WINDOWS3)

res = get_ours(envname);
return res;
}

achar *agetenv(const gchar *envname)
	{
	gchar *grslt = w_shad_getenv(envname);
#if	HAS_UNICODE
#define	MAX_ENV_LEN	1024
	static achar agetenv_buf[MAX_ENV_LEN];
	achar *arslt = agetenv_buf;
	size_t len;

	if(0 == grslt)
		return 0;
	len = gstrlen(grslt);
	if(MAX_ENV_LEN <= len)
		limit();
	a_memcpy_g(arslt, grslt, 1 + len);
	return arslt;
#else //HAS_UNICODE
	return grslt;
#endif//HAS_UNICODE
	}

char * LIBCALL shad_getenv(const char *envname)
{
#if HAS_UNICODE

gchar*in =galloca(strlen(envname)+1);
gchar*ip = in;

static char out[256];
char*op = out;

const char*ep = envname;

gchar *vp;

while (*ep)
	*ip++=*ep++;
*ip=0;

vp=w_shad_getenv(in);

if (0==vp) return 0;

while (*vp)
	*op++=(char)*vp++;

return out;

#else
return w_shad_getenv(envname);
#endif
}

// National language friendly 1-1 AV to ANSI mapping	

UCH av_ansi_dflt[] =
{
//	   AV APL-char		     NB	     ->	ANSI	   ANSI-char	     

	/* 00 Null			*/	0x00,	/* ctrl: NUL	     */
	/* 01 Backspace			*/	0x08,	/* ctrl: BS	     */
	/* 02 Line feed			*/	0x0a,	/* ctrl: LF	     */
	/* 03 Carriage return		*/	0x0d,	/* ctrl: CR	     */
	/* 04 Space			*/	0x20,	/* <space>	     */
	/* 05 Form feed			*/	0x0c,	/* ctrl: FF	     */
	/* 06			(spare)	*/	0x06,	/* ctrl: ACK	     */
	/* 07 Bell			*/	0x07,	/* ctrl: BEL	     */
	/* 08 Escape			*/	0x1b,	/* ctrl: ESC	     */
	/* 09 Tab			*/	0x09,	/* ctrl: HT	     */
	/* 0a I-Beam			*/	0x0b,	/* ctrl: VT	     */
	/* 0b Stile bar			*/	0x0e,	/* ctrl: SO	     */
	/* 0c Percent			*/	0x25,	/* % percent sign    */
	/* 0d APL-Quote			*/	0x27,	/* ' apostrophe	     */
	/* 0e Alpha			*/	0xb8,	/* , spacing cedilla */
	/* 0f Omega			*/	0xbe,	/*   three quarters  */

	/* 10 Underscore		*/	0x5f,	/* _ underscore	     */
	/* 11 Lower a			*/	0x61,	/* a small letter    */
	/* 12 Lower b			*/	0x62,	/* b small letter    */
	/* 13 Lower c			*/	0x63,	/* c small letter    */
	/* 14 Lower d			*/	0x64,	/* d small letter    */
	/* 15 Lower e			*/	0x65,	/* e small letter    */
	/* 16 Lower f			*/	0x66,	/* f small letter    */
	/* 17 Lower g			*/	0x67,	/* g small letter    */
	/* 18 Lower h			*/	0x68,	/* h small letter    */
	/* 19 Lower i			*/	0x69,	/* i small letter    */
	/* 1a Lower j			*/	0x6a,	/* j small letter    */
	/* 1b Lower k			*/	0x6b,	/* k small letter    */
	/* 1c Lower l			*/	0x6c,	/* l small letter    */
	/* 1d Lower m			*/	0x6d,	/* m small letter    */
	/* 1e Lower n			*/	0x6e,	/* n small letter    */
	/* 1f Lower o			*/	0x6f,	/* o small letter    */

	/* 20 Lower p			*/	0x70,	/* p small letter    */
	/* 21 Lower q			*/	0x71,	/* q small letter    */
	/* 22 Lower r			*/	0x72,	/* r small letter    */
	/* 23 Lower s			*/	0x73,	/* s small letter    */
	/* 24 Lower t			*/	0x74,	/* t small letter    */
	/* 25 Lower u			*/	0x75,	/* u small letter    */
	/* 26 Lower v			*/	0x76,	/* v small letter    */
	/* 27 Lower w			*/	0x77,	/* w small letter    */
	/* 28 Lower x			*/	0x78,	/* x small letter    */
	/* 29 Lower y			*/	0x79,	/* y small letter    */
	/* 2a Lower z			*/	0x7a,	/* z small letter    */
	/* 2b			(spare)	*/	0x01,	/* ctrl: SOH	     */
	/* 2c			(spare)	*/	0x02,	/* ctrl: STX	     */
	/* 2d High Minus		*/	0xaf,	/*   spacing macron  */
	/* 2e Period			*/	0x2e,	/* . period	     */
	/* 2f Zilde			*/	0xab,	/*   left guillemet  */

	/* 30 Zero			*/	0x30,	/* 0 digit zero	     */
	/* 31 One			*/	0x31,	/* 1 digit one	     */
	/* 32 Two			*/	0x32,	/* 2 digit two	     */
	/* 33 Three			*/	0x33,	/* 3 digit three     */
	/* 34 Four			*/	0x34,	/* 4 digit four	     */
	/* 35 Five			*/	0x35,	/* 5 digit five	     */
	/* 36 Six			*/	0x36,	/* 6 digit six	     */
	/* 37 Seven			*/	0x37,	/* 7 digit seven     */
	/* 38 Eight			*/	0x38,	/* 8 digit eight     */
	/* 39 Nine			*/	0x39,	/* 9 digit nine	     */
	/* 3a			(spare)	*/	0x03,	/* ctrl: ETX	     */
	/* 3b Int'l currency sumbol	*/	0xa4,	/*   currency sign   */
	/* 3c Japanese Yen		*/	0xa5,	/*   yen sign	     */
	/* 3d Dollar			*/	0x24,	/* $ dollar sign     */
	/* 3e Pound Sterling		*/	0xa3,	/*   pound sign	     */
	/* 3f Cent			*/	0xa2,	/*   cent sign	     */

	/* 40 Delta			*/	0x91,	/* ctrl: reserved    */
	/* 41 Upper _A			*/	0x41,	/* A capital letter  */
	/* 42 Upper B			*/	0x42,	/* B capital letter  */
	/* 43 Upper C			*/	0x43,	/* C capital letter  */
	/* 44 Upper D			*/	0x44,	/* D capital letter  */
	/* 45 Upper E			*/	0x45,	/* E capital letter  */
	/* 46 Upper F			*/	0x46,	/* F capital letter  */
	/* 47 Upper G			*/	0x47,	/* G capital letter  */
	/* 48 Upper H			*/	0x48,	/* H capital letter  */
	/* 49 Upper I			*/	0x49,	/* I capital letter  */
	/* 4a Upper J			*/	0x4a,	/* J capital letter  */
	/* 4b Upper K			*/	0x4b,	/* K capital letter  */
	/* 4c Upper L			*/	0x4c,	/* L capital letter  */
	/* 4d Upper M			*/	0x4d,	/* M capital letter  */
	/* 4e Upper N			*/	0x4e,	/* N capital letter  */
	/* 4f Upper O			*/	0x4f,	/* O capital letter  */

	/* 50 Upper P			*/	0x50,	/* P capital letter  */
	/* 51 Upper Q			*/	0x51,	/* Q capital letter  */
	/* 52 Upper R			*/	0x52,	/* R capital letter  */
	/* 53 Upper S			*/	0x53,	/* S capital letter  */
	/* 54 Upper T			*/	0x54,	/* T capital letter  */
	/* 55 Upper U			*/	0x55,	/* U capital letter  */
	/* 56 Upper V			*/	0x56,	/* V capital letter  */
	/* 57 Upper W			*/	0x57,	/* W capital letter  */
	/* 58 Upper X			*/	0x58,	/* X capital letter  */
	/* 59 Upper Y			*/	0x59,	/* Y capital letter  */
	/* 5a Upper Z			*/	0x5a,	/* Z capital letter  */
	/* 5b			(spare)	*/	0x04,	/* ctrl: EOT	     */
	/* 5c			(spare)	*/	0x05,	/* ctrl: ENQ	     */
	/* 5d Lower y acute		*/	0xfd,	/* y acute	     */
	/* 5e Middle Dot		*/	0xb7,	/* Middle Dot	     */
	/* 5f			(spare)	*/	0x7f,	/* ctrl: delete	     */

	/* 60 Delta Underscore		*/	0x8f,	/* ctrl: reserved    */
	/* 61 A Underscore		*/	0xc1,	/* A acute	     */
	/* 62 B Underscore		*/	0xc2,	/* A circumflex	     */
	/* 63 C Underscore		*/	0xc3,	/* A tilde	     */
	/* 64 D Underscore		*/	0xc7,	/* C cedilla	     */
	/* 65 E Underscore		*/	0xc8,	/* E grave	     */
	/* 66 F Underscore		*/	0xca,	/* E circumflex	     */
	/* 67 G Underscore		*/	0xcb,	/* E diaeresis	     */
	/* 68 H Underscore		*/	0xcc,	/* I grave	     */
	/* 69 I Underscore		*/	0xcd,	/* I acute	     */
	/* 6a J Underscore		*/	0xce,	/* I circumflex	     */
	/* 6b K Underscore		*/	0xcf,	/* I diaeresis	     */
	/* 6c L Underscore		*/	0xd0,	/*   capital ETH     */
	/* 6d M Underscore		*/	0xd2,	/* O grave	     */
	/* 6e N Underscore		*/	0xd3,	/* O acute	     */
	/* 6f O Underscore		*/	0xd4,	/* O circumflex	     */

	/* 70 P Underscore		*/	0xd5,	/* O tilde	     */
	/* 71 Q Underscore		*/	0xd9,	/* U grave	     */
	/* 72 R Underscore		*/	0xda,	/* U acute	     */
	/* 73 S Underscore		*/	0xdb,	/* U circumflex	     */
	/* 74 T Underscore		*/	0xdd,	/* Y acute	     */
	/* 75 U Underscore		*/	0xde,	/*   capital THORN   */
	/* 76 V Underscore		*/	0xe3,	/* a tilde	     */
	/* 77 W Underscore		*/	0xec,	/* i grave	     */
	/* 78 X Underscore		*/	0xf0,	/*   small eth	     */
	/* 79 Y Underscore		*/	0xf2,	/* o grave	     */
	/* 7a Z Underscore		*/	0xf5,	/* o tilde	     */
	/* 7b Left Brace		*/	0x7b,	/* { left brace	     */
	/* 7c Left Tack			*/	0x80,	/* ctrl: reserved    */
	/* 7d Right Brace		*/	0x7d,	/* } right brace     */
	/* 7e RightTack			*/	0x81,	/* ctrl: reserved    */
	/* 7f Squish			*/	0xa6,	/*   broken vert bar */

	/* 80 Each			*/	0xa8,	/* spacing diaeresis */
	/* 81 Upper A grave		*/	0xc0,	/* A grave	     */
	/* 82 Upper A diaeresis		*/	0xc4,	/* A diaeresis	     */
	/* 83 Upper A ring		*/	0xc5,	/* A ring	     */
	/* 84 Upper AE			*/	0xc6,	/* AE		     */
	/* 85 Commute			*/	0xfe,	/* small thorn	     */
	/* 86 Upper E acute		*/	0xc9,	/* E acute	     */
	/* 87 Upper N tilde		*/	0xd1,	/* N tilde	     */
	/* 88 Upper O diaeresis		*/	0xd6,	/* O diaeresis	     */
	/* 89 Upper O slash		*/	0xd8,	/* O slash	     */
	/* 8a Upper U diaeresis		*/	0xdc,	/* U diaeresis	     */
	/* 8b German double-S		*/	0xdf,	/*   small sharp s   */
	/* 8c Lower a grave		*/	0xe0,	/* a grave	     */
	/* 8d Lower a acute		*/	0xe1,	/* a acute	     */
	/* 8e Lower a circumflex	*/	0xe2,	/* a circumflex	     */
	/* 8f Lower a diaeresis		*/	0xe4,	/* a diaeresis	     */

	/* 90 Lower a ring		*/	0xe5,	/* a ring	     */
	/* 91 Lower ae			*/	0xe6,	/* ae		     */
	/* 92 Lower c cedilla		*/	0xe7,	/* c cedilla	     */
	/* 93 Lower e grave		*/	0xe8,	/* e grave	     */
	/* 94 Lower e acute		*/	0xe9,	/* e acute	     */
	/* 95 Lower e circumflex	*/	0xea,	/* e circumflex	     */
	/* 96 Lower e diaeresis		*/	0xeb,	/* e diaeresis	     */
	/* 97 Lower i acute		*/	0xed,	/* i acute	     */
	/* 98 Lower i circumflex	*/	0xee,	/* i circumflex	     */
	/* 99 Lower i diaeresis		*/	0xef,	/* i diaeresis	     */
	/* 9a Lower n tilde		*/	0xf1,	/* n tilde	     */
	/* 9b Left bracket		*/	0x5b,	/* [ left bracket    */
	/* 9c Slash			*/	0x2f,	/* / slash	     */
	/* 9d Slash bar			*/	0x9a,	/* ctrl: reserved    */
	/* 9e Slope			*/	0x5c,	/* \ backslash	     */
	/* 9f Slope bar			*/	0x99,	/* ctrl: reserved    */

	/* a0 Less			*/	0x3c,	/* < less-than sign  */
	/* a1 Less or Equal		*/	0x88,	/* ctrl: reserved    */
	/* a2 Equal			*/	0x3d,	/* = equals sign     */
	/* a3 Greater or Equal		*/	0x89,	/* ctrl: reserved    */
	/* a4 Greater			*/	0x3e,	/* > greater-than    */
	/* a5 Not Equal			*/	0xac,	/*   not sign	     */
	/* a6 Or			*/	0x9f,	/* ctrl: reserved    */
	/* a7 And			*/	0x5e,	/* ^ spacing c'flex  */
	/* a8 Minus			*/	0x2d,	/* - hyphen-minus    */
	/* a9 Plus			*/	0x2b,	/* + plus sign	     */
	/* aa Divide			*/	0xf7,	/*   division sign   */
	/* ab Times			*/	0xd7,	/*   multiply sign   */
	/* ac Query			*/	0x3f,	/* ? question mark   */
	/* ad Epsilon			*/	0xb9,	/*   superscript 1   */
	/* ae Rho			*/	0xbd,	/*   one half	     */
	/* af Tilde			*/	0x7e,	/* ~ tilde	     */

	/* b0 Up Arrow			*/	0x86,	/* ctrl: reserved    */
	/* b1 Down Arrow		*/	0x87,	/* ctrl: reserved    */
	/* b2 Iota			*/	0xbc,	/*   one quarter     */
	/* b3 Circle			*/	0xb1,	/*   plus or minus   */
	/* b4 Star			*/	0x2a,	/* * asterisk	     */
	/* b5 Up stile			*/	0x97,	/* ctrl: reserved    */
	/* b6 Down stile		*/	0x98,	/* ctrl: reserved    */
	/* b7 Del			*/	0x92,	/* ctrl: reserved    */
	/* b8 Jot			*/	0xb0,	/*   degree sign     */
	/* b9 Left parenthesis		*/	0x28,	/* ( left paren	     */
	/* ba Left Shoe			*/	0x9b,	/* ctrl: reserved    */
	/* bb Right Shoe		*/	0x9c,	/* ctrl: reserved    */
	/* bc Up Shoe			*/	0x9d,	/* ctrl: reserved    */
	/* bd Down Shoe			*/	0x9e,	/* ctrl: reserved    */
	/* be Down Tack, Decode		*/	0x83,	/* ctrl: reserved    */
	/* bf Up Tack, Encode		*/	0x82,	/* ctrl: reserved    */

	/* c0 APL Stile			*/	0x7c,	/* | vertical bar    */
	/* c1 Semi-colon		*/	0x3b,	/* ; semicolon	     */
	/* c2 Comma			*/	0x2c,	/* , comma	     */
	/* c3 Nor			*/	0x8b,	/* ctrl: reserved    */
	/* c4 Nand			*/	0x8a,	/* ctrl: reserved    */
	/* c5 Downgrade			*/	0x94,	/* ctrl: reserved    */
	/* c6 Upgrade			*/	0x93,	/* ctrl: reserved    */
	/* c7 Transpose			*/	0xb3,	/*   superscript 3   */
	/* c8 Rotate			*/	0xb2,	/*   superscript 2   */
	/* c9 Rotate bar		*/	0xb4,	/*   spacing acute   */
	/* ca Log			*/	0xb5,	/*   micron sign     */
	/* cb Domino			*/	0x8e,	/* ctrl: reserved    */
	/* cc Shriek			*/	0x21,	/* ! exclamation mark*/
	/* cd Thorn, format		*/	0x95,	/* ctrl: reserved    */
	/* ce Hydrant, execute		*/	0x96,	/* ctrl: reserved    */
	/* cf Del tilde			*/	0x90,	/* ctrl: reserved    */

	/* d0 Cat bar			*/	0xae,	/*   trade mark sign */
	/* d1 Match			*/	0xad,	/*   soft hyphen     */
	/* d2 Natch			*/	0xbb,   /*   right guillemet */
	/* d3 Lower o acute		*/	0xf3,	/* o acute	     */
	/* d4 Lower o circumflex	*/	0xf4,	/* o circumflex	     */
	/* d5 Lower o diaeresis		*/	0xf6,	/* o diaeresis	     */
	/* d6 Lower o slash		*/	0xf8,	/* o slash	     */
	/* d7 Ascii double quote	*/	0x22,	/* " quotation mark  */
	/* d8 Hash			*/	0x23,	/* # number sign     */
	/* d9			(spare)	*/	0x1e,	/* ctrl: RS	     */
	/* da Ampersand			*/	0x26,	/* & ampersand	     */
	/* db Ascii single quote	*/	0x1c,	/* ctrl: FS	     */
	/* dc Line draw lower right	*/	0x10,	/* ctrl: DLE	     */
	/* dd Line draw upper right	*/	0x11,	/* ctrl: DC1	     */
	/* de Line draw upper left	*/	0x12,	/* ctrl: DC2	     */
	/* df Line draw lower left	*/	0x13,	/* ctrl: DC3	     */

	/* e0 Line draw middle cross	*/	0x14,	/* ctrl: DC4	     */
	/* e1 Line draw horizontal bar	*/	0x15,	/* ctrl: NAK	     */
	/* e2 Line draw left  T		*/	0x16,	/* ctrl: SYN	     */
	/* e3 Line draw right T		*/	0x17,	/* ctrl: ETB	     */
	/* e4 Line draw lower T		*/	0x18,	/* ctrl: CAN	     */
	/* e5 Line draw upper T		*/	0x19,	/* ctrl: EM	     */
	/* e6 Line draw vertical bar	*/	0x1a,	/* ctrl: SUB	     */
	/* e7 Ascii at			*/	0x40,	/* @ commercial at   */
	/* e8 Lower u grave		*/	0xf9,	/* u grave	     */
	/* e9 Lower u acute		*/	0xfa,	/* u acute	     */
	/* ea Lower u circumflex	*/	0xfb,	/* u circumflex	     */
	/* eb Ascii caret		*/	0x0f,	/* ctrl: SI	     */
	/* ec Lower u diaeresis		*/	0xfc,	/* u diaeresis	     */
	/* ed Ascii back quote		*/	0x60,	/* ` spacing grave   */
	/* ee Ascii split bar		*/	0x1d,	/* ctrl: GS	     */
	/* ef Paragraph sign		*/	0xb6,	/* paragraph sign    */

	/* f0 Colon			*/	0x3a,	/* : colon	     */
	/* f1 Epsilon underbar (Find)	*/	0xba,	/*   masculine ord'l */
	/* f2 Inverted query		*/	0xbf,	/*   inverted query  */
	/* f3 Inverted shriek		*/	0xa1,	/*   inverted exclm  */
	/* f4 Diamond			*/	0xaa,	/*   feminine ordn'l */
	/* f5 Left Arrow		*/	0x84,	/* ctrl: reserved    */
	/* f6 Right Arrow		*/	0x85,	/* ctrl: reserved    */
	/* f7 Lamp			*/	0xa9,	/*   copyright sign  */
	/* f8 Right parenthesis		*/	0x29,	/* ) right paren     */
	/* f9 Right bracket		*/	0x5d,	/* ] right bracket   */
	/* fa			(spare)	*/	0x1f,	/* ctrl: US	     */
	/* fb			(spare)	*/	0xa0,	/*   non-break space */
	/* fc Section sign		*/	0xa7,	/* section sign	     */
	/* fd Quad			*/	0x8c,	/* ctrl: reserved    */
	/* fe Quote Quad		*/	0x8d,	/* ctrl: reserved    */
	/* ff Lower y diaeresis		*/	0xff	/* y diaeresis	     */

,/*C234*/ 255,  0,  0	// Native File			Red on...
,/*B234*/   0,  0,  0	// Native File			...Black

,/*C235*/ 255,  0,  0	// Native File (Read-only)	Red on...
,/*B235*/ 130,130,130	// Native File (Read-only)	...Dark Grey

,/*C236*/   0,255,  0	// Char Array			Green on...
,/*B236*/   0,  0,  0	// Char Array			...Black

,/*C237*/   0,255,  0	// Char Array (Read-only)	Green on...
,/*B237*/ 130,130,130	// Char Array			...Dark Grey

,/*C238*/ 255,255,255	// Numeric Array			White on...
,/*B238*/   0,  0,  0	// Numeric Array			...Black

,/*C239*/ 255,255,255	// Numeric Array	(Read-only)  	White on...
,/*B239*/ 130,130,130	// Numeric Array (Read-only)	...Dark Grey

,/*C240*/   0,255,255	// Mixed Array			Cyan on ...
,/*B240*/   0,  0,  0	// Mixed Array			...Black

,/*C241*/   0,255,255	// Mixed Array (Read-only)	Cyan on ...
,/*B241*/ 130,130,130	// Mixed Array (Read-only)	...Dark Grey

,/*C242*/   0,255,255	// Nested Array			Cyan on ...
,/*B242*/   0,  0,  0	// Nested Array			...Black

,/*C243*/   0,255,255	// Nested Array (Read-only)	Cyan on ...
,/*B243*/ 130,130,130	// Nested Array (Read-only)	...Dark Grey

,/*C244*/ 255,255,255	// QuadOR Object			White on...
,/*B244*/ 255,  0,  0	// QuadOR Object			...Red	

,/*C245*/ 255,255,255	// QuadOR Object	(Read-only)	White on...
,/*B245*/ 255,  0,  0	// QuadOR Object	(Read-only)	...Red	

,/*C248*/ 255,255,255	// Edit Function	(Read-only)	White on...
,/*B248*/   0,128,128	// Edit Function	(Read-only)	...Dark Cyan

,/*C249*/ 255,  0,  0	// marked line			Red on...
,/*B249*/ 255,255,255	// marked line			...White

,/*C250*/   0,  0,  0	// session			Black on...
,/*B250*/ 255,255,255	// session			...White


,/*C252*/ 255,255,  0	// Suspended Function		Yellow on...
,/*B252*/   0,  0,  0	// Suspended Function		...Black

,/*C253*/ 255,255,  0	// Pendent Function		Yellow on...
,/*B253*/ 128,128,128	// Pendent Function		...Dark Grey

,/*C254*/   0,  0,  0	// Edit Function			White on...
,/*B254*/ 192,192,192	// Edit Function			...Blue

,/*C255*/ 255,255,255	// Hilite			White on...
,/*B255*/ 255,  0,  0	// Hilite			...Red

,/*C0*/  192,192,192	// White (dim)
,/*B0*/    0,  0,  0	// Black
,/*C1*/    0,  0,128	// Dk Blue
,/*B1*/    0,  0,128	// Dk Blue
,/*C2*/    0,128,  0	// Dk Green
,/*B2*/    0,128,  0	// Dk Green
,/*C3*/    0,130,130	// Dk Cyan
,/*B3*/    0,130,130	// Dk Cyan
,/*C4*/  128,  0,  0	// Dk Red
,/*B4*/  128,  0,  0	// Dk Red
,/*C5*/  128,  0,128	// Dk Magenta
,/*B5*/  128,  0,128	// Dk Magenta 
,/*C6*/  128,128,  0	// Dk Yellow
,/*B6*/  128,128,  0	// Dk Yellow 
,/*C7*/    0,  0,  0	// Black
,/*B7*/  192,192,192	// White (dim)
,/*C8*/  255,255,255	// White (Intense)
,/*B8*/  255,255,255	// White (Intense)
,/*C9*/    0,  0,255	// Blue
,/*B9*/    0,  0,255	// Blue
,/*C10*/   0,255,  0	// Green
,/*B10*/   0,255,  0	// Green
,/*C11*/   0,255,255	// Cyan
,/*B11*/   0,255,255	// Cyan
,/*C12*/ 255,  0,  0	// Red
,/*B12*/ 255,  0,  0	// Red
,/*C13*/ 255,  0,255	// Magenta
,/*B13*/ 255,  0,255	// Magenta
,/*C14*/ 255,255,  0	// Yellow
,/*B14*/ 255,255,  0	// Yellow
,/*C15*/ 130,130,130	// Grey
,/*B15*/ 130,130,130	// Grey

,/*C129*/ 255,  0,  0	// Syntax Error			
,/*B129*/ 255,255,255	// Syntax Error		
,/*C130*/   0,  0,255	// Syntax Line Numbers
,/*B130*/ 255,255,255	// Syntax Line Numbers
,/*C131*/   0,  0,128	// Syntax System Function
,/*B131*/ 255,255,255	// Syntax System Function
,/*C132*/   0,  0,255	// Syntax Primitive	
,/*B132*/ 255,255,255	// Syntax Primitive
,/*C133*/   0,  0,128	// Syntax Label	
,/*B133*/ 255,255,255	// Syntax Label	
,/*C134*/ 128,128,128	// Syntax Local Name	
,/*B134*/ 255,255,255	// Syntax Local Name
,/*C135*/   0,  0,  0	// Syntax Global Name
,/*B135*/ 255,255,255	// Syntax Global Name
,/*C136*/ 128,128,128	// Syntax Numeric Const
,/*B136*/ 255,255,255	// Syntax Numeric Const
,/*C137*/   0,128,  0	// Syntax Comment
,/*B137*/ 255,255,255	// Syntax Comment
,/*C138*/ 128,  0,  0	// Syntax Control Str.	
,/*B138*/ 255,255,255	// Syntax Control Str.	
,/*C139*/   0,128,128	// Syntax String Const	
,/*B139*/ 255,255,255	// Syntax String Const	
,/*C140*/ 128,  0,128	// Syntax local Sys Var	
,/*B140*/ 255,255,255	// Syntax local Sys Var	
,/*C141*/   0,  0,  0	// Syntax Global Sys Var
,/*B141*/ 255,255,255	// Syntax Global Sys Var
,/*C142*/ 128,128,128	// Syntax name in Dfn
,/*B142*/ 255,255,255	// Syntax name in Dfn
,/*C143*/   0,  0,255	// D-Op {dyadic}
,/*B143*/ 255,255,255	// D-Op {dyadic}
,/*C144*/   0,  0,255	// D-Op {monadic}
,/*B144*/ 255,255,255	// D-Op {monadic}

,/*C100*/   0,128,  0	// Status Info	
,/*B100*/ 255,255,255	// Status Info	
,/*C101*/ 255,  0,  0	// Status Error
,/*B101*/ 255,255,255	// Status Error
,/*C102*/   0,  0,255	// Status Warning
,/*B102*/ 255,255,255	// Status Warning
,/*CRNL*/ 13, 10	// carriage return, line feed - for file io
};

// 'Raw' String Translation Functions:

static UCH *RawAnsi = 0;		// Raw ANSI input  for ansi2av	
static UCH *RawAnso = av_ansi_dflt;	// Raw ANSI output for av2ansi	

void av2ansi(UCH *cptr, size_t size)
{
	while (size--)
	{
		unsigned ch = *cptr;
		*cptr++ = RawAnso[ch];
	}
}

static void MakeRawXtab(void)
{
	UCH code;
	int indx;

	RawAnsi = malloc(256);

	if(0 == RawAnsi) error(10);	// limit()

	// note that the 1 - 1 nature of the table makes initialisation 
	// unnecessary

	for (indx=256; indx--;)
	{
		code = RawAnso[indx];
		RawAnsi[code] = indx;
	}
}

void ansi2av(UCH *cptr, size_t size)
{
	if (RawAnsi == 0)			// initialise first time used.
		MakeRawXtab();

	while (size--)
	{
		unsigned ch = *cptr;
		*cptr++ = RawAnsi[ch];
	}
}

#define	RAWOFF(n)	(256 + (3 * n))

static struct Sas raw_sas[] =
	{
		{ C0 - 234, RAWOFF( 0), 3 }
	,	{ B0 - 234, RAWOFF( 1), 3 }
	,	{ C0 - 235, RAWOFF( 2), 3 }
	,	{ B0 - 235, RAWOFF( 3), 3 }
	,	{ C0 - 236, RAWOFF( 4), 3 }
	,	{ B0 - 236, RAWOFF( 5), 3 }
	,	{ C0 - 237, RAWOFF( 6), 3 }
	,	{ B0 - 237, RAWOFF( 7), 3 }
	,	{ C0 - 238, RAWOFF( 8), 3 }
	,	{ B0 - 238, RAWOFF( 9), 3 }
	,	{ C0 - 239, RAWOFF(10), 3 }
	,	{ B0 - 239, RAWOFF(11), 3 }
	,	{ C0 - 240, RAWOFF(12), 3 }
	,	{ B0 - 240, RAWOFF(13), 3 }
	,	{ C0 - 241, RAWOFF(14), 3 }
	,	{ B0 - 241, RAWOFF(15), 3 }
	,	{ C0 - 242, RAWOFF(16), 3 }
	,	{ B0 - 242, RAWOFF(17), 3 }
	,	{ C0 - 243, RAWOFF(18), 3 }
	,	{ B0 - 243, RAWOFF(19), 3 }
	,	{ C0 - 244, RAWOFF(20), 3 }
	,	{ B0 - 244, RAWOFF(21), 3 }
	,	{ C0 - 245, RAWOFF(22), 3 }
	,	{ B0 - 245, RAWOFF(23), 3 }
	,	{ C0 - 248, RAWOFF(24), 3 }
	,	{ B0 - 248, RAWOFF(25), 3 }
	,	{ C0 - 249, RAWOFF(26), 3 }
	,	{ B0 - 249, RAWOFF(27), 3 }
	,	{ C0 - 250, RAWOFF(28), 3 }
	,	{ B0 - 250, RAWOFF(29), 3 }
	,	{ C0 - 252, RAWOFF(30), 3 }
	,	{ B0 - 252, RAWOFF(31), 3 }
	,	{ C0 - 253, RAWOFF(32), 3 }
	,	{ B0 - 253, RAWOFF(33), 3 }
	,	{ C0 - 254, RAWOFF(34), 3 }
	,	{ B0 - 254, RAWOFF(35), 3 }
	,	{ C0 - 255, RAWOFF(36), 3 }
	,	{ B0 - 255, RAWOFF(37), 3 }
	,	{ C0 -   0, RAWOFF(38), 3 }
	,	{ B0 -   0, RAWOFF(39), 3 }
	,	{ C0 -   1, RAWOFF(40), 3 }
	,	{ B0 -   1, RAWOFF(41), 3 }
	,	{ C0 -   2, RAWOFF(42), 3 }
	,	{ B0 -   2, RAWOFF(43), 3 }
	,	{ C0 -   3, RAWOFF(44), 3 }
	,	{ B0 -   3, RAWOFF(45), 3 }
	,	{ C0 -   4, RAWOFF(46), 3 }
	,	{ B0 -   4, RAWOFF(47), 3 }
	,	{ C0 -   5, RAWOFF(48), 3 }
	,	{ B0 -   5, RAWOFF(49), 3 }
	,	{ C0 -   6, RAWOFF(50), 3 }
	,	{ B0 -   6, RAWOFF(51), 3 }
	,	{ C0 -   7, RAWOFF(52), 3 }
	,	{ B0 -   7, RAWOFF(53), 3 }
	,	{ C0 -   8, RAWOFF(54), 3 }
	,	{ B0 -   8, RAWOFF(55), 3 }
	,	{ C0 -   9, RAWOFF(56), 3 }
	,	{ B0 -   9, RAWOFF(57), 3 }
	,	{ C0 -  10, RAWOFF(58), 3 }
	,	{ B0 -  10, RAWOFF(59), 3 }
	,	{ C0 -  11, RAWOFF(60), 3 }
	,	{ B0 -  11, RAWOFF(61), 3 }
	,	{ C0 -  12, RAWOFF(62), 3 }
	,	{ B0 -  12, RAWOFF(63), 3 }
	,	{ C0 -  13, RAWOFF(64), 3 }
	,	{ B0 -  13, RAWOFF(65), 3 }
	,	{ C0 -  14, RAWOFF(66), 3 }
	,	{ B0 -  14, RAWOFF(67), 3 }
	,	{ C0 -  15, RAWOFF(68), 3 }
	,	{ B0 -  15, RAWOFF(69), 3 }
	,	{ C0 - 129, RAWOFF(70), 3 }
	,	{ B0 - 129, RAWOFF(71), 3 }
	,	{ C0 - 130, RAWOFF(72), 3 }
	,	{ B0 - 130, RAWOFF(73), 3 }
	,	{ C0 - 131, RAWOFF(74), 3 }
	,	{ B0 - 131, RAWOFF(75), 3 }
	,	{ C0 - 132, RAWOFF(76), 3 }
	,	{ B0 - 132, RAWOFF(77), 3 }
	,	{ C0 - 133, RAWOFF(78), 3 }
	,	{ B0 - 133, RAWOFF(79), 3 }
	,	{ C0 - 134, RAWOFF(80), 3 }
	,	{ B0 - 134, RAWOFF(81), 3 }
	,	{ C0 - 135, RAWOFF(82), 3 }
	,	{ B0 - 135, RAWOFF(83), 3 }
	,	{ C0 - 136, RAWOFF(84), 3 }
	,	{ B0 - 136, RAWOFF(85), 3 }
	,	{ C0 - 137, RAWOFF(86), 3 }
	,	{ B0 - 137, RAWOFF(87), 3 }
	,	{ C0 - 138, RAWOFF(88), 3 }
	,	{ B0 - 138, RAWOFF(89), 3 }
	,	{ C0 - 139, RAWOFF(90), 3 }
	,	{ B0 - 139, RAWOFF(91), 3 }
	,	{ C0 - 140, RAWOFF(92), 3 }
	,	{ B0 - 140, RAWOFF(93), 3 }
	,	{ C0 - 141, RAWOFF(94), 3 }
	,	{ B0 - 141, RAWOFF(95), 3 }
	,	{ C0 - 142, RAWOFF(96), 3 }
	,	{ B0 - 142, RAWOFF(97), 3 }
	,	{ C0 - 143, RAWOFF(98), 3 }
	,	{ B0 - 143, RAWOFF(99), 3 }
	,	{ C0 - 144, RAWOFF(100), 3 }
	,	{ B0 - 144, RAWOFF(101), 3 }
	,	{ C0 - 100, RAWOFF(102), 3 }
	,	{ B0 - 100, RAWOFF(103), 3 }
	,	{ C0 - 101, RAWOFF(104), 3 }
	,	{ B0 - 101, RAWOFF(105), 3 }
	,	{ C0 - 102, RAWOFF(106), 3 }
	,	{ B0 - 102, RAWOFF(107), 3 }

	};

static short unsigned raw_offsets[CC+1] =
	{
	 0,  1,  2,  3,  4,  5,  6,  7,  8,  9
,	10, 11, 12, 13, 14, 15, 16, 17, 18, 19
,	20, 21, 22, 23, 24, 25, 26, 27, 28, 29
,	30, 31, 32, 33, 34, 35, 36, 37, 38, 39
,	40, 41, 42, 43, 44, 45, 46, 47, 48, 49
,	50, 51, 52, 53, 54, 55, 56, 57, 58, 59
,	60, 61, 62, 63, 64, 65, 66, 67, 68, 69
,	70, 71, 72, 73, 74, 75, 76, 77, 78, 79
,	80, 81, 82, 83, 84, 85, 86, 87, 88, 89
,	90, 91, 92, 93, 94, 95, 96, 97, 98, 99
,	100, 101, 102, 103, 104, 105, 106, 107, 108, 109
,	110, 111, 112, 113, 114, 115, 116, 117, 118, 119
,	120, 121, 122, 123, 124, 125, 126, 127, 128, 129
,	130, 131, 132, 133, 134, 135, 136, 137, 138, 139
,	140, 141, 142, 143, 144, 145, 146, 147, 148, 149
,	150, 151, 152, 153, 154, 155, 156, 157, 158, 159
,	160, 161, 162, 163, 164, 165, 166, 167, 168, 169
,	170, 171, 172, 173, 174, 175, 176, 177, 178, 179
,	180, 181, 182, 183, 184, 185, 186, 187, 188, 189
,	190, 191, 192, 193, 194, 195, 196, 197, 198, 199
,	200, 201, 202, 203, 204, 205, 206, 207, 208, 209
,	210, 211, 212, 213, 214, 215, 216, 217, 218, 219
,	220, 221, 222, 223, 224, 225, 226, 227, 228, 229
,	230, 231, 232, 233, 234, 235, 236, 237, 238, 239
,	240, 241, 242, 243, 244, 245, 246, 247, 248, 249
,	250, 251, 252, 253, 254, 255
,	0					// ..., CC
	};

static short unsigned redirected_offsets[CC+1] =
	{
	 0,  1,  RAWOFF(108),  RAWOFF(108),  4,  5,  6,  7,  8,  9
,	10, 11, 12, 13, 14, 15, 16, 17, 18, 19
,	20, 21, 22, 23, 24, 25, 26, 27, 28, 29
,	30, 31, 32, 33, 34, 35, 36, 37, 38, 39
,	40, 41, 42, 43, 44, 45, 46, 47, 48, 49
,	50, 51, 52, 53, 54, 55, 56, 57, 58, 59
,	60, 61, 62, 63, 64, 65, 66, 67, 68, 69
,	70, 71, 72, 73, 74, 75, 76, 77, 78, 79
,	80, 81, 82, 83, 84, 85, 86, 87, 88, 89
,	90, 91, 92, 93, 94, 95, 96, 97, 98, 99
,	100, 101, 102, 103, 104, 105, 106, 107, 108, 109
,	110, 111, 112, 113, 114, 115, 116, 117, 118, 119
,	120, 121, 122, 123, 124, 125, 126, 127, 128, 129
,	130, 131, 132, 133, 134, 135, 136, 137, 138, 139
,	140, 141, 142, 143, 144, 145, 146, 147, 148, 149
,	150, 151, 152, 153, 154, 155, 156, 157, 158, 159
,	160, 161, 162, 163, 164, 165, 166, 167, 168, 169
,	170, 171, 172, 173, 174, 175, 176, 177, 178, 179
,	180, 181, 182, 183, 184, 185, 186, 187, 188, 189
,	190, 191, 192, 193, 194, 195, 196, 197, 198, 199
,	200, 201, 202, 203, 204, 205, 206, 207, 208, 209
,	210, 211, 212, 213, 214, 215, 216, 217, 218, 219
,	220, 221, 222, 223, 224, 225, 226, 227, 228, 229
,	230, 231, 232, 233, 234, 235, 236, 237, 238, 239
,	240, 241, 242, 243, 244, 245, 246, 247, 248, 249
,	250, 251, 252, 253, 254, 255
,	0					// ..., CC
	};

static UCH raw_lengths[CC+1] =
	{
	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
,	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
,	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
,	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
,	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
,	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
,	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
,	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
,	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
,	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
,	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
,	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
,	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
,	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
,	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
,	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
,	0					// ..., CC
	};

static UCH redirected_lengths[CC+1] =
	{
	1,1,2,2,1,1,1,1,1,1,1,1,1,1,1,1
,	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
,	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
,	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
,	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
,	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
,	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
,	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
,	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
,	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
,	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
,	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
,	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
,	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
,	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
,	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
,	0					// ..., CC
	};

static struct Translate raw_output =
	{
	1		// give it an initial refcount then we won't free it
,	{ 0 }		// unique_fid
,	1		// just one mode
#if	HAS_UNICODE
,	0		// no overridden values
,	{ 0 }		// array of 256 zeros
,	8		// utf8
#else //HAS_UNICODE
,	0		// no utf
#endif//HAS_UNICODE
,	(sizeof(raw_sas) / sizeof(struct Sas))		// colours
,	raw_sas
,	raw_offsets
,	raw_lengths
,	av_ansi_dflt
	};

static struct Out_fd raw_fd =
	{
	&raw_output
			// rest of structure is 0
	};

static struct Translate redirected_output =
	{
	1		// give it an initial refcount then we won't free it
,	{ 0 }		// unique_fid
,	1		// just one mode
#if	HAS_UNICODE
,	0		// no overridden values
,	{ 0 }		// array of 256 zeros
,	8		// utf8
#else //HAS_UNICODE
,	0		// no utf
#endif//HAS_UNICODE
,	(sizeof(raw_sas) / sizeof(struct Sas))		// colours
,	raw_sas
,	redirected_offsets
,	redirected_lengths
,	av_ansi_dflt
	};

static result_t redirected_input_result[] = { ER };

static Mmap redirected_input_mode =
	{
	0		// not temporary mode
,	0		// don't strip top bit
,	0		// not a tty - may override this at run time
,	0		// utfwidth
,		{	// all direct except carriage return
		0xff, 0xfb, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
	,	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
	,	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
	,	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
		}
,		{	// actual map - we only keep the output side so:
		0	// fill this in at runtime
		}
,	2		// two rows
,	2		// two cols
,	redirected_input_result
,		{
/*row0*/	13, 10	// standard destruction of junk CR on file
/*row2*/,	13, 0	// hit enter from a keyboard
		}
	};

static Imap redirected_input_map =
	{
		{
		0
		}
,	1		// one mode
,	0		// weak int - not used
,	0		// strong int - not used
,	0		// prog mode - there is only one
,	0		// user mode - there is only one
,	0		// utfwidth - no utf
,	0		// caps on
,	0		// number of labels
,		{	// is_lower - not useful
		0
		}
,		{	// caps xchng - not useful
		0
		}
,		{
		&redirected_input_mode
		}	
,	0
	};

#if	GUI && (defined(MAINWIN) || WINE)

static UCH ScanMap[256];

static void s_fname	// develop an input file name 
	(
	gchar *tp		// null or target file name 
,	gchar *buf		// where to put the file name 
,	unsigned errflag		// is this an error 
	)
	{
	t_fname(tp, GSTR("APLKEYS"), GSTR("aplkeys"), GSTR("APLS  "), GSTR(".SCN"), buf, errflag);
	}

static void ReadScanMap(void)
{
gchar filename[MAXFCHARS];
UCH *buf,*cp,*ep;
ubyte VirtMap[256];
int VirtKey,ScanCode,fh,comma;
int save_test_file = test_file;	// has already been bumped. This code assumes
				// only one stream for keys/scans

mod_fname = 0;			// SCN files don't use MOD: delete history
if
	(
	is_user_io(0)		// input not redirected
&&	shad_getenv("APLS")	// scn file specified
	)
	{
	test_file = 0;
	fname=s_fname; (*fname)(0, filename, 0);

	fh = t_fopen(filename, GSTR("\n"));
	if (-1 == fh)
		x_error(NOSCANFILE);
	else
		{
		cp=buf=alloca(trans_size);
		if (buf) 
			{
			memset(VirtMap,0,256);
			if (trans_size!=read(fh,buf,trans_size))
				(*x_error)(CANT_READ);

			TrimDosFile(buf);
			if(buf[trans_size-1] != '\n')
				(*x_error)(NOT_ASCII);

			while (cp<buf+trans_size)
				{
				comma=VirtKey=ScanCode=0;
				ep=strchr(cp,'\n');
				curr_start_of_line = ep-1;

				cp=white(cp,ep);
				if('+'==*cp)cp=comment(cp,ep);
				if(','==*cp)cp+=comma=1;
				if(cp!=ep) cp=white (cp,ep);
				if(cp!=ep) cp=number(cp,ep),VirtKey=lastnumber;
				if(cp!=ep) cp=white (cp,ep);
				if('+'==*cp)cp=comment(cp,ep);
				if(cp!=ep) cp=number(cp,ep),ScanCode=lastnumber;
				if(cp!=ep) cp=white (cp,ep);
				cp=comment(cp,ep);

				if (VirtKey && ScanCode)
					{
					if (ScanMap[ScanCode])
						(*x_error)(DUP_STRING);

					if (0==comma && VirtMap[VirtKey])
						(*x_error)(DUP_STRING);
					
					ScanMap[ScanCode]=VirtKey;
					VirtMap[VirtKey]=1;
					}
				cp=ep+1;
				}
			}
		else
			(*x_error)(FILE2BIG);
		close(fh);
		}
	}
test_file = save_test_file;

}


int MapScanCode(int sc,int vk)
{
return ScanMap[sc] ? ScanMap[sc] : vk;
}

#else //GUI && (defined(MAINWIN) || WINE)

int MapScanCode(int sc,int vk) { return vk; }

#endif//GUI && (defined(MAINWIN) || WINE)

#ifndef rmemcpy

void rmemcpy(ubyte *to,ubyte * fm,size_t n)				// reverse memcpy	     
{
	to += n;				// NOTE, great benefits will 
	fm += n;				// almost certainly accrue   
						// if this and the next fun  
	while (n--)				// are replaced by corresp.  
		*--to = *--fm;			// library equivalents       
}

#endif // rmemcpy 

#ifndef fmemcpy

void fmemcpy(ubyte *to, ubyte *fm,size_t n)				// forward memcpy	     
{
	while (n--)
		*to++ = *fm++;
}
#endif

size_t key_to_codes(int key, ubyte *codes)
{
Imap *imp = in_fd[0].map;

if (imp)
	{
	uns thismode = in_fd[0].curr_mode[0];
	size_t i;
	uns av = (uns)key;

	Mmap *map= imp->mode[thismode];

	for (i=0;i<256;i++)
		{
		if (map->char_or_index[i]==av)
			{
			if (BIT_READ(map->direct,i))
				{
				codes[0]=(ubyte)i;
				return 1;
				}
			}
		}

	for (i=0;i<map->rows;i++)		
		{
		if (map->result[i]==av)
			{
			ubyte *rp = map->strs + i*map->cols;
			strcpy(codes,rp);
			return strlen(codes);
			}
		}
	}
return 0;
}

bool IsLikeNewLine(uns c)
{
switch (c)
	{
#if HAS_UNICODE
	case 0xd:		// CARRIAGE RETURN
	case 0x85:		// NEL
#if 0
	case 0xc:		// FORM FEED
	case 0xa:		// LINE FEED
	case 0xb:		// VERTICAL TAB
	case 0x2028:		// LINE SEPARATOR
	case 0x2029:		// PARAGRAPH SEPARATOR
#endif
		return 1;
#else
	case AV_NEWLINE:
#if 0
	case AV_LF:
#endif
		return 1;
#endif
	default:
		break;
	}
return 0;
}

int zero_is_input_fd(void)
{
	return 0 != in_fd[0].map;
}

#if	!AP_BUILD

achar LineEnding(void)
{
#if HAS_UNICODE
	return avu_uni(3);	// quadav[3]
#else
	return 3;		// quadav[3]
#endif
}

#endif//!AP_BUILD

#if	0	// some partially developed code for reading win.dot
		// to produce WINCHARS
/*
 B.N.F.	For reading a win.dot (classic)
	<line>		::=<white><lline><newline>
	<lline>		::=<statement>|<implied comment>
	<comment>	::=<null>|+<anything>

	<statement>	::=<number><colon><byte_number><white><comment>
	<white>		::=<null>|<white delim>|<white>
	<white delim>	::=<space>|<tab>|<cr>
	<number>	::=<dec number>|<hex number>
	<byte number>	::=<number>
	<decnumber>	::=<dec digit><dec digits>
	<dec digits>	::=<null>|<dec digit><dec digits>
	<hexnumber>	::=0x<hex digit><hex digits>
	<hex digit>	::=0|1|...|9|a|...|f|A|...|F
	<hex digits>	::=<null>|<hex digit><hex digits>
	<colon>		::=<white><colon><white>

	not expanded
		<null>
		<anything>
		<space>
		<tab>
		<cr>	N.B. not newline
 */

typedef	struct
	{
	uint16_t winchars[256];
	uint16_t bodgechars[256];
	} WIN_BODGE_CHARS;

typedef UCH *WPARSEFN
	(
	WIN_BODGE_CHARS *t
,	UCH *cp
,	UCH *ep
	);

#define W_DECLARE(name)			\
	static UCH *name		\
		(			\
		WIN_BODGE_CHARS *t	\
	,	UCH *cp			\
	,	UCH *ep			\
		)			\

static WPARSEFN
	w_lline
,	w_statement
,	w_header
,	w_str_def
,	w_strand
;

typedef enum
	{
	ILLEGAL_VALUE = 0xffff
	} ILLEGAL_VALUE_t;

W_DECLARE(w_statement)
	{
	UCP wp = cp;
	ubyte avside;
	ubyte fontside;

	cp = byte_number(cp , ep);
	if(wp == cp)
		return cp;			// not a statement
	avside = lastnumber;
	cp = colon(cp, ep);
	wp = cp;
	cp = byte_number(cp, ep);
	fontside = lastnumber;
	if(wp == cp)
		(*x_error)(NUM_WOT);	// didn't get a number
	cp = white(cp, ep);
	cp = comment(cp, ep);
	if(ILLEGAL_VALUE == t->winchars[avside])
		t->winchars[avside] = fontside;
	if(ILLEGAL_VALUE == t->bodgechars[avside])
		t->bodgechars[fontside] = avside;
	}

W_DECLARE(w_lline)
	{
	UCP wp = cp;

	if(wp == cp) cp = w_statement(t, cp, ep);
	if(wp == cp) 
		return ep;	// implied comment
	return cp;
	}

#endif//0

//	Copyright (c) 1982 Dyadic Systems Limited	
